/*
 * This file is part of the "OCAPI" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005 OPeNDAP, Inc.
 * Author: John Chamberlain <jchamberlain at opendap.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>		/* move to utility */
#include "buffers.h"
StringBuffer sbDefault;		/* this is returned if out of memory occurs */

#define DEFAULT_BUFFER_SIZE 80


StringBuffer *sbNew(const char *sStringOrNULL)
{
    unsigned int iBufferSize, iContentLength;
    char *sBuffer;
    StringBuffer *sb;
    if (sStringOrNULL == NULL) {
	iContentLength = 0;
	iBufferSize = DEFAULT_BUFFER_SIZE;
	sBuffer = malloc(iBufferSize);
	if (sBuffer == NULL)
	    return NULL;	/* &sbDefault; out of memory ? */
	*sBuffer = 0;
    } else {
	iBufferSize = strlen(sStringOrNULL) + 1;	/* plus null terminator */
	iContentLength = iBufferSize - 1;
	sBuffer = malloc(iBufferSize);
	if (sBuffer == NULL)
	    return NULL;	/* &sbDefault; out of memory ? */
	strcpy(sBuffer, sStringOrNULL);
    }
    sb = malloc(sizeof(StringBuffer));
    if (sb == NULL)
	return NULL;		/* &sbDefault; out of memory ? */
    sb->iBufferSize = iBufferSize;
    sb->iContentLength = iContentLength;
    sb->sContent = sBuffer;
    return sb;
}

void sbFree(StringBuffer * sb)
{
    if (sb == NULL)
	return;
    if (sb->sContent != NULL)
	free(sb->sContent);
    free(sb);
    return;
}

StringBuffer *sbNew_size(unsigned int iBufferSize)
{
    char *sBuffer;
    StringBuffer *sb = malloc(sizeof(StringBuffer));
    if (sb == NULL)
	return NULL;		/* &sbDefault; out of memory ? */
    if (iBufferSize == 0)
	iBufferSize = DEFAULT_BUFFER_SIZE;	/* default */
    sBuffer = malloc(iBufferSize);
    *sBuffer = 0;		/* null terminator */
    sb->iBufferSize = iBufferSize;
    sb->iContentLength = 0;
    sb->sContent = sBuffer;
    return sb;
}

StringBuffer *sbNew_int(const char *sStringOrNULL, int i)
{
    StringBuffer *sb = sbNew(NULL);
    sbAppend_int(sb, sStringOrNULL, i);
    return sb;
}

void sbClear(StringBuffer * sb)
{
    *(sb->sContent) = 0;
    sb->iContentLength = 0;
}

void sbAppend(StringBuffer * sb, const char *sToBeAppended)
{
    sbInsert_at(sb, sToBeAppended, sb->iContentLength);
}

void sbInsert(StringBuffer * sb, const char *sInsert)
{
    sbInsert_at(sb, sInsert, 0);
}

/* Uses printf to add a string to the message */
void sbInsert_string(StringBuffer * sb, const char *sFormatString,
		     const char *s)
{
    unsigned int lenFormatString, lenSuppliedString, lenCombinedString;
    char *sCombinedString;
    if (sFormatString == NULL)
	sFormatString = "[internal error: no format string]";
    if (s == NULL)
	s = "[null]";
    lenFormatString = strlen(sFormatString);
    lenSuppliedString = strlen(s);
    lenCombinedString = lenFormatString + lenSuppliedString;
    sCombinedString = malloc(lenCombinedString + 1);	/* plus one for null terminator */
    if (sCombinedString == NULL)
	return;
    sprintf(sCombinedString, sFormatString, s);
    *(sCombinedString + lenCombinedString) = 0;	/* terminate */
    sbInsert(sb, sCombinedString);
    free(sCombinedString);
}

/* Uses printf to add an int to the message */
void sbInsert_int(StringBuffer * sb, const char *sFormatString,
		  const int i)
{
    unsigned int lenFormatString, lenSuppliedString, lenCombinedString;
    char *sCombinedString;
    if (sFormatString == NULL)
	sFormatString = "[internal error: no format string]";
    lenFormatString = strlen(sFormatString);
    lenSuppliedString = 48;	/* should be sufficient */
    lenCombinedString = lenFormatString + lenSuppliedString;
    sCombinedString = malloc(lenCombinedString + 1);	/* plus one for null terminator */
    if (sCombinedString == NULL)
	return;
    sprintf(sCombinedString, sFormatString, i);
    *(sCombinedString + lenCombinedString) = 0;	/* terminate */
    sbInsert(sb, sCombinedString);
    free(sCombinedString);
}

void sbInsert_at(StringBuffer * sb, const char *sInsert,
		 unsigned int posInsertionLocation0)
{
    unsigned int lenExistingContent, iAppendLength, iNewBufferSize;
    register unsigned int pos;
    if (sb == NULL)
	return;

    if (posInsertionLocation0 > sb->iContentLength)
	posInsertionLocation0 = sb->iContentLength;
    lenExistingContent = sb->iContentLength;
    iAppendLength = strlen(sInsert);
    iNewBufferSize = sb->iBufferSize;
    if (sb->iBufferSize < lenExistingContent + iAppendLength + 1) {	/* need to enlarge buffer */
	char *sNewBuffer;
	while (iNewBufferSize < lenExistingContent + iAppendLength + 1) {	/* extra space is for accommodation of null terminator */
	    if (iNewBufferSize == 0) {
		iNewBufferSize = iAppendLength + 1;	/* 1 byte is for terminator */
	    } else {
		iNewBufferSize *= 2;
	    }
	}
	sNewBuffer = malloc(iNewBufferSize);
	if (sNewBuffer == NULL)
	    return;		/* out of memory probably */

	memcpy(sNewBuffer, sb->sContent, lenExistingContent);

	*(sNewBuffer + lenExistingContent) = 0;
	free(sb->sContent);	/* free existing content */
	sb->sContent = sNewBuffer;
	sb->iBufferSize = iNewBufferSize;
    }

    /* shift portion after the insertion */
    if (lenExistingContent > 0) {
	pos = lenExistingContent - 1;
	while (1) {
	    if (pos < posInsertionLocation0)
		break;
	    *(sb->sContent + iAppendLength + pos) = *(sb->sContent + pos);
	    if (pos == 0)
		break;
	    pos--;
	}
    }

    /* copy string to be inserted */
    memcpy(sb->sContent + posInsertionLocation0, sInsert, iAppendLength);

    sb->iContentLength += iAppendLength;

    /* ensure null terminator is present */
    *(sb->sContent + sb->iContentLength) = 0;

    return;
}

void sbAppend_char(StringBuffer * sb, const char c)
{
    if (sb == NULL)
	return;
    if (sb->iBufferSize < sb->iContentLength + 1 + 1) {	/* need to enlarge buffer, one for char, one for terminator */
	unsigned int iNewBufferSize = sb->iBufferSize;
	char *sNewBuffer;
#if 0
	register unsigned int pos;
#endif
	while (iNewBufferSize < sb->iContentLength + 1 + 1) {	/* extra space is for accommodation of null terminator */
	    if (iNewBufferSize == 0) {
		iNewBufferSize = 80;	/* allocate a sizable buffer even though only a single char is being appended */
	    } else {
		iNewBufferSize = sb->iBufferSize * 2;
	    }
	}
	sNewBuffer = malloc(iNewBufferSize);
	if (sNewBuffer == NULL)
	    return;		/* out of memory probably */


	memcpy(sNewBuffer, sb->sContent, sb->iContentLength);

	if (sb->sContent != NULL)
	    free(sb->sContent);	/* free existing content */
	sb->sContent = sNewBuffer;
	sb->iBufferSize = iNewBufferSize;
    }
    *(sb->sContent + sb->iContentLength) = c;
    *(sb->sContent + sb->iContentLength + 1) = 0;	/* terminate */
    sb->iContentLength++;
    return;
}

/* Uses printf to add an int to the message */
void sbAppend_int(StringBuffer * sb, const char *sFormatString,
		  const int i)
{
    unsigned int lenFormatString;
    char *sCombinedString;
    if (sFormatString == NULL)
	return;
    lenFormatString = strlen(sFormatString);
    sCombinedString = malloc(lenFormatString + 48);
    if (sCombinedString == NULL)
	return;
    sprintf(sCombinedString, sFormatString, i);
    sbAppend(sb, sCombinedString);
    free(sCombinedString);
}

/* Uses printf to add a long to the message */
void sbAppend_long(StringBuffer * sb, const char *sFormatString,
		   const long n)
{
    unsigned int lenFormatString;
    char *sCombinedString;
    if (sFormatString == NULL)
	return;
    lenFormatString = strlen(sFormatString);
    sCombinedString = malloc(lenFormatString + 48);
    if (sCombinedString == NULL)
	return;
    sprintf(sCombinedString, sFormatString, n);
    sbAppend(sb, sCombinedString);
    free(sCombinedString);
}

/* Uses printf to add a string to the message */
void sbAppend_string(StringBuffer * sb, const char *sFormatString, char *s)
{
    unsigned int lenFormatString, lenSuppliedString, lenCombinedString;
    char *sCombinedString;
    if (sFormatString == NULL)
	return;
    if (s == NULL)
	s = "[null]";
    lenFormatString = strlen(sFormatString);
    lenSuppliedString = strlen(s);
    lenCombinedString = lenFormatString + lenSuppliedString;
    sCombinedString = malloc(lenCombinedString + 1);	/* plus one for null terminator */
    if (sCombinedString == NULL)
	return;
    sprintf(sCombinedString, sFormatString, s);
    *(sCombinedString + lenCombinedString) = 0;	/* terminate */
    sbAppend(sb, sCombinedString);
    free(sCombinedString);
}

/* Uses printf to add a pointer to the message */
void sbAppend_pointer(StringBuffer * sb, const char *sFormatString,
		      void *p)
{
    unsigned int lenFormatString;
    char *sCombinedString;
    if (sFormatString == NULL)
	return;
    lenFormatString = strlen(sFormatString);
    sCombinedString = malloc(lenFormatString + 48);
    if (sCombinedString == NULL)
	return;
    sprintf(sCombinedString, sFormatString, p);
    sbAppend(sb, sCombinedString);
    free(sCombinedString);
}

/* Compares a string to the buffer. Not case-sensitive. Returns 0 if false, 1 if true. */
int sbCompare(StringBuffer * sb, const char *sComparisonString)
{
    char *cSB;
    char *cCompare;
    if (sb == NULL)
	return 0;
    if (sComparisonString == NULL && sb->sContent == NULL)
	return 1;
    if (sComparisonString == NULL || sb->sContent == NULL)
	return 0;
    cSB = sb->sContent;
    cCompare = (char *) sComparisonString;
    for (;;) {
	char charSB = *cSB;
	char charCompare = *cCompare;
	if (charSB > 0x60 && charSB < 0x7B)
	    charSB -= 0x20;	/* convert to upper case */
	if (charCompare > 0x60 && charCompare < 0x7B)
	    charCompare -= 0x20;	/* convert to upper case */
	if (charSB != charCompare)
	    return 0;
	if (charSB == 0 || charCompare == 0)
	    return 1;
	cSB++;
	cCompare++;
    }
}

/* returns NULL if out of memory */
char *sbCopyContent(StringBuffer * sb)
{
    unsigned int len = sb->iContentLength;
    char *sCopy = malloc(len + 1);
    if (sCopy == NULL)
	return NULL;
    memcpy(sCopy, sb->sContent, len);
    *(sCopy + len) = 0;		/* null terminator */
    return sCopy;
}

char *sbCopyContentSection(StringBuffer * sb, unsigned int pos0,
			   unsigned int length)
{
    char *sCopy = malloc(length + 1);
    if (sCopy == NULL)
	return NULL;
    for (pos0 = 0; pos0 < length; pos0++)
	*(sCopy + pos0) = *(sb->sContent + pos0);
    *(sCopy + length) = 0;	/* null terminator */
    return sCopy;
}

void sbDump(StringBuffer * sb, FILE * file)
{
    if (file == NULL)
	file = stderr;
    if (sb == NULL) {
	fputs("[no buffer]", file);
	return;
    }
    fprintf(file, "Buffer %p:\n", (void *) sb);
    fprintf(file, "capacity: %d\n", sb->iBufferSize);
    fprintf(file, "length: %d\n", sb->iContentLength);
    if (sb->sContent == NULL) {
	fprintf(file, "content: [null]\n");
    } else if (*(sb->sContent) == 0) {
	fprintf(file, "content: [blank]\n");
    } else {
	fprintf(file, "content:\n%s\n", sb->sContent);
    }
}

/* if capacity is zero, a default capacity of 80 will be used */
Buffer *bufferNew(unsigned int iCapacity)
{
    unsigned int iContentLength;
    char *sContentBuffer;
    Buffer *pNewBuffer;
    iContentLength = 0;
    if (iCapacity == 0)
	iCapacity = DEFAULT_BUFFER_SIZE;

    // Other code (e.g. findError) does a strlen on
    // sContentBuffer and if the buffer is full, then it
    // will read one byte past the capacity; so add an extra byte.
    sContentBuffer = malloc(iCapacity+1);
    if (sContentBuffer == NULL)
	return NULL;		/* &sbDefault; out of memory ? */
    memset((void*)sContentBuffer, 0, iCapacity+1); /* zero buffer */

    pNewBuffer = malloc(sizeof(Buffer));
    if (pNewBuffer == NULL)
	return NULL;		/* &sbDefault; out of memory ? */
    pNewBuffer->pcharContent = sContentBuffer;
    pNewBuffer->iCapacity = iCapacity;
    pNewBuffer->iContentLength = 0;
    pNewBuffer->iContentPosition0 = 0;
    return pNewBuffer;
}

/* returns the number of appended chars, if it is 0, you ran out of memory */
int bufferAppend(Buffer * pBuffer, char *pDataToAppend,
		 unsigned int posStart, unsigned int iLength)
{
    unsigned int iBytesFree;

    /* resize buffer if necessary */
    iBytesFree =
	pBuffer->iCapacity - pBuffer->iContentLength -
	pBuffer->iContentPosition0;
    if (iLength > iBytesFree) {
	char *pcharReallocatedBuffer;
	unsigned int iNewBufferSize = pBuffer->iCapacity;
	do {
	    iNewBufferSize *= 2;
	    iBytesFree =
		iNewBufferSize - pBuffer->iContentLength -
		pBuffer->iContentPosition0;
	} while (iLength >= iBytesFree);	/* make sure the buffer is always larger than the input */
	pcharReallocatedBuffer = (char *) realloc(pBuffer->pcharContent, iNewBufferSize);	/* realloc automatically frees any old storage */
	if (pcharReallocatedBuffer == NULL) {
	    return 0;
	} else {
	    pBuffer->pcharContent = pcharReallocatedBuffer;
	    pBuffer->iCapacity = iNewBufferSize;
	    {			/* zero buffer */
		/* not done for efficiency reasons */
	    }
	}
    }

    /* copy data to this buffer */
    {
	char *pBufferOffset;
	pBufferOffset =
	    pBuffer->pcharContent + pBuffer->iContentPosition0 +
	    pBuffer->iContentLength;
	memcpy(pBufferOffset, pDataToAppend + posStart, iLength);
	pBuffer->iContentLength += iLength;
    }

    /* copy data to this buffer -- OLD INEFFICIENT VERSION
       {
       unsigned int posBuffer;
       char* pBufferOffset = pBuffer->pcharContent + pBuffer->iContentPosition0 + pBuffer->iContentLength;
       for( posBuffer = 0; posBuffer < iLength; posBuffer++ ) *(pBufferOffset + posBuffer) = *(pDataToAppend + posBuffer + posStart);
       pBuffer->iContentLength += iLength;
       }
     */

    return iLength;
}

/* if the content position is more than halfway past the capacity,
 * or the buffer is empty then any content is moved to the front of the
 * buffer and the position pointer is moved to the beginning
 */
void bufferPack(Buffer * pBuffer)
{
    if (pBuffer == NULL)
	return;
    if (pBuffer->iContentPosition0 > pBuffer->iCapacity / 2) {
	if (pBuffer->iContentLength > 0) {
	    register unsigned int xChar;
	    char *pBufferOffset =
		(char *) pBuffer->pcharContent +
		pBuffer->iContentPosition0;
	    char *pBufferStart = (char *) pBuffer->pcharContent;
	    for (xChar = 0; xChar < pBuffer->iContentLength; xChar++)
		*(pBufferStart + xChar) = *(pBufferOffset + xChar);
	}
	pBuffer->iContentPosition0 = 0;
    }
}
/*places the buffer just packaged on the heap*/
void bufferDump(Buffer * pBuffer, FILE * file, unsigned int posFrom,
		unsigned int posTo)
{
    if (pBuffer == NULL) {
	printf("[buffer is null]");
    } else {
	int zMultiLineContent = 0;	/* false */
	fprintf(file, "buffer capacity: %d\n", pBuffer->iCapacity);
	fprintf(file, "content length: %d\n", pBuffer->iContentLength);
	fprintf(file, "content position: %d\n",
		pBuffer->iContentPosition0);
	if (pBuffer->iContentLength > 60)
	    zMultiLineContent = 1;	/* true */
	fprintf(file, "content: ");
	if (zMultiLineContent == 1)
	    fprintf(file, "\n********************************\n");
	{
	    unsigned int posBuffer;
	    unsigned int iLineChars = 0;
	    unsigned char cCurrent;
	    if (posFrom == 0)
		posFrom = 1;
	    if (posTo == 0)
		posTo = pBuffer->iContentLength;
	    if (posTo < posFrom)
		posTo = posFrom;
	    if (posTo > pBuffer->iContentLength)
		posTo = pBuffer->iContentLength;
	    posBuffer = pBuffer->iContentPosition0 + posFrom - 1;	/* posFrom is one-based */
	    for (; posBuffer < posTo; posBuffer++) {
		cCurrent = *(unsigned char *) (pBuffer->pcharContent + posBuffer);
		if (cCurrent > 32 && cCurrent < 127) {
		    fputc(cCurrent, file);
		    iLineChars++;
		} else {
		    fprintf(file, ".%i", cCurrent);
		    iLineChars += 3;
		}
		if (iLineChars >= 60) {
		    fprintf(file, "\n");
		    iLineChars = 0;
		}
	    }
	}
	if (zMultiLineContent == 1)
	    fprintf(file, "\n********************************\n");
    }
}

char *bufferCopyAsString(Buffer * pBuffer)
{
    register unsigned int pos;
    unsigned int len = pBuffer->iContentLength;
    char *sCopy = malloc(len + 1);
    if (sCopy == NULL)
	return NULL;
    for (pos = 0; pos < len; pos++)
	*(sCopy + pos) = *((char *) pBuffer->pcharContent + pBuffer->iContentPosition0 + pos);
    *(sCopy + len) = 0;		/* null terminator */
    return sCopy;
}

/* returns 0 if the search string does not start the text and 1 if it does
 * ignores leading whitespace
 * not case-sensitive
 */
int bufferStartsWith(Buffer * bufferText, char *sSearchString)
{
    register unsigned int posBuffer;
    unsigned int lenBuffer = bufferText->iContentLength;
    unsigned int posSearch = 0;
    unsigned int lenSearch;
    if (sSearchString == NULL || bufferText == NULL)
	return 0;
    for (posBuffer = 0; posBuffer < lenBuffer; posBuffer++) {
	char c = *((char *) bufferText->pcharContent + bufferText->iContentPosition0 + posBuffer);
	if (c != ' ' && c != '\t' && c != '\r' && c != '\n')
	    break;
    }
    if (posBuffer == lenBuffer)
	return 0;
    lenSearch = strlen(sSearchString);
    for (posSearch = 0; posSearch < lenSearch; posSearch++) {
	char cSearch = *(sSearchString + posSearch);
	char cBuffer =
	    *((char *) bufferText->pcharContent +
	      bufferText->iContentPosition0 + posBuffer);
	if (cSearch > 96 && cSearch < 123)
	    cSearch -= 32;
	if (cBuffer > 96 && cBuffer < 123)
	    cBuffer -= 32;
	if (cSearch != cBuffer)
	    return 0;
	posBuffer++;
    }
    return 1;
}

void vTestInsert(void)
{
    StringBuffer *sb = sbNew(NULL);
    sbAppend(sb, "123456789\n");
    printf(sb->sContent);
    sbInsert_at(sb, "abc", 0);
    printf("abc at 0: %s\n", sb->sContent);
    sbInsert_at(sb, "def", 10);
    printf("def at 10: %s\n", sb->sContent);
    sbInsert_at(sb, "gh", 15);
    printf("gh at 15: %s\n", sb->sContent);
}
/*places the buffer just packaged in the new location*/
Buffer *bufferLoadFile(const char *sFilename)
{
    Buffer *pBuffer;
    unsigned int iFileSize;
    FILE *pFile;
    iFileSize = 0;
    pFile = fopen(sFilename, "rb");
    if (pFile == NULL) {
	return NULL;
    }
    fseek(pFile, 0, SEEK_END);
    iFileSize = ftell(pFile);
    fseek(pFile, 0, SEEK_SET);
    pBuffer = bufferNew((unsigned int) iFileSize);
    if (iFileSize ==
	fread(pBuffer->pcharContent, sizeof(char), iFileSize, pFile)) {
	pBuffer->iContentLength = iFileSize;
    } else {
	return NULL;
    }
    fclose(pFile);
    return pBuffer;
}
