/*
 * This file is part of the "Ocapi" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005 OPeNDAP, Inc.
 * Author: Dennis Heimbigner <jchamberlain at opendap.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */

#include "parse.h"

#define DQUOTE '"'
/*Establish all of the array types*/
eARRAY_TYPE
arrayTypeOf(eNODE_TYPE t)
{
    switch (t) {
    case NODE_Byte: return ARRAY_Byte;
    case NODE_Int16: return ARRAY_Int16;
    case NODE_UInt16: return ARRAY_UInt16;
    case NODE_Int32: return ARRAY_Int32;
    case NODE_UInt32: return ARRAY_UInt32;
    case NODE_Float32: return ARRAY_Float32;
    case NODE_Float64: return ARRAY_Float64;
    case NODE_String: return ARRAY_String;
    case NODE_URL: return ARRAY_URL;
    case NODE_Sequence: return ARRAY_Sequence;
    case NODE_Array: return ARRAY_Array;
    case NODE_Grid: return ARRAY_Grid;
    case NODE_Structure: return ARRAY_Structure;
    case NODE_root:
    default:
    return 0;
    }
}

/* Note: In order to track names that are on the parse stack
   versus dup'd names; the arg name => dup'd, the arg name name0 => stacked.*/
/* Make a NodeContent Array*/
NodeContent_Array*
makeNCArray(eNODE_TYPE t, char* name)
{
    int len;
    NodeContent_Array* nca
        = (NodeContent_Array*)malloc(sizeof(NodeContent_Array));
    nca->sArrayName = name;
    nca->eType = arrayTypeOf(t);
    len = stackDepth(dapdimsizes);
    nca->ctDimensions = len;
    if(nca->ctDimensions > 0) {
        int i;
        nca->sDimensionName1
            = (char**)malloc((len+1)*sizeof(char*)); // starts at 1 not 0
        nca->sDimensionName1[0] = (char*)NULL;
        for(i=0;i<len;i++)
            nca->sDimensionName1[i+1] = (char*)stackGet(dapdimnames,i);
        nca->iDimensionLengthArray1
            = (unsigned int*)malloc((len+1)*sizeof(int));
        nca->iDimensionLengthArray1[0] = 0;
        for(i=0;i<len;i++)
            nca->iDimensionLengthArray1[i+1]
		=(unsigned int)(long)stackGet(dapdimsizes,i);
    } else {
        nca->sDimensionName1 = NULL;
        nca->iDimensionLengthArray1 = NULL;
    }
    nca->array = NULL;
    stackClear(dapdimnames); stackClear(dapdimsizes);
    return nca;
}
/* Make a DAP Node*/
DAPNode*
makeDAPNode(eNODE_TYPE t, char* name, int nsubnodes)
{
    DAPNode* node = (DAPNode*)malloc(sizeof(DAPNode));
    node->eType = t;
    node->nodeSuper = NULL;
    node->sName = name;
    node->nEstimatedSize = 0;
    node->attributes = NULL;
    node->iSubNodeCount = nsubnodes;
    if(node->iSubNodeCount > 0) {
        unsigned int i;
	int stackbase = stackDepth(dapstack) - nsubnodes;// base of the subnodes
        node->nodeSub=(DAPNode**)malloc(sizeof(DAPNode*)*node->iSubNodeCount);
        /* do the subnode threading*/
        for(i=0;i<node->iSubNodeCount;i++) {
	    DAPNode* sub = (DAPNode*)stackGet(dapstack,stackbase+i);
	    sub->nodeSuper = node;
            node->nodeSub[i] = sub;
        }
	stackSetdepth(dapstack,stackbase); /* remove nodes from stack*/
    } else
        node->nodeSub = NULL;
    /* strings and urls appear to require special handling*/
    if(t == NODE_String || t == NODE_URL) {
	NodeContent_String* ncs
	    = (NodeContent_String*)malloc(sizeof(NodeContent_String));
	ncs->iLength = 0;
	ncs->sContent = NULL;
	node->nodeContent = ncs;
    } else
        node->nodeContent = NULL;
    return node;
}
/* Make an Array on the parse stack*/
DAPNode*
makeArray0(char* name0, eNODE_TYPE basetype, int nsubnodes)
{
    DAPNode* node;
    if(stackDepth(dapdimsizes) > 0) {
        NodeContent_Array* nca = makeNCArray(basetype,nulldup(name0));
        node = makeDAPNode(NODE_Array,nulldup(name0),nsubnodes);
        node->nodeContent = nca;
    } else {
        node = makeDAPNode(basetype,nulldup(name0),nsubnodes);
    }
    return node;
}
/* Make an Array*/
void
makeArray(char* name0, eNODE_TYPE basetype)
{
    DAPNode* node = makeArray0(name0,basetype,0);
    stackPush(dapstack,(elem_t)node);
}
/* Make a Structure*/
void
makeStructure(char* name0, int nsubnodes)
{
    DAPNode* node;
    node = makeDAPNode(NODE_Structure,
		       (stackDepth(dapdimsizes) > 0 ? NULL: nulldup(name0)),
		       nsubnodes);
    if(stackDepth(dapdimsizes) > 0) {
	stackPush(dapstack,(elem_t)node);
        node = makeArray0(name0,NODE_Structure,1);
    }
    stackPush(dapstack,(elem_t)node);
}

/*
    Note that the current ocapi parser chokes if the
    basetype of either the ARRAY or the MAPs are
    compound types.
    I choose to handle them by making them subnodes.
    For backward compatibility, if they are not compound,
    then the values and maps fields of the NODEContent_Grid
    object are filled in.
*/
/* Make a Grid*/
void
makeGrid(char* name0, int nmaps)
{
    int i;
    int stackbase;
    int simple;
    DAPNode* node;
	DAPNode *gridarray;
	NodeContent_Grid* ncg;

    /* Grid is just below maps in the stack
       get index into stack of grid*/
    stackbase = stackDepth(dapstack) - (nmaps + 1);
    gridarray = (DAPNode*)stackGet(dapstack,stackbase);
    ncg = (NodeContent_Grid*)malloc(sizeof(NodeContent_Grid));
    /* Check that the grid is a non-compound array*/
    simple = 1; /* assume yes*/
    if(gridarray->eType == NODE_Array) {
        NodeContent_Array* nca
	    = (NodeContent_Array*)gridarray->nodeContent;
        /* Check the ARRAY TYPE*/
        switch (nca->eType) {
        case ARRAY_Array:
        case ARRAY_Sequence:
        case ARRAY_Grid:
        case ARRAY_Structure:
        simple = 0; /* not ok*/
        break;
        default: break;
        }
    }
    if(simple) {
        ncg->values
	    = cloneNCArray((NodeContent_Array*)gridarray->nodeContent);
    } else {
        ncg->values = NULL;
    }
    /* see if all the maps are primitive*/
    simple=1; // assume yes
    for(i=0;i<nmaps;i++) {
        DAPNode* map = (DAPNode*)stackGet(dapstack,stackbase+1+i);
		NodeContent_Array* nca;

        /* All of these better be true arrays:*/
        if(map->eType != NODE_Array) {
			simple = 0; break;
		}
        nca = (NodeContent_Array*)map->nodeContent;
        /* Check the ARRAY TYPE*/
        switch (nca->eType) {
        case NODE_Array:
        case NODE_Sequence:
        case NODE_Grid:
        case NODE_Structure:
        case NODE_root:
        case NODE_List:
        simple = 0; /* not ok*/
        break;
        default: break;
        }
    }
    if(simple) {
        /* Fill in ncg->maps1*/
        ncg->ctMappings = nmaps;
        ncg->maps1
            = (NodeContent_Array**)malloc(sizeof(NodeContent_Array*)
                                          *(1+ncg->ctMappings));
        /* Get the arrays from the maps*/
        ncg->maps1[0] = NULL;
        for(i=0;i<nmaps;i++) {
            DAPNode* map = (DAPNode*)stackGet(dapstack,stackbase+1+i);
            NodeContent_Array* nca
		= cloneNCArray((NodeContent_Array*)map->nodeContent);
            ncg->maps1[i+1] = nca; /* remember: 1 based*/
        }
    } else {
        ncg->ctMappings = 0;
        ncg->maps1 = NULL;
    }
    /* create the grid node or the array*/
    node = makeDAPNode(NODE_Grid,
		       (stackDepth(dapdimsizes)>0?NULL:nulldup(name0)),
		       nmaps+1);
    node->nodeContent = ncg;
    if(stackDepth(dapdimsizes) > 0) {
	stackPush(dapstack,(elem_t)node);
        node = makeArray0(name0,NODE_Grid,1);
    }
    stackPush(dapstack,(elem_t)node);
}

/*
    As with grids and structures, original ocapi parser
    chokes.
*/
void
makeSequence(char* name0, int nsubnodes)
{
    int i;
    int stackbase;
    DAPNode* node;
    NodeContent_Sequence* ncs;
    /* For Sequences, we need to partly fill in the nodecontent
       to be back compatible.*/
    ncs=(NodeContent_Sequence*)malloc(sizeof(NodeContent_Sequence));
    ncs->ctRows = 0;
    ncs->aFieldValue = NULL;
    ncs->ctFields = nsubnodes;
    ncs->aeFieldType = (eNODE_TYPE*)malloc(sizeof(eNODE_TYPE)*nsubnodes);
    /* get index of first subnode in stack*/
    stackbase = stackDepth(dapstack) - nsubnodes;
    for(i=0;i<nsubnodes;i++) {
        ncs->aeFieldType[i]
	    = ((DAPNode*)stackGet(dapstack,stackbase+i))->eType;
    }
    node = makeDAPNode(NODE_Sequence,
		       (stackDepth(dapdimsizes)>0?NULL:nulldup(name0)),
		       nsubnodes);
    node->nodeContent = ncs;
    /* Subsume the Stack to an array if required*/
    if(stackDepth(dapdimnames) > 0) {
        stackPush(dapstack,(elem_t)node); /* Stack becomes array subnode*/
        node = makeArray0(name0,NODE_Sequence,1);
    }
    stackPush(dapstack,(elem_t)node);
}

void add_attributevalue(char* value, eNODE_TYPE etype)
{
    if(!checktype(value,etype)) {
	 dap_parse_error("illegal attribute value type",value,
			dap_line_num);
	 stackPush(dapvalues,(elem_t)NULL);
    } else
       stackPush(dapvalues,(elem_t)nulldup(value));
}

/* the goal of typecheck is to see that the val string can
   be converted to a legal instance of the type.*/
int checktype(char* val, eNODE_TYPE etype)
{
    char* ptr;
    long iv;
    double dv;
    int ok = 1;

    switch (etype) {
    case NODE_Byte:
	iv = strtol(val,&ptr,0); /* 0=>auto determine base*/
	if((iv == 0 && val == ptr) || *ptr != '\0') ok=0;
        else if(iv > DODS_UCHAR_MAX || iv < DODS_UCHAR_MIN) ok=0;
	break;
    case NODE_Int16:
	iv = strtol(val,&ptr,0); /* 0=>auto determine base*/
	if((iv == 0 && val == ptr) || *ptr != '\0') ok=0;
        else if(iv > DODS_SHRT_MAX || iv < DODS_SHRT_MIN) ok=0;
	break;
    case NODE_UInt16:
	iv = strtol(val,&ptr,0); /* 0=>auto determine base*/
	if((iv == 0 && val == ptr) || *ptr != '\0') ok=0;
        else if(iv > DODS_USHRT_MAX || iv < 0) ok=0;
	break;
    case NODE_Int32:
	iv = strtol(val,&ptr,0); /* 0=>auto determine base*/
	if((iv == 0 && val == ptr) || *ptr != '\0') ok=0;
        else if(iv > DODS_INT_MAX || iv < DODS_INT_MIN) ok=0;
	break;
    case NODE_UInt32:
	iv = strtol(val,&ptr,0); /* 0=>auto determine base*/
	if((iv == 0 && val == ptr) || *ptr != '\0') ok=0;
        else if(iv > DODS_UINT_MAX || iv < 0) ok=0;
	break;
    case NODE_Float32:
        /* strtof fails on some machines apparently*/
    case NODE_Float64:
	dv = strtod(val,&ptr);
	if((dv == 0 && val == ptr) || *ptr != '\0') ok=0;
	break;
    case NODE_String:
    case NODE_URL:
	break;
    default: break;
    }
    return ok;
}

void
makeAttr(char* name0, eATTRIBUTE_TYPE etype)
{
    DAPAttribute* attr
	= (DAPAttribute*)malloc(sizeof(DAPAttribute));
    attr->eType = etype;
    attr->sName = nulldup(name0);
    attr->ctValues = stackDepth(dapvalues);
    attr->pValueList = makeValuelist(etype);
    stackPush(dapstack,(elem_t)attr);
}

int
isattribute(void* p)
{
    /* distinguish DAPAttribute instance
       from DAPAttributeTable instance*/
    DAPAttribute* dap = (DAPAttribute*)p;
    unsigned int pa = (unsigned int)(dap->eType);
    /* Check the initial value to see if it looks like an attribute type*/
    if((pa & ATTRMASK) == ATTRTAG) return 1;
    return 0;
}


/* Given a stack containing only DAPAttribute pointers;
   Construct a DAPTable containing them
*/
DAPAttributeTable*
makeAttrtable(Stack* attrs)
{
    int i;
    DAPAttributeTable* table;
    int nsubnodes = stackDepth(attrs);
    if(nsubnodes == 0) return NULL;
    table=(DAPAttributeTable*)malloc(sizeof(DAPAttributeTable));
    table->ctAttributes = nsubnodes;
    table->pNameList = (char**)malloc(sizeof(char*)*nsubnodes);
    table->pAttributeList
	=(DAPAttribute**)malloc(sizeof(DAPAttribute*)*nsubnodes);
    for(i=0;i<nsubnodes;i++) {
        DAPAttribute* attr = (DAPAttribute*)stackGet(attrs,i);
        table->pAttributeList[i] = attr;
        table->pNameList[i] = nulldup(attr->sName);
    }
    return table;
}


/*
The stack at this point contains a mix of two
kinds of nodes
1. DAPAttribute* if it is a pure attribute instance
2. DAPNode* if it is a set of attributes

collectAttributes does the following:
1. collect all the DAPAttributes (if any) into a DAPAttributeTable.
2. remove DAPAttributes from the stack
3. return the created table and the revised nsubnodes
*/

DAPAttributeTable*
collectAttributes(int* pnsubnodes)
{
    int i,j;
    int stackbase;
    int nsubnodes = *pnsubnodes;
    DAPAttributeTable* table;
    stackbase = stackDepth(dapstack) - nsubnodes;
    /* collect non-table attributes*/
    stackClear(daptemp);
    for(i=0,j=0;i<nsubnodes;i++) {
	void* p = (void*)stackGet(dapstack,stackbase+i);
	if(!isattribute(p)) j++; else stackPush(daptemp,(elem_t)p);
    }
    table = makeAttrtable(daptemp);
    /* collect table attributes*/
    stackClear(daptemp);
    for(i=0,j=0;i<nsubnodes;i++) {
	void* p = (void*)stackGet(dapstack,stackbase+i);
	if(!isattribute(p)) stackPush(daptemp,(elem_t)p);
    }
    /* pop nodes off stack; subnodes are still in temp*/
    stackSetdepth(dapstack,stackbase);
    /* push table nodes back onto the stack*/
    for(i=0;i<stackDepth(daptemp);i++) stackPush(dapstack,stackGet(daptemp,i));
    *pnsubnodes = stackDepth(daptemp);
    return table;
}

/* Invoke makeAttrnode*/
void
makeAttrnode(char* name0, int stackbase)
{
    int nsubnodes = stackDepth(dapstack) - stackbase;
    DAPAttributeTable* table = collectAttributes(&nsubnodes);
    DAPNode* node = makeDAPNode(NODE_Structure,nulldup(name0),nsubnodes);
    node->attributes = table;
    stackPush(dapstack,(elem_t)node);
}

/*
The current OCAPI storage mechanism for attributes
values is rather baroque.  It is a two level mechanism
in which there is a vector of pointers to the individual
values in the Stack of values assigned to the
attribute. The memory for each value in the Stack is
separately malloc'd.
*/
/* Invoke makeValuelist*/
void**
makeValuelist(eATTRIBUTE_TYPE  eat)
{
    int count;
    void** ptrvec;
    int i;
    long iv;
    float f;
    double d;
    char* ptr;
    char* sval;

    count  = stackDepth(dapvalues);
    ptrvec = (void**)malloc(count*sizeof(void*));
    for(i=0;i<count;i++) {
        sval = (char*)stackGet(dapvalues,i);
	switch (eat) {
	case ATTRIBUTE_Byte: {
	    unsigned char* p = (unsigned char*)malloc(sizeof(unsigned char));
	    iv = strtol(sval,&ptr,0);
	    *p = (unsigned char)iv;
	    ptrvec[i] = p;
	} break;
	case ATTRIBUTE_Int16: {
	    short* p = (short*)malloc(sizeof(short));
	    iv = strtol(sval,&ptr,0);
	    *p = (short)iv;
	    ptrvec[i] = p;
	} break;
	case ATTRIBUTE_UInt16: {
	    unsigned short* p
		= (unsigned short*)malloc(sizeof(unsigned short));
	    iv = strtol(sval,&ptr,0);
	    *p = (unsigned short)iv;
	    ptrvec[i] = p;
	} break;
	case ATTRIBUTE_Int32: {
	    int* p = (int*)malloc(sizeof(int));
	    iv = strtol(sval,&ptr,0);
	    *p = (int)iv;
	    ptrvec[i] = p;
	} break;
	case ATTRIBUTE_UInt32: {
	    unsigned int* p = (unsigned int*)malloc(sizeof(unsigned int));
	    iv = strtol(sval,&ptr,0);
	    *p = (unsigned int)iv;
	    ptrvec[i] = p;
	} break;
	case ATTRIBUTE_Float32: {
	    float* p = (float*)malloc(sizeof(float));
	    /* strtof fails on some machines apparently*/
	    d = strtod(sval,&ptr);
	    f = (float)d;
	    *p = f;
	    ptrvec[i] = p;
	} break;
	case ATTRIBUTE_Float64: {
	    double* p = (double*)malloc(sizeof(float));
	    d = strtod(sval,&ptr);
	    *p = d;
	    ptrvec[i] = p;
	} break;
	case ATTRIBUTE_String:
	case ATTRIBUTE_URL: {
	    char* str = NULL;
	    /* Strip the surrounding quotes.*/
	    if(sval != NULL) {
		int len = strlen(sval);
		str = strdup(sval);
		if(sval[0] == DQUOTE && sval[len-1] == DQUOTE) {
		  int i;
		  for(i=0;i<len-2;i++) str[i] = sval[i+1];
		  str[len-2] = '\0';
		}
	    }
	    ptrvec[i] = str;
	  } break;
	  default: break;
	}
	free(sval);
    }
    return ptrvec;
}
/*to copy the NC array made above*/
NodeContent_Array*
cloneNCArray(NodeContent_Array* nca)
{
    int len;
    NodeContent_Array* clone
        = (NodeContent_Array*)malloc(sizeof(NodeContent_Array));
    clone->sArrayName = nulldup(nca->sArrayName);
    clone->eType = nca->eType;
    len = nca->ctDimensions;
    clone->ctDimensions = len;
    if(len > 0) {
        int i;
        clone->sDimensionName1
            = (char**)malloc((len+1)*sizeof(char*)); // starts at 1 not 0
        clone->sDimensionName1[0] = (char*)NULL;
        for(i=1;i<=len;i++) {
	    char* s = nca->sDimensionName1[i];
            clone->sDimensionName1[i] = nulldup(s);
	}
        clone->iDimensionLengthArray1
            = (unsigned int*)malloc((len+1)*sizeof(int));
        clone->iDimensionLengthArray1[0] = 0;
        for(i=1;i<=len;i++)
            clone->iDimensionLengthArray1[i]
		= nca->iDimensionLengthArray1[i];
    } else {
        clone->sDimensionName1 = NULL;
        clone->iDimensionLengthArray1 = NULL;
    }
    clone->array = nca->array;
    return clone;
}

void
makeRoot(eSTRUCTURE_TYPE etype, char* name0, int nsubnodes)
{
    DAPNode* root;
    DAPObject* oroot;
    root = makeDAPNode(NODE_root,nulldup(name0),nsubnodes);
    oroot = (DAPObject*)malloc(sizeof(DAPObject));
    oroot->eStructureType = etype;
    oroot->iEstimatedSize = 0;
    oroot->nodeRoot = root;
    stackPush(dapstack,(elem_t)oroot);
}

void
makeAttrroot(int stackbase)
{
    int nsubnodes = stackDepth(dapstack) - stackbase;
    DAPAttributeTable* table = collectAttributes(&nsubnodes);
    if(table != NULL) {
        /* It is not quite clear how OCAPI handles the case
           of top level non-nested attributes.
           Assume they are accumulated into an extra node*/
        DAPNode* singletons = makeDAPNode(NODE_Structure,NULL,0);
        singletons->attributes = table;
        stackPush(dapstack,(elem_t)singletons);
	nsubnodes++;
    }
    /* In any case make the root node*/
    makeRoot(STRUCTURE_DAS,"Attributes",nsubnodes);
}

int
checkdupattr(int stackbase)
{
    int i;
    int nsubnodes = stackDepth(dapstack) - stackbase;
    DAPAttribute* attr = (DAPAttribute*)stackTop(dapstack);
    for(i=0;i<nsubnodes-1;i++) {
        DAPAttribute* attr2;
	void* p = (void*)stackGet(dapstack,stackbase+i);
	if(!isattribute(p)) continue;
	attr2 = (DAPAttribute*)p;
	if(attr->sName == NULL && attr2->sName == NULL) return 1;
	if(attr->sName == NULL || attr2->sName == NULL) return 0;
	if(strcmp(attr->sName,attr2->sName)==0) return 1;
    }
    return 0;
}
