/* 
 * This file is part of the "Ocapi" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005 OPeNDAP, Inc.
 * Author: John Chamberlain <jchamberlain at opendap.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */

/*
   Scanner for the DDS. This file works with gnu's flex scanner generator. It
   returns either a defined token such as DATASET, LIST, SEQUENCE, etc.
   or one of the single character tokens `{', `}', `;', `=' or `\n' as
   integers. In the case of an ID, the scanner stores a pointer to the lexeme
   in yylval (whose type is char *).

   The scanner discards all comment text.

   The scanner is not reentrant, but can share name spaces with other
   scanners.
   
   Note:
   1) The `defines' file dds.tab.h is built using `bison -d'.
   2) Define YY_DECL such that the scanner is called `ddslex'.
   3) When bison builds the dds.tab.h file, it uses `dds' instead of `yy' for
   variable name prefixes (e.g., yylval --> ddslval).
   
   jhrg 8/29/94 
*/

%{

#include "parse.h"
#include "dap.tab.h"

/* #include "escaping.h" necesssary ? */

#ifdef YY_PROTO
#define YY_DECL int daplex YY_PROTO(( void ))
#else
#define YY_DECL int daplex (void)
#endif
#define YY_INPUT(buf,result,max_size)(result = OCAPI_Input( buf, max_size ))
#define YY_FATAL_ERROR(msg) fatal_lexer_error(msg);

extern ParserBuffer* dapParserState;

int dap_datahdr = 0;
int dap_line_num = 1;
int dap_start_line = 1;

extern void fatal_lexer_error( char* msg );
extern void vResetScanner();

void store_word();
void store_string();
%}

%option noyywrap
%option nounput
%x comment
%x quote

BYTE 		BYTE|Byte|byte
INT16 		INT16|Int16|int16
UINT16 		UINT16|UInt16|uint16
INT32 		INT32|Int32|int32
UINT32 		UINT32|UInt32|uint32
FLOAT32 	FLOAT32|Float32|float32
FLOAT64 	FLOAT64|Float64|float64
STRING 		STRING|String|string
URL 		URL|Url|url

DATASET 	DATASET|Dataset|dataset 
SEQUENCE 	SEQUENCE|Sequence|sequence
STRUCTURE 	STRUCTURE|Structure|structure
GRID 		GRID|Grid|grid
ARRAY 		ARRAY|Array|array
MAPS 		MAPS|Maps|maps
ATTR 	attributes|Attributes|ATTRIBUTES
ALIAS   ALIAS|Alias|alias
DATAHDR   Data:[\r]?\n

/* See das.lex for comments about the characters allowed in a WORD.
   10/31/2001 jhrg */

WORD        [-+a-zA-Z0-9_/%.\\*][-+a-zA-Z0-9_/%.\\#*]*

NEVER       [^\-+a-zA-Z0-9_/%.\\#;,(){}[\]]

%%

"{"         return (int)*yytext;
"}"         return (int)*yytext;
"["         return (int)*yytext;
"]"         return (int)*yytext;
":"         return (int)*yytext;
";"         return (int)*yytext;
"="         return (int)*yytext;
","         return (int)*yytext;

{BYTE}			store_word(); return SCAN_BYTE;
{INT16}			store_word(); return SCAN_INT16;
{UINT16}		store_word(); return SCAN_UINT16;
{INT32}			store_word(); return SCAN_INT32;
{UINT32}		store_word(); return SCAN_UINT32;
{FLOAT32}		store_word(); return SCAN_FLOAT32;
{FLOAT64}		store_word(); return SCAN_FLOAT64;
{STRING}		store_word(); return SCAN_STRING;
{URL}			store_word(); return SCAN_URL;
{DATASET}		store_word(); return SCAN_DATASET;
{SEQUENCE}		store_word(); return SCAN_SEQUENCE;
{STRUCTURE}		store_word(); return SCAN_STRUCTURE;
{GRID}			store_word(); return SCAN_GRID;
{ARRAY}			store_word(); return SCAN_ARRAY;
{MAPS}			store_word(); return SCAN_MAPS;
{ATTR}			store_word(); return SCAN_ATTR;
{ALIAS}			store_word(); return SCAN_ALIAS;

{WORD}          store_word(); return SCAN_WORD;

[ \t\r]+
\n                  ++dap_line_num;
<INITIAL><<EOF>>    yy_init = 1; dap_line_num = 1; yyterminate();

"#"                 BEGIN(comment);
<comment>[^\n]*
<comment>\n         ++dap_line_num; BEGIN(INITIAL);
<comment><<EOF>>    yy_init = 1; dap_line_num = 1; yyterminate();

\"			BEGIN(quote); dap_start_line = dap_line_num; yymore();
<quote>[^"\r\n\\]*	yymore();
<quote>[^"\r\n\\]*\n	yymore(); ++dap_line_num;
<quote>[^"\r\n\\]*\r\n	yymore(); ++dap_line_num;
<quote>\\.		yymore();
<quote>\"		{ 
    			  BEGIN(INITIAL); 
			  store_string();
			  return SCAN_WORD;
                        }
<quote><<EOF>>		{
                          char msg[256];
			  sprintf(msg,
				  "Unterminated quote (starts on line %d)\n",
				  dap_start_line);
			  YY_FATAL_ERROR(msg);
                        }

{DATAHDR}		{dap_datahdr = 1;
			 yyterminate();}

{NEVER}  {
			if( yytext ){	/* suppress msgs about `' chars */
			fprintf(stderr, "Character %x is not allowed and has been ignored\n", *yytext);
         }
}
%%

void store_word(){
    strncpy(daplval.word, yytext, ID_MAX-1);
    daplval.word[ID_MAX-1] = '\0';
}

int toOctal(int c)
{
    switch (c) {
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7':
        return (c - '0');
    default: break;
    }
    return -1;
}

int toHex(int c)
{
    switch (c) {
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8':case '9':
        return (c - '0');
    case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
        return ((c - 'a') + 0x0a);
    case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
        return ((c - 'A') + 0x0a);
    default: break;
    }
    return -1;
}

void store_string(){
    int i;
    int stop=0;
    int complete=0;
    int c;
    char* p = yytext;
    p++; // skip leading double quote
    for(i=0;i<ID_MAX-1;i++) {
	c = *p++;
        if(c == '"') {
	    complete=1;
	    break;
	} else if(c == '\\') {
	    c = *p++;
	    if(c == '\0') break;
	    // extend as needed
	    switch (c) {
            case 'r': c = '\r'; break;
            case 'n': c = '\n'; break;
            case 't': c = '\t'; break;
            case 'v': c = '\v'; break;
            case 'f': c = '\f'; break;
	    case 'x': { // collect two hex digits
		int d0,d1;
	        d0 = *p++; if(d0=='\0') {stop=1; break;}
	        d1 = *p++; if(d1=='\0') {stop=1; break;}
		d0 = toHex(d0);
		d1 = toHex(d1);
		if(d0 < 0 || d1 < 0) {stop=1; break;}
		c = (d0<<4 | d1);
	    }; break;
	    case '0': { // collect three octal digits
		int d0,d1,d2;
	        d0 = *p++; if(d0=='\0') {stop=1; break;}
	        d1 = *p++; if(d1=='\0') {stop=1; break;}
	        d2 = *p++; if(d2=='\0') {stop=1; break;}
		d0 = toOctal(d0);
		d1 = toOctal(d1);
		d2 = toOctal(d2);
		if(d0 < 0 || d1 < 0 || d2 < 0) {stop=1; break;}
		c = (d0<<6 | d1<<3 | d2);
	    }; break;
	    default: break;	    
	    }
	}
	if(stop) break;
	daplval.word[i]=(char)c;
    }      
    daplval.word[i] = '\0';
    if(!complete)
	dap_parse_error("Malformed string",yytext,dap_line_num);
}

void vResetScanner(){
	daprestart(NULL);
	BEGIN(0);
}

void fatal_lexer_error( char* msg ){
	fputs( msg, stderr );
        dapin = NULL;
}
