/*
 * This file is part of the "OCAPI" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005 OPeNDAP, Inc.
 * Author: John Chamberlain <jchamberlain at opendap.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */

/* #define DODS_DEBUG */

/*#include "config.h"*/


#include "global.h"
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_RPC_TYPES_H
#include <rpc/types.h>
#endif

#if HAVE_RPC_XDR_H
#include <rpc/xdr.h>
#else
#include <xdr.h>
#endif


#define macroQuadPad( i ) ( i%4 == 0 ? 0 : 4 - (i % 4) )	/* XDR frames always have a number of bytes divisible by 4 */

#define	DODS_INT_MAX	2147483647	/* from DODSLimits.h */
const unsigned int MAX_STRING_LENGTH = 32767;	/* = DODS_UINT_MAX-1; from DODSLimits.h */
const int DODS_MAX_ARRAY = DODS_INT_MAX;	/* from Array.h in dods/dap */
const unsigned char SEQUENCE_MARKER_StartOfInstance = 0x5A;
const unsigned char SEQUENCE_MARKER_EndOfSequence = 0xA5;

eRESULT read_Node(DAPNode * nodeTemplate, void **nodeContent_Supplied,
		  Buffer * pBuffer, StringBuffer * sbError);

eRESULT read_String(NodeContent_String * pStringContent, Buffer * pBuffer,
		    StringBuffer * sbError);
eRESULT read_Array(NodeContent_Array * pArray, Buffer * pBuffer,
		   StringBuffer * sbError);
eRESULT read_Grid(NodeContent_Grid * pGridContent, Buffer * pBuffer,
		  StringBuffer * sbError);
eRESULT read_Sequence(DAPNode * nodeSequence,
		      NodeContent_Sequence * pNodeContent_Sequence,
		      Buffer * pBuffer, StringBuffer * sbError);
eRESULT read_Record(DAPNode * nodeSequence,
		    NodeContent_Sequence * pNodeContent_Sequence,
		    int xRow0, Buffer * pBuffer, StringBuffer * sbError);
eRESULT read_String_raw(char **hString, Buffer * pBuffer,
			StringBuffer * sbError);
void *clone_NodeTemplate(DAPNode * pnodeTemplate, StringBuffer * sbError);
eRESULT clone_ArrayTemplate(DAPNode * pTemplateNode,
			    NodeContent_Array ** hContent_array,
			    StringBuffer * sbError);
int zIsPrimitiveType(eNODE_TYPE node_type);

extern int zIsInsideSequence(DAPNode * node);	/* defined in ocapi.c */
extern char *render_NODE_TYPE(eNODE_TYPE e);	/* defined in ocapi.c */

/*this is called in zFile_LoadStructure, zConnection_Load, read_Record*/
eRESULT read_Node(DAPNode * nodeTemplate, void **nodeContent_Supplied,
		  Buffer * pBuffer, StringBuffer * sbError)
{
    void **hNodeContent;
    unsigned int xSubNode, iSubNodeCount_Template;
    XDR xhandle;
    unsigned int ctReadBytes;	/* total number of bytes read by XDR */
    DBG(printf("In read_Node: received buffer at %p with address %p\n", (void *) pBuffer,
    			(void *) &pBuffer));

    if (nodeTemplate == NULL) {
	return SUCCESS;
    }

    iSubNodeCount_Template = nodeTemplate->iSubNodeCount;
    if (nodeTemplate->eType == NODE_root) {
    	for (xSubNode = 1; xSubNode <= iSubNodeCount_Template; xSubNode++) {
    		DAPNode *pSubNode = *(nodeTemplate->nodeSub + (xSubNode - 1));
    		if (pSubNode == NULL) {
    			sbAppend_int(sbError, "invalid root node, template sub-node %d does not exist",
			     xSubNode);
    			return FAILURE;
    		}
    		else {
    			if (read_Node(pSubNode, NULL, pBuffer, sbError) == FAILURE) {
    				sbInsert_string(sbError, "%s\": ", pSubNode->sName);
    				sbInsert_int(sbError, "error reading sub-node %d of root named \"",
    						xSubNode);
    				return FAILURE;
    			}
    		}
    	}
	return SUCCESS;
    }

    if (nodeContent_Supplied == NULL) {

    	/* If there is no supplied content and this is a sequence then the content
    	 * will be that of the template node. In this case it should be verified
    	 * that this is the root of the sequence. */
    	if (nodeTemplate->eType == NODE_Sequence) {
    		if (zIsInsideSequence(nodeTemplate)) {
    			sbAppend(sbError,
    			"internal inconsistency, attempt to read a nested sequence but no content node was supplied");
    			return FAILURE;
    		}
    		/* it is the base node of the sequence */
    		else {
    			hNodeContent = &(nodeTemplate->nodeContent);
    		}
    	}
    	else {
    		hNodeContent = &(nodeTemplate->nodeContent);
    	}
    }
    else {

	hNodeContent = nodeContent_Supplied;

    }
    DBG(int xByte;
	printf("reading node %s, bytes from buffer at %p:\n", nodeTemplate->sName, (void *) pBuffer);
	for (xByte = pBuffer->iContentPosition0; xByte < pBuffer->iContentPosition0 + 16; xByte++) {
		printf("  %02i: %02u", xByte, *((unsigned char *) pBuffer->pcharContent + xByte));
		if (xByte % 8 == 0) printf("\n");}
	printf("\n");)

	switch (nodeTemplate->eType) {
	case NODE_Byte:;
	    ctReadBytes = 4; /* bytes are sent as 4 byte packages (yup) */
	    *hNodeContent = malloc(ctReadBytes);
	    if (*hNodeContent == NULL) {
		sbAppend(sbError, "unable to allocate byte buffer");
		return FAILURE;
	    }
#if 1
	{
			xdrmem_create(&xhandle, pBuffer->pcharContent + pBuffer->iContentPosition0,
					ctReadBytes, XDR_DECODE);
			if (xdr_char(&xhandle, *hNodeContent) == TRUE) {
				 pBuffer->iContentPosition0 += ctReadBytes + macroQuadPad(ctReadBytes);
				 pBuffer->iContentLength -= ctReadBytes + macroQuadPad(ctReadBytes);
			} else {
			    sbAppend(sbError,"failed to read DBYTE in field");
				return FAILURE;
			}
	    }
	    break;
#else
    {
    	*(unsigned char *) (*hNodeContent) = *(unsigned char *) (pBuffer->pcharContent + pBuffer->iContentPosition0);	/* copy the value in the buffer to the content memory */
        pBuffer->iContentPosition0 += ctReadBytes + macroQuadPad(ctReadBytes);
        pBuffer->iContentLength -= ctReadBytes + macroQuadPad(ctReadBytes);
        break;
#endif

    case NODE_Int16:;
	    ctReadBytes = 4;
	    *hNodeContent = malloc(ctReadBytes);
	    if (*hNodeContent == NULL) {
		sbAppend(sbError,
			 "unable to allocate signed short buffer");
		return FAILURE;
	    }
	    xdrmem_create(&xhandle,
			  pBuffer->pcharContent +
			  pBuffer->iContentPosition0, ctReadBytes,
			  XDR_DECODE);
	    if (xdr_short(&xhandle, *hNodeContent) == TRUE) {
		pBuffer->iContentPosition0 +=
		    ctReadBytes + macroQuadPad(ctReadBytes);
		pBuffer->iContentLength -=
		    ctReadBytes + macroQuadPad(ctReadBytes);
	    } else {
		sbAppend(sbError, "short data translation error");
		return FAILURE;
	    }
	    break;
	case NODE_UInt16:;
	    ctReadBytes = 4;
	    *hNodeContent = malloc(ctReadBytes);
	    if (*hNodeContent == NULL) {
		sbAppend(sbError,
			 "unable to allocate unsigned short buffer");
		return FAILURE;
	    }
	    xdrmem_create(&xhandle,
			  pBuffer->pcharContent +
			  pBuffer->iContentPosition0, ctReadBytes,
			  XDR_DECODE);
	    if (xdr_u_short(&xhandle, *hNodeContent) == TRUE) {
		pBuffer->iContentPosition0 +=
		    ctReadBytes + macroQuadPad(ctReadBytes);
		pBuffer->iContentLength -=
		    ctReadBytes + macroQuadPad(ctReadBytes);
	    } else {
		sbAppend(sbError, "ushort data translation error");
		return FAILURE;
	    }
	    break;
	case NODE_Int32:;
	    ctReadBytes = 4;
	    *hNodeContent = malloc(ctReadBytes);
	    if (*hNodeContent == NULL) {
		sbAppend(sbError, "unable to allocate signed int buffer");
		return FAILURE;
	    }
	    xdrmem_create(&xhandle,
			  pBuffer->pcharContent +
			  pBuffer->iContentPosition0, ctReadBytes,
			  XDR_DECODE);
	    if (xdr_int(&xhandle, *hNodeContent) == TRUE) {
		pBuffer->iContentPosition0 +=
		    ctReadBytes + macroQuadPad(ctReadBytes);
		pBuffer->iContentLength -=
		    ctReadBytes + macroQuadPad(ctReadBytes);
	    } else {
		sbAppend(sbError, "int data translation error");
		return FAILURE;
	    }
	    break;
	case NODE_UInt32:;
	    ctReadBytes = 4;
	    *hNodeContent = malloc(ctReadBytes);
	    if (*hNodeContent == NULL) {
		sbAppend(sbError,
			 "unable to allocate unsigned int buffer");
		return FAILURE;
	    }
	    xdrmem_create(&xhandle,
			  pBuffer->pcharContent +
			  pBuffer->iContentPosition0, ctReadBytes,
			  XDR_DECODE);
	    if (xdr_u_int(&xhandle, *hNodeContent) == TRUE) {
		pBuffer->iContentPosition0 +=
		    ctReadBytes + macroQuadPad(ctReadBytes);
		pBuffer->iContentLength -=
		    ctReadBytes + macroQuadPad(ctReadBytes);
	    } else {
		sbAppend(sbError, "uint data translation error");
		return FAILURE;
	    }
	    break;
	case NODE_Float32:;
	    ctReadBytes = 4;
	    *hNodeContent = malloc(ctReadBytes);
	    if (*hNodeContent == NULL) {
		sbAppend(sbError, "unable to allocate float32 buffer");
		return FAILURE;
	    }
	    xdrmem_create(&xhandle,
			  pBuffer->pcharContent +
			  pBuffer->iContentPosition0, ctReadBytes,
			  XDR_DECODE);
	    if (xdr_float(&xhandle, *hNodeContent) == TRUE) {
		pBuffer->iContentPosition0 += ctReadBytes;
		pBuffer->iContentLength -= ctReadBytes;
	    } else {
		sbAppend(sbError, "float32 data translation error");
		return FAILURE;
	    }
	    break;
	case NODE_Float64:;
	    ctReadBytes = 8;
	    *hNodeContent = malloc(ctReadBytes);
	    if (*hNodeContent == NULL) {
		sbAppend(sbError, "unable to allocate float64 buffer");
		return FAILURE;
	    }
	    xdrmem_create(&xhandle,
			  pBuffer->pcharContent +
			  pBuffer->iContentPosition0, ctReadBytes,
			  XDR_DECODE);
	    if (xdr_double(&xhandle, *hNodeContent) == TRUE) {
		pBuffer->iContentPosition0 += ctReadBytes;
		pBuffer->iContentLength -= ctReadBytes;
	    } else {
		sbAppend(sbError, "float64 data translation error");
		return FAILURE;
	    }
	    break;
	case NODE_String:;
	case NODE_URL:;
	    if (read_String
		((NodeContent_String *) (*hNodeContent), pBuffer,
		 sbError) == FAILURE) {
		sbInsert(sbError, "error reading string/URL: ");
		return FAILURE;
	    }
	    break;
	case NODE_root:;
	    /* there is no data to be loaded for this node type */
	    break;
	case NODE_Structure:;
	    /* there is no data to be loaded for this node type */
	    break;
	case NODE_Sequence:;
	    if (read_Sequence
		(nodeTemplate, (NodeContent_Sequence *) (*hNodeContent),
		 pBuffer, sbError) == FAILURE) {
		sbInsert(sbError, "error reading sequence: ");
		return FAILURE;
	    }
	    break;
	case NODE_Array:;
	    if (read_Array
		((NodeContent_Array *) (*hNodeContent), pBuffer,
		 sbError) == FAILURE) {
		sbInsert(sbError, "error reading array: ");
		return FAILURE;
	    }
	    break;
	case NODE_Grid:;
	    if (read_Grid
		((NodeContent_Grid *) (*hNodeContent), pBuffer,
		 sbError) == FAILURE) {
		sbInsert(sbError, "error reading grid: ");
		return FAILURE;
	    }
	    break;
	default:
	    sbAppend_string(sbError, "%s: ", nodeTemplate->sName);
	    sbAppend_int(sbError, "unknown node type while reading %d",
			 nodeTemplate->eType);
	    return FAILURE;
	}

    /* recursively process any sub nodes in a structure */
    /* if the node is a structure node and the node content is NULL then that means the
     * structure is not in a sequence, otherwise it is a structure in a sequence */
    if (nodeTemplate->eType == NODE_Structure) {
	if (nodeContent_Supplied == NULL) {
	    /* then we are not in a sequence */
	} else {		/* we are inside a sequence, validate the sub node count for the content */
	    NodeContent_Structure *nodeStructureContent =

	// remember nodeContent_Supplied is a pointer to a pointer
		(NodeContent_Structure *) *nodeContent_Supplied;

	    if (nodeStructureContent->ctMembers == iSubNodeCount_Template) {
		/* content node matches template node */
	    } else {
		sbAppend_int(sbError,
			     "validation error, sub-node count of template %i does not match",
			     iSubNodeCount_Template);
		sbAppend_int(sbError,
			     " sub-node count of structure content %i",
			     nodeStructureContent->ctMembers);
		return FAILURE;
	    }
	}
	for (xSubNode = 1; xSubNode <= iSubNodeCount_Template; xSubNode++) {
	    DAPNode *pSubNode = *(nodeTemplate->nodeSub + (xSubNode - 1));
	    if (pSubNode == NULL) {
		sbAppend_int(sbError,
			     "template sub-node %d does not exist",
			     xSubNode);
		return FAILURE;
	    } else {
		void *nodeSubContent;
		if (nodeContent_Supplied == NULL) {
		    nodeSubContent = NULL;
		} else {
		    NodeContent_Structure *nodeStructureContent =
			(NodeContent_Structure *) *nodeContent_Supplied;

		    nodeSubContent =
			nodeStructureContent->array_NodeContent +
			(xSubNode - 1);
		}
		if (read_Node(pSubNode, nodeSubContent, pBuffer, sbError)
		    == FAILURE) {
		    sbInsert_int(sbError, "error reading sub-node %d, ",
				 xSubNode);
		    sbInsert_string(sbError, "%s: ", pSubNode->sName);
		    return FAILURE;
		}
	    }
	}
    }

    return SUCCESS;
}

eRESULT read_String(NodeContent_String * pStringContent, Buffer * pBuffer,
		    StringBuffer * sbError)
{
    if (pStringContent == NULL) {
	sbAppend(sbError,
		 "internal error, string content buffer unexpectedly missing");
	return FAILURE;
    }
    if (read_String_raw(&pStringContent->sContent, pBuffer, sbError) ==
	SUCCESS) {
	int iStringLength = strlen(pStringContent->sContent);
	pStringContent->iLength = iStringLength;
	return SUCCESS;
    } else {
	return FAILURE;
    }
}

int zIsPrimitiveType(eNODE_TYPE node_type)
{
    switch (node_type) {
    case NODE_Byte:
    case NODE_Int16:
    case NODE_UInt16:
    case NODE_Int32:
    case NODE_UInt32:
    case NODE_Float32:
    case NODE_Float64:
	return TRUE;
    default:
	return FALSE;
    }
}

/* the format of an XDR string is a 4-byte Motorola int followed by a series of 7-bit chars
 * here we support 8-bit chars additionally since they are often found in real implementations */
eRESULT read_String_raw(char **hString, Buffer * pBuffer,
			StringBuffer * sbError)
{
#if 0
    unsigned int posString;
#endif
    unsigned int iStringSize;
    char *pStringSize4;
    char *pStringContentPointer;
    char *sNewString;
    /* XDR xhandle; */

/* debugging
{
int xByte;
printf("reading node, bytes:\n");
for( xByte = 0; xByte < 100; xByte++ ){
	printf("  %02i: %c", xByte, *((unsigned char*)pBuffer->pcharContent + pBuffer->iContentPosition0 + xByte) );
	if( xByte % 10 == 0 ) printf("\n");
}
printf("\n");
}
*/

    /* determine size of string */
    pStringSize4 = pBuffer->pcharContent + pBuffer->iContentPosition0;
    iStringSize =
	*pStringSize4 * 0x1000000 + *(pStringSize4 + 1) * 0x10000 +
	*(pStringSize4 + 2) * 0x100 + *(pStringSize4 + 3);
    pBuffer->iContentPosition0 += 4;	/* advance buffer position past string size preamble */
    pBuffer->iContentLength -= 4;
    if (iStringSize > MAX_STRING_LENGTH) {
	sbAppend_int(sbError,
		     "string size, %i, exceeds maximum (possible data corruption)",
		     iStringSize);
	return FAILURE;
    }

    pStringContentPointer =
	pBuffer->pcharContent + pBuffer->iContentPosition0;
    sNewString = malloc(iStringSize + 1);	/* add an extra space for a null terminator */
    if (sNewString == NULL) {
	sbAppend(sbError, "unable to allocate memory for string");
	return FAILURE;
    }

    /* XDR does not support UNICODE, it uses simple C-strings */
    memcpy(sNewString, pStringContentPointer, iStringSize);

    *(sNewString + iStringSize) = '\0';	/* null terminate string */
    *hString = sNewString;
    pBuffer->iContentPosition0 += iStringSize + macroQuadPad(iStringSize);
    pBuffer->iContentLength -= iStringSize + macroQuadPad(iStringSize);
    return SUCCESS;

/* old non-working methodology
	xdrmem_create( &xhandle, pBuffer->pcharContent + pBuffer->iContentPosition0, MAX_STRING_LENGTH, XDR_DECODE );
	if( xdr_string( &xhandle, hString, MAX_STRING_LENGTH ) == TRUE ){
		unsigned int lenString = strlen( *hString );
		pBuffer->iContentPosition0 += lenString + macroQuadPad(lenString);
		pBuffer->iContentLength    -= lenString + macroQuadPad(lenString);
		return SUCCESS;
	} else {
		sbAppend( sbError, "string data translation error" );
		return FAILURE;
	}
*/
}

eRESULT read_Grid(NodeContent_Grid * pGridContent, Buffer * pBuffer,
		  StringBuffer * sbError)
{
    int ctMapping;
    int xMapping;
    NodeContent_Array *grid_values;
    NodeContent_Array *grid_map_current;
    if (pGridContent == NULL) {	/* the content should have been created during DDS parsing */
	sbAppend(sbError, "node content does not exist");
	return FAILURE;
    }
    ctMapping = pGridContent->ctMappings;
    if (ctMapping <= 0) {	/* the mapping count should have been established during DDS parsing */
	sbAppend_int(sbError, "mapping count (%d) is invalid", ctMapping);
	return FAILURE;
    }
    grid_values = pGridContent->values;
    if (read_Array(grid_values, pBuffer, sbError) == FAILURE) {
	sbInsert(sbError, "error reading value array: ");
	return FAILURE;
    }
    for (xMapping = 1; xMapping <= ctMapping; xMapping++) {
	grid_map_current = pGridContent->maps1[xMapping];
	if (read_Array(grid_map_current, pBuffer, sbError) == FAILURE) {
	    sbInsert_int(sbError, "error reading mapping array %d: ",
			 xMapping);
	    return FAILURE;
	}
    }
    return SUCCESS;
}

eRESULT read_Array(NodeContent_Array * pArray, Buffer * pBuffer,
		   StringBuffer * sbError)
{
    XDR xhandle;
    unsigned int ctArrayBytes_in;	/* total number of bytes read by XDR; the length of the xdr prepended length plus the array itself */
    unsigned int iSizeofLengthPrefix = sizeof(u_int);
    int eArrayType;
    /* Changed type from u_int to int; u_int returned a warning. jhrg */
    unsigned int lenArray_OPeNDAP = 0;	/* OPeNDAP prepends a length */
    /* lenArray_XDR is unsigned since that's what xdr_array expects. I cast
     * when it's passed to xdr_int (which expects an int*. jhrg */
    unsigned int lenArray_XDR = 0;	/* XDR prepends a length */
    if (pArray == NULL) {
	sbAppend(sbError, "internal error, array content pointer missing");
	return FAILURE;
    }
    eArrayType = pArray->eType;

    /* read the length prepended by OPeNDAP */
    xdrmem_create(&xhandle,
		  pBuffer->pcharContent + pBuffer->iContentPosition0,
		  iSizeofLengthPrefix, XDR_DECODE);
    if (xdr_int(&xhandle, (int *) &lenArray_OPeNDAP) == TRUE) {
	pBuffer->iContentPosition0 += iSizeofLengthPrefix;
	pBuffer->iContentLength -= iSizeofLengthPrefix;
    } else {
	sbAppend(sbError, "failed to read OPeNDAP length");
	return FAILURE;
    }

    if (eArrayType == ARRAY_String || eArrayType == ARRAY_URL) {
	/* then there is no xdr length header */
    } else {

	/* read the length as prepended by XDR */
	xdrmem_create(&xhandle,
		      pBuffer->pcharContent + pBuffer->iContentPosition0,
		      sizeof(DINT32), XDR_DECODE);
	if (xdr_int(&xhandle, (int *) &lenArray_XDR) == TRUE) {
	    /* leave the length in the buffer because xdr_array needs to read it too */
	} else {
	    sbAppend(sbError, "failed to read XDR length");
	    return FAILURE;
	}

	/* validate that the lengths are the same */
	if (lenArray_OPeNDAP != lenArray_XDR) {
	    sbAppend_int(sbError,
			 "validation failure, OPeNDAP array length (%u) ",
			 lenArray_OPeNDAP);
	    sbAppend_int(sbError, "does not match XDR array length (%u)",
			 lenArray_XDR);
	    return FAILURE;
	}
    }

    if (lenArray_OPeNDAP == 0)
	return SUCCESS;		/* TODO - make warning ? */

    /* todo consider validating array length against dimensions (probably a good idea) */

    /* todo -- this reads array in one swoop, it should be progressive so that entire array does not have
     * to be loaded in buffer todo */
    if (eArrayType == ARRAY_Byte) {	/* no translation needed: bytes will not change since byte order does not effect their machine representation */
	char *pServerData;
	char *pClientArray;
	ctArrayBytes_in = iSizeofLengthPrefix + lenArray_XDR;
	pServerData =
	    pBuffer->pcharContent + pBuffer->iContentPosition0 +
	    iSizeofLengthPrefix;
	if (lenArray_XDR > (unsigned int) DODS_MAX_ARRAY) {
	    sbAppend_int(sbError,
			 "unable to read byte array of length %u, exceeds maximum array size",
			 lenArray_XDR);
	    return FAILURE;
	}
	pClientArray = malloc(lenArray_XDR);
	if (pClientArray == NULL) {
	    sbAppend_int(sbError, "failed to allocate memory for %u bytes",
			 lenArray_XDR);
	    return FAILURE;
	}
	/* old method: for( xByte = 0; xByte < lenArray_XDR; xByte++ ) *(pClientArray + xByte) = *(pServerData + xByte); */
	memcpy(pClientArray, pServerData, lenArray_XDR);
	pArray->array = pClientArray;
	pBuffer->iContentPosition0 += ctArrayBytes_in + macroQuadPad(ctArrayBytes_in);	/* modulo 4 accounts for XDR's 4-byte framing */
	pBuffer->iContentLength -=
	    ctArrayBytes_in + macroQuadPad(ctArrayBytes_in);
    } else if (eArrayType == ARRAY_String || eArrayType == ARRAY_URL) {
	unsigned int xString;
	NodeContent_String **pClientArray;
	pClientArray =
	    (NodeContent_String **) malloc(lenArray_OPeNDAP *
					   sizeof(NodeContent_String *));
	if (pClientArray == NULL) {
	    sbAppend_int(sbError,
			 "failed to allocate memory for %i string pointers",
			 lenArray_OPeNDAP);
	    return FAILURE;
	}
	for (xString = 1; xString <= lenArray_OPeNDAP; xString++) {
	    NodeContent_String *pContent_string =
		(NodeContent_String *) malloc(sizeof(NodeContent_String));
	    if (pContent_string == NULL) {
		sbAppend_int(sbError,
			     "failed to allocate memory for string/URL content for string %i",
			     xString);
		return FAILURE;
	    }
	    pContent_string->iLength = 0;
	    pContent_string->sContent = NULL;
	    if (!read_String(pContent_string, pBuffer, sbError)) {	/* acquire string content from buffer */
		sbInsert_int(sbError, "failed to read string/URL %i: ",
			     xString);
		return FAILURE;
	    }
	    *(pClientArray + xString - 1) = pContent_string;	/* put pointer to string content in array */
	}
	pArray->array = (void *) pClientArray;	/* assign the newly built array to the content pointer for the array */
    } else {			/* todo xdr_array is kind of bogus, we should do this ourself since it just runs through flat */
	u_int elsize;		/* size in bytes of each output element */
	u_int iInWidth;		/* size in bytes of each input element */
	xdrproc_t elproc;	/* xdr routine to handle each element */
	switch (eArrayType) {
	case ARRAY_Byte:	/* not used -- handled above */
	    iInWidth = 1;
	    elsize = sizeof(DBYTE);
	    elproc = (xdrproc_t) xdr_char;
	    break;
	case ARRAY_Int16:
	    iInWidth = 4;
	    elsize = sizeof(DINT16);
	    elproc = (xdrproc_t) xdr_short;
	    break;
	case ARRAY_UInt16:
	    iInWidth = 4;
	    elsize = sizeof(DUINT16);
	    elproc = (xdrproc_t) xdr_u_short;
	    break;
	case ARRAY_Int32:
	    iInWidth = 4;
	    elsize = sizeof(DINT32);
	    elproc = (xdrproc_t) xdr_int;
	    break;
	case ARRAY_UInt32:
	    iInWidth = 4;
	    elsize = sizeof(DUINT32);
	    elproc = (xdrproc_t) xdr_u_int;
	    break;
	case ARRAY_Float32:
	    iInWidth = 4;
	    elsize = sizeof(DFLOAT32);
	    elproc = (xdrproc_t) xdr_float;
	    break;
	case ARRAY_Float64:
	    iInWidth = 8;
	    elsize = sizeof(DFLOAT64);
	    elproc = (xdrproc_t) xdr_double;
	    break;
	default:
	    sbAppend(sbError, "unrecognized array type");
	    return FAILURE;
	    /* Originally this was marked 'todo.' I hacked
	     * this error message in place because an
	     * unknown type would send the code down into
	     * the next block with a bunch of uninitialized
	     * values. jhrg 8/14/06 */
	}
	{
	    ctArrayBytes_in =
		iSizeofLengthPrefix + lenArray_XDR * iInWidth;
	    if (ctArrayBytes_in > pBuffer->iContentLength) {
		sbAppend_int(sbError, "read is for %i bytes, ",
			     ctArrayBytes_in);
		sbAppend_int(sbError,
			     "but there are only %i in the buffer",
			     pBuffer->iContentLength);
		return FAILURE;
	    }
	    xdrmem_create(&xhandle,
			  pBuffer->pcharContent +
			  pBuffer->iContentPosition0, ctArrayBytes_in,
			  XDR_DECODE);
	    if (xdr_array
		(&xhandle, &pArray->array, &lenArray_XDR, DODS_MAX_ARRAY,
		 elsize, elproc) == TRUE) {
		pBuffer->iContentPosition0 +=
		    ctArrayBytes_in + macroQuadPad(ctArrayBytes_in);
		pBuffer->iContentLength -=
		    ctArrayBytes_in + macroQuadPad(ctArrayBytes_in);
	    } else {
		sbAppend_int(sbError, "failed to read array of length %d",
			     lenArray_XDR);
		return FAILURE;
	    }
	}
    }
    return SUCCESS;
}

/* Sequences are read and stored differently than the other structures to optimize the
 * internal data structure. If sequences were read and stored according to their
 * natural definition then a function call would take place on every field and
 * every row. Instead we would like to handle all the primitive types in a
 * sequence as a single matrix. Thus there needs to be only one function call
 * per row of the sequence for reading and only function call per level of the
 * sequence when using the data. This makes the processing faster proportionally to the
 * number of elements in the sequence. Thus if the sequence has a million
 * elements it gets processed a million times faster this way than if a
 * function call were done on every record/field.
 *
 * Another issue is that the standard content structures for arrays etc cannot
 * store the array content for a sequence because the sequence node actually
 * contains a vector of arrays, not a single array. Thus it must be read and
 * stored different. Rather than create special content types for arrays/grids/strings
 * the solution is to handle them differently.
 *
 * To manage this the entire content of the sequence is stored only in the
 * node content of the top-most sequence. The node content of all elements below
 * the top-most sequence will be empty. For this reason when a sequence is printed
 * all the logic in print_Sequence_toFile() is only applied to the content of the top-most
 * sequence and if a sequence is encountered its sub-nodes are not recursed further
 * by the node handling logic.
 *
 * As a consequence of this methodology this function takes two key arguments:
 * nodeSequence = the sequence being processed
 * pNodeContent_Sequence = the content structure the data will be added to
 * If the sequence is a top-level sequence then the content node will be the
 * content node of nodeSequence, but if the sequence is an inner (nested)
 * sequence then the content node will NOT be content node of the sequence.
 * Instead it will be a pointer to the field vector of the parent sequence and
 * there will be many such pointers--one for each row in the vector.
 *
 * If the sequence field is a structure then the content node will be a pointer
 * to a vector of the nodes in the structure
 *
 */
eRESULT read_Sequence(DAPNode * nodeSequence,
		      NodeContent_Sequence * pNodeContent_Sequence,
		      Buffer * pBuffer, StringBuffer * sbError)
{
    DAPNode *nodeTemplateField;
    unsigned int xField, xRow, iRowCapacity, ctFields;

    DBG(printf
	("reading sequence %p with content at %p\n", (void *) nodeSequence,
	 (void *) pNodeContent_Sequence);
	)
	if (pNodeContent_Sequence == NULL) {
	sbAppend(sbError, "node content is missing");
	return FAILURE;
    }
    if (pBuffer->iContentLength < 3) {	/* TODO - this will need to upgraded when dynamic load is installed */
	sbAppend(sbError, "sequence is incomplete");
	return FAILURE;
    }

    DBG(printf
	("content for sequence named %s at %p has %i fields\n",
	 nodeSequence->sName, (void *) pNodeContent_Sequence,
	 (void *) nodeSequence->iSubNodeCount);
	)

	/* determine the number of fields in the sequence */
	ctFields = nodeSequence->iSubNodeCount;
    pNodeContent_Sequence->ctFields = ctFields;
    if (ctFields == 0)
	return SUCCESS;		/* the field vectors will remain null in this case */

    /* free any pre-existing field types from the content */
    if (pNodeContent_Sequence->aeFieldType != NULL) {
	free(pNodeContent_Sequence->aeFieldType);
    }

    /* allocate memory for the field type vector and assign the type values from the template */
    pNodeContent_Sequence->aeFieldType =
	malloc(sizeof(eNODE_TYPE) * ctFields);
    if (pNodeContent_Sequence->aeFieldType == NULL) {
	sbAppend(sbError,
		 "failed to allocate memory for sequence's field type vector");
	return FAILURE;
    }
    for (xField = 1; xField <= ctFields; xField++) {	/* initialize types from template */
	nodeTemplateField = *(nodeSequence->nodeSub + (xField - 1));
	*(pNodeContent_Sequence->aeFieldType + xField - 1) = nodeTemplateField->eType;
    }

    /* allocate memory for the field value vectors */
    pNodeContent_Sequence->aFieldValue = malloc(sizeof(void *) * ctFields);
    if (pNodeContent_Sequence->aFieldValue == NULL) {
	sbAppend(sbError,
		 "failed to allocate memory for sequence's field value vector");
	return FAILURE;
    }

    memset(pNodeContent_Sequence->aFieldValue, 0, ctFields * sizeof(pNodeContent_Sequence->aFieldValue));

    xRow = 0;
    iRowCapacity = 0;
    while (TRUE) {
	unsigned char cMarker =
	    *(unsigned char *) (pBuffer->pcharContent +
				pBuffer->iContentPosition0);
	pBuffer->iContentPosition0 += 4;	/* only the first byte is the marker -- ignore rest of the word */
	pBuffer->iContentLength -= 4;
	DBG(printf
	    ("content position is now %i\n", pBuffer->iContentPosition0);
	    )
	    if (cMarker == SEQUENCE_MARKER_StartOfInstance) {
	    xRow++;
	    DBG(printf("BOS marker found, reading row %i\n", xRow);
		)
		if (xRow > iRowCapacity) {	/* enlarge field vectors */
		unsigned int ctRows_desired;
		if (iRowCapacity <= 0) {
		    ctRows_desired = 1000;
		} else {
		    ctRows_desired = iRowCapacity * 2;
		}
		for (xField = 1; xField <= ctFields; xField++) {
		    int iWordSize;
		    char *pNewVector;	/* the copying is done by char (=byte) */
		    char *pOldVector;
		    eNODE_TYPE *pFieldTypeVector;
		    eNODE_TYPE eFieldType;
		    pFieldTypeVector = pNodeContent_Sequence->aeFieldType;
		    if (pFieldTypeVector == NULL) {
			sbAppend_int(sbError,
				     "field type vector does not exist even though there are %d fields",
				     ctFields);
			return FAILURE;
		    }
		    eFieldType = *(pFieldTypeVector + xField - 1);
		    DBG(printf("In read_Node: eFieldType: %x\n", eFieldType);)
		    switch (eFieldType) {
		    case NODE_Byte:
			iWordSize = sizeof(DBYTE);
			break;
		    case NODE_Int16:
			iWordSize = sizeof(DINT16);
			break;
		    case NODE_UInt16:
			iWordSize = sizeof(DUINT16);
			break;
		    case NODE_Int32:
			iWordSize = sizeof(DINT32);
			break;
		    case NODE_UInt32:
			iWordSize = sizeof(DUINT32);
			break;
		    case NODE_Float32:
			iWordSize = sizeof(DFLOAT32);
			break;
		    case NODE_Float64:
			iWordSize = sizeof(DFLOAT64);
			break;
		    case NODE_String:
		    case NODE_URL:
		    case NODE_Array:
		    case NODE_Grid:
		    case NODE_Sequence:
		    case NODE_Structure:;
			iWordSize = sizeof(void *);
			break;
		    case NODE_root:;
			sbAppend(sbError,
				 "invalid content, root node inside sequence");
			return FAILURE;
		    default:
			sbAppend(sbError,
				 "invalid content, undefined field type");
			return FAILURE;
		    }
		    {		/* allocate and zero the new space because could be used for string which xdr requires to be zeroed */
			int iAllocationSize = iWordSize * ctRows_desired;
			pNewVector = malloc(iAllocationSize);
			DBG(printf
			    ("created value array at %p for field %i type is %x\n",
			     pNewVector, xField, eFieldType);
			    )
			    if (pNewVector == NULL) {
			    sbAppend_int(sbError,
					 "out of memory attempting to allocate an additional %i bytes",
					 iAllocationSize);
			    sbAppend_int(sbError, " for row %i", xRow);
			    sbAppend_int(sbError, " for field %i", xField);
			    return FAILURE;
			}
			memset(pNewVector, 0, iAllocationSize);
		    }
		    {

			pOldVector =
			    *(pNodeContent_Sequence->aFieldValue +
			      (xField - 1));
		        if(pOldVector != NULL) {
			    memcpy((void*)pNewVector,(void*)pOldVector,pNodeContent_Sequence->ctRows * iWordSize);
			    free(pOldVector);
			}

			*(pNodeContent_Sequence->aFieldValue +
			  (xField - 1)) = pNewVector;
		    }
		    /* TODO consider shrinking array to fit after all rows have been read in */
		}
		iRowCapacity = ctRows_desired;
	    }
	    if (read_Record(nodeSequence, pNodeContent_Sequence, xRow - 1, pBuffer,
		 sbError) == SUCCESS) {
		pNodeContent_Sequence->ctRows = xRow;
	    } else {
		sbInsert_int(sbError, "error reading row %d: ", xRow);
		return FAILURE;
	    }
	} else if (cMarker == SEQUENCE_MARKER_EndOfSequence) {
	    DBG(printf("end of sequence encountered\n");
		)
		break;		/* break out of loop -- done */
	} else {
	    sbAppend_int(sbError,
			 "marker not found for row %i, sequence may be corrupt",
			 xRow);
	    return FAILURE;
	}
    }
    return SUCCESS;
}

eRESULT read_Record(DAPNode * nodeSequence,
		    NodeContent_Sequence * pNodeContent_Sequence,
		    int xRow0, Buffer * pBuffer, StringBuffer * sbError)
{
    XDR xhandle;
    unsigned int xField;
    unsigned int ctReadBytes;	/* total number of bytes read by XDR */
    for (xField = 1; xField <= pNodeContent_Sequence->ctFields; xField++) {	/* todo ripping through a blind double pointer array like this is risky -- it is key to add internal validation codes to the OCAPI */
	eNODE_TYPE eFieldType, eFieldType_Template;
	void *pVector;
	DAPNode *pTemplateFieldNode;

	/* determine template parameters */
	pTemplateFieldNode = *(nodeSequence->nodeSub + xField - 1);
	eFieldType_Template = pTemplateFieldNode->eType;

	/* determine content parameters */
	eFieldType = *(pNodeContent_Sequence->aeFieldType + (xField - 1));
	pVector = *(pNodeContent_Sequence->aFieldValue + (xField - 1));

	DBG(printf
	    ("reading sequence %p with content at %p, row %u, field %u, field vector is: %p\n",
	     (void *) nodeSequence, (void *) pNodeContent_Sequence,
	     xRow0 + 1, xField, (void *) pVector);
	    )

	    /* validate that the template node matches the content node being passed in */
	    if (eFieldType_Template != eFieldType) {
	    sbAppend_int(sbError,
			 "template %i node type does not match content type",
			 eFieldType_Template);
	    return FAILURE;
	}

	switch (eFieldType) {
	case NODE_Byte:;	/* bytes are stored directly in the vector with no xdr translation */
	DBG(printf("In read_Record, NODE_Byte\n");)
#if 1
	{
			ctReadBytes = 4;	/* bytes are sent as 4 byte packages (yup) */
			xdrmem_create(&xhandle, pBuffer->pcharContent + pBuffer->iContentPosition0,
					ctReadBytes, XDR_DECODE);
			if (xdr_char(&xhandle, (DBYTE *)pVector + xRow0) == TRUE) {
			    pBuffer->iContentPosition0 += ctReadBytes;
			    pBuffer->iContentLength -= ctReadBytes;
			} else {
			    sbAppend_int(sbError,
					 "failed to read DBYTE in field %d",
					 xField);
			    return FAILURE;
			}
	    }
	    break;
#else
    {
		*((char *) pVector + xRow0) = *(char *) (pBuffer->pcharContent + pBuffer->iContentPosition0);
		pBuffer->iContentPosition0 += 4;
		pBuffer->iContentLength -= 4;
	    }
	    break;
#endif
	case NODE_Int16:;
	    {
		ctReadBytes = 4;	/* shorts are sent as 4 byte packages */
		xdrmem_create(&xhandle,
			      pBuffer->pcharContent +
			      pBuffer->iContentPosition0, ctReadBytes,
			      XDR_DECODE);
		if (xdr_short(&xhandle, (DINT16 *) pVector + xRow0) ==
		    TRUE) {
		    pBuffer->iContentPosition0 += ctReadBytes;
		    pBuffer->iContentLength -= ctReadBytes;
		} else {
		    sbAppend_int(sbError,
				 "failed to read DINT16 in field %d",
				 xField);
		    return FAILURE;
		}
	    }
	    break;
	case NODE_UInt16:;
	    {
		ctReadBytes = 4;
		xdrmem_create(&xhandle,
			      pBuffer->pcharContent +
			      pBuffer->iContentPosition0, ctReadBytes,
			      XDR_DECODE);
		if (xdr_u_short(&xhandle, (DUINT16 *) pVector + xRow0) ==
		    TRUE) {
		    pBuffer->iContentPosition0 += ctReadBytes;
		    pBuffer->iContentLength -= ctReadBytes;
		} else {
		    sbAppend_int(sbError,
				 "failed to read DUINT16 in field %d",
				 xField);
		    return FAILURE;
		}
	    }
	    break;
	case NODE_Int32:;
	    {
		ctReadBytes = 4;
		xdrmem_create(&xhandle,
			      pBuffer->pcharContent +
			      pBuffer->iContentPosition0, ctReadBytes,
			      XDR_DECODE);
		if (xdr_int(&xhandle, (DINT32 *) pVector + xRow0) == TRUE) {
		    pBuffer->iContentPosition0 += ctReadBytes;
		    pBuffer->iContentLength -= ctReadBytes;
		} else {
		    sbAppend_int(sbError,
				 "failed to read DINT32 in field %d",
				 xField);
		    return FAILURE;
		}
	    }
	    break;
	case NODE_UInt32:;
	    {
		ctReadBytes = 4;
		xdrmem_create(&xhandle,
			      pBuffer->pcharContent +
			      pBuffer->iContentPosition0, ctReadBytes,
			      XDR_DECODE);
		if (xdr_u_int(&xhandle, (DUINT32 *) pVector + xRow0) ==
		    TRUE) {
		    pBuffer->iContentPosition0 += ctReadBytes;
		    pBuffer->iContentLength -= ctReadBytes;
		} else {
		    sbAppend_int(sbError,
				 "failed to read DUINT32 in field %d",
				 xField);
		    return FAILURE;
		}
	    }
	    break;
	case NODE_Float32:;
	    {
		ctReadBytes = 4;
		xdrmem_create(&xhandle,
			      pBuffer->pcharContent +
			      pBuffer->iContentPosition0, ctReadBytes,
			      XDR_DECODE);
		if (xdr_float(&xhandle, (DFLOAT32 *) pVector + xRow0) ==
		    TRUE) {
		    pBuffer->iContentPosition0 += ctReadBytes;
		    pBuffer->iContentLength -= ctReadBytes;
		} else {
		    sbAppend_int(sbError,
				 "failed to read DFLOAT32 in field %d",
				 xField);
		    return FAILURE;
		}
	    }
	    break;
	case NODE_Float64:;
	    {
		double *pfloat64 = (double *) pVector;
		ctReadBytes = 8;
		xdrmem_create(&xhandle,
			      pBuffer->pcharContent +
			      pBuffer->iContentPosition0, ctReadBytes,
			      XDR_DECODE);
		if (xdr_double(&xhandle, pfloat64 + xRow0) == TRUE) {
		    pBuffer->iContentPosition0 += ctReadBytes;
		    pBuffer->iContentLength -= ctReadBytes;
		} else {
		    sbAppend_int(sbError,
				 "failed to read DFLOAT64 in field %d",
				 xField);
		    return FAILURE;
		}
		DBG(printf
		    ("stored float64 with value %f at %p\n",
		     *(pfloat64 + xRow0), pfloat64 + xRow0);
		    )
	    }
	    break;
	case NODE_String:;
	case NODE_URL:;
	    {
		/* allocate memory for string content and read the string into it */
		NodeContent_String *pContent_string =
		    (NodeContent_String *)
		    malloc(sizeof(NodeContent_String));
		if (pContent_string == NULL) {
		    sbAppend_int(sbError,
				 "failed to allocate memory for string in field %d",
				 xField);
		    return FAILURE;
		}
		pContent_string->iLength = 0;
		pContent_string->sContent = NULL;
		if (read_String(pContent_string, pBuffer, sbError) ==
		    FAILURE) {
		    sbInsert_int(sbError,
				 "error reading string in field %d: ",
				 xField);
		    return FAILURE;
		}

		/* assign content to vector */
		*((NodeContent_String **) pVector + xRow0) =
		    pContent_string;

	    }
	    break;
	case NODE_Array:;
	    {
		/* duplicate the template content for the array */
		NodeContent_Array *pTemplateArrayContent =
		    pTemplateFieldNode->nodeContent;
		unsigned int xDimension = 0;
		NodeContent_Array *pContent_array =
		    malloc(sizeof(NodeContent_Array));
		if (pContent_array == NULL) {
		    sbAppend_int(sbError,
				 "failed to allocate memory for array in field %d",
				 xField);
		    return FAILURE;
		}
		pContent_array->sArrayName =
		    sCloneString(pTemplateArrayContent->sArrayName);
		pContent_array->eType = pTemplateArrayContent->eType;
		pContent_array->ctDimensions =
		    pTemplateArrayContent->ctDimensions;
		pContent_array->iDimensionLengthArray1 =
		    malloc(sizeof(unsigned int) *
			   (pContent_array->ctDimensions + 1));
		if (pContent_array->iDimensionLengthArray1 == NULL) {
		    sbAppend_int(sbError,
				 "failed to allocate memory for array dim size in field %d",
				 xField);
		    return FAILURE;
		}
		pContent_array->sDimensionName1 =
		    malloc(sizeof(char *) *
			   (pContent_array->ctDimensions + 1));
		if (pContent_array->sDimensionName1 == NULL) {
		    sbAppend_int(sbError,
				 "failed to allocate memory for array dim names in field %d",
				 xField);
		    return FAILURE;
		}
		for (; xDimension <= pContent_array->ctDimensions; xDimension++) {
		    pContent_array->iDimensionLengthArray1[xDimension] =
			pTemplateArrayContent->	iDimensionLengthArray1[xDimension];
		    pContent_array->sDimensionName1[xDimension] =sCloneString(pTemplateArrayContent->
				     sDimensionName1[xDimension]);
		}
		pContent_array->array = NULL;

		/* read the data from the buffer into the newly created array and store it in the vector */
		if (read_Array(pContent_array, pBuffer, sbError) ==
		    FAILURE) {
		    sbInsert_int(sbError,
				 "error reading array in field %d: ",
				 xField);
		    return FAILURE;
		}
		*((NodeContent_Array **) pVector + xRow0) = pContent_array;
	    }
	    break;
	case NODE_Grid:;
	    {
		/* duplicate the template content for the grid */
		NodeContent_Array *pValueArrayContent;
		NodeContent_Grid *pContent_grid;
		NodeContent_Grid *pTemplateGridContent =
		    pTemplateFieldNode->nodeContent;
		unsigned int ctValueDimensions;
		unsigned int xMapping;
		pContent_grid = malloc(sizeof(NodeContent_Grid));
		if (pContent_grid == NULL) {
		    sbAppend_int(sbError,
				 "failed to allocate memory for grid in field %d",
				 xField);
		    return FAILURE;
		}
		pContent_grid->ctMappings =
		    pTemplateGridContent->ctMappings;
		pContent_grid->values = NULL;

		/* allocate memory for the mappings vector */
		pContent_grid->maps1 =
		    malloc(sizeof(NodeContent_Array **) *
			   (pContent_grid->ctMappings + 1));
		if (pContent_grid->maps1 == NULL) {
		    sbAppend_int(sbError,
				 "failed to allocate memory for grid mappings vector in field %d",
				 xField);
		    return FAILURE;
		}
		pContent_grid->maps1[0] = NULL;	/* one-based array, first entry is null */

		for (xMapping = 1; xMapping <= pContent_grid->ctMappings;
		     xMapping++) {
		    NodeContent_Array *pMapArrayContent;
		    NodeContent_Array *pTemplateMapArray;
		    unsigned int xMapDimension;
		    unsigned int ctDimensions;

		    /* determine the template content */
		    pTemplateMapArray =
			pTemplateGridContent->maps1[xMapping];
		    if (pTemplateMapArray == NULL) {
			sbAppend_int(sbError,
				     "no content found for template's grid mapping array %i",
				     xMapping);
			return FAILURE;
		    }

		    /* create a new mapping array */
		    pMapArrayContent = malloc(sizeof(NodeContent_Array));
		    if (pMapArrayContent == NULL) {
			sbAppend_int(sbError,
				     "failed to allocate memory for mapping array %i",
				     xMapping);
			sbAppend_int(sbError, " in field %i", xField);
			return FAILURE;
		    }

		    /* size the dimension info */
		    ctDimensions = pTemplateMapArray->ctDimensions;
		    pMapArrayContent->iDimensionLengthArray1 =
			malloc(sizeof(unsigned int) * (ctDimensions + 1));
		    if (pMapArrayContent->iDimensionLengthArray1 == NULL) {
			sbAppend_int(sbError,
				     "failed to allocate memory for mapping array %i dimension sizes",
				     xMapping);
			sbAppend_int(sbError, " in field %i", xField);
			return FAILURE;
		    }
		    pMapArrayContent->sDimensionName1 =
			malloc(sizeof(char *) * (ctDimensions + 1));
		    if (pMapArrayContent->sDimensionName1 == NULL) {
			sbAppend_int(sbError,
				     "failed to allocate memory for mapping array %i dimension names",
				     xMapping);
			sbAppend_int(sbError, " in field %i", xField);
			return FAILURE;
		    }

		    /* clone the template info to the new mapping array */
		    pMapArrayContent->eType = pTemplateMapArray->eType;
		    pMapArrayContent->ctDimensions = ctDimensions;
		    pMapArrayContent->sArrayName =
			sCloneString(pTemplateMapArray->sArrayName);
		    pMapArrayContent->array = NULL;	/* it is crucial to initialize to NULL, otherwise the XDR routines will not know to allocate memory for it later on */
		    for (xMapDimension = 1;
			 xMapDimension <= pMapArrayContent->ctDimensions;
			 xMapDimension++) {
			pMapArrayContent->
			    iDimensionLengthArray1[xMapDimension] =
			    pTemplateMapArray->
			    iDimensionLengthArray1[xMapDimension];
			if (pTemplateMapArray->
			    sDimensionName1[xMapDimension] == NULL) {
			    pMapArrayContent->
				sDimensionName1[xMapDimension] = NULL;
			} else {
			    pMapArrayContent->
				sDimensionName1[xMapDimension] =
				sCloneString(pTemplateMapArray->
					     sDimensionName1
					     [xMapDimension]);
			}
		    }

		    pContent_grid->maps1[xMapping] = pMapArrayContent;
		}

		/* create a new values array */
		pValueArrayContent = malloc(sizeof(NodeContent_Array));
		if (pValueArrayContent == NULL) {
		    sbAppend_int(sbError,
				 "failed to allocate memory for value array in field %i",
				 xField);
		    return FAILURE;
		}

		/* size the dimension info vectors in the new values array */
		ctValueDimensions =
		    pTemplateGridContent->values->ctDimensions;
		pValueArrayContent->iDimensionLengthArray1 =
		    malloc(sizeof(unsigned int) * (ctValueDimensions + 1));
		if (pValueArrayContent->iDimensionLengthArray1 == NULL) {
		    sbAppend_int(sbError,
				 "failed to allocate memory for value array dimension sizes in field %i",
				 xField);
		    return FAILURE;
		}
		pValueArrayContent->sDimensionName1 =
		    malloc(sizeof(char *) * (ctValueDimensions + 1));
		if (pValueArrayContent->sDimensionName1 == NULL) {
		    sbAppend_int(sbError,
				 "failed to allocate memory for value array dimension names in field %i",
				 xField);
		    return FAILURE;
		}

		/* clone the template info into the new value array */
		{
		    unsigned int xValueDimension;
		    NodeContent_Array *pTemplateValueArray =
			pTemplateGridContent->values;
		    pValueArrayContent->eType = pTemplateValueArray->eType;
		    pValueArrayContent->ctDimensions = ctValueDimensions;
		    pValueArrayContent->sArrayName =
			sCloneString(pTemplateValueArray->sArrayName);
		    pValueArrayContent->array = NULL;	/* it is crucial to initialize to NULL, otherwise the XDR routines will not know to allocate memory for it later on */
		    for (xValueDimension = 1;
			 xValueDimension <= ctValueDimensions;
			 xValueDimension++) {
			pValueArrayContent->
			    iDimensionLengthArray1[xValueDimension] =
			    pTemplateValueArray->
			    iDimensionLengthArray1[xValueDimension];
			if (pTemplateValueArray->
			    sDimensionName1[xValueDimension] == NULL) {
			    pValueArrayContent->
				sDimensionName1[xValueDimension] = NULL;
			} else {
			    pValueArrayContent->
				sDimensionName1[xValueDimension] =
				sCloneString(pTemplateValueArray->
					     sDimensionName1
					     [xValueDimension]);
			}
		    }
		}

		pContent_grid->values = pValueArrayContent;

		/* read the values from the buffer into the newly created content structure */
		if (read_Grid(pContent_grid, pBuffer, sbError) == FAILURE) {
		    sbInsert_int(sbError,
				 "error reading grid in field %d: ",
				 xField);
		    return FAILURE;
		}

		*((NodeContent_Grid **) pVector + xRow0) = pContent_grid;
	    }
	    break;
	case NODE_Sequence:;
	    {
		DAPNode *pNodeSubSequence =
		    *(nodeSequence->nodeSub + (xField - 1));
		NodeContent_Sequence *pContent_sequence;

		/* verify that template node is a sequence */
		if (pNodeSubSequence->eType != NODE_Sequence) {
		    sbAppend_int(sbError, "validation error, field %i of ",
				 xField);
		    sbAppend_string(sbError,
				    "template node %s is not a sequence as expected",
				    nodeSequence->sName);
		    return FAILURE;
		}

		/* create a blank sequence content structure */
		pContent_sequence = malloc(sizeof(NodeContent_Sequence));
		if (pContent_sequence == NULL) {
		    sbAppend_int(sbError,
				 "failed to allocate memory for sequence in field %d",
				 xField);
		    return FAILURE;
		}
		pContent_sequence->ctRows = 0;
		pContent_sequence->ctFields = 0;
		pContent_sequence->aeFieldType = NULL;
		pContent_sequence->aFieldValue = NULL;

		if (read_Sequence
		    (pNodeSubSequence, pContent_sequence, pBuffer,
		     sbError) == FAILURE) {
		    sbInsert_int(sbError,
				 "error reading sub-sequence in field %d: ",
				 xField);
		    return FAILURE;
		}

		if (pContent_sequence->aeFieldType == NULL) {
		    sbInsert_int(sbError,
				 "validation error reading sub-sequence in field %d, no field type vector created",
				 xField);
		    return FAILURE;
		}
		if (pContent_sequence->aFieldValue == NULL) {
		    sbInsert_int(sbError,
				 "validation error reading sub-sequence in field %d, no value type vector created",
				 xField);
		    return FAILURE;
		}


		// The call above to read_Sequence may cause the
		// pNodeContent_Sequence->aFieldValue to be
		// reallocated, in which case, we have to recompute
		// pVector.
		pVector = *(pNodeContent_Sequence->aFieldValue + (xField - 1));

		*((NodeContent_Sequence **) pVector + xRow0) =
		    pContent_sequence;

		DBG(printf
		    ("stored sequence at %p in location %p\n",
		     (void *) pContent_sequence,
		     (void *) ((NodeContent_Sequence **) pVector + xRow0));
		    )

	    }
	    break;
/*
typedef struct {
	unsigned int   ctMembers; 	The number of members of the Structure.
	eNODE_TYPE*    array_eNodeType; 	An array of node types, one for each Structure member. 	These may be any \c eNODE_TYPE except \c root.
	void**          array_NodeContent;      This array contains a pointer to an \c array_NodeContent structure for each Structure member.
} NodeContent_Structure;
*/
	case NODE_Structure:;
	    {
		NodeContent_Structure *pContent_structure;
		unsigned int xMember, ctStructureMembers;

		/* allocate a content node for the structure */
		pContent_structure = malloc(sizeof(NodeContent_Structure));
		if (pContent_structure == NULL) {
		    sbAppend_int(sbError,
				 "failed to allocate memory for structure in field %d",
				 xField);
		    return FAILURE;
		}

		/* create the content node for the structure */
		ctStructureMembers = pTemplateFieldNode->iSubNodeCount;
		pContent_structure->ctMembers = ctStructureMembers;
		pContent_structure->array_eNodeType =
		    /* pedantic but correct*/
		    malloc(sizeof(eNODE_TYPE) * ctStructureMembers);

		if (pContent_structure->array_eNodeType == NULL) {
		    sbAppend_int(sbError,
				 "failed to allocate memory for type array of structure in field %d",
				 xField);
		    return FAILURE;
		}
		pContent_structure->array_NodeContent =

		    /* pedantic*/
		    malloc(sizeof(void *) * ctStructureMembers);

		if (pContent_structure->array_NodeContent == NULL) {
		    sbAppend_int(sbError,
				 "failed to allocate memory for node content array of structure in field %d",
				 xField);
		    return FAILURE;
		}

		DBG(printf
		    ("reading structure with %i members\n",
		     ctStructureMembers);
		    )
		    for (xMember = 1; xMember <= ctStructureMembers;
			 xMember++) {
		    DAPNode *pnodeTemplate_Member =
			*(pTemplateFieldNode->nodeSub + xMember - 1);
		    pContent_structure->array_eNodeType[xMember - 1] =
			pnodeTemplate_Member->eType;

	        pContent_structure->array_NodeContent[xMember - 1]
			    = clone_NodeTemplate(pnodeTemplate_Member,sbError);

		    DBG(printf("reading structure member %i\n", xMember);
			)
			if (read_Node
			    (pnodeTemplate_Member,
			     pContent_structure->
			     array_NodeContent+xMember - 1, pBuffer,

			     sbError) == FAILURE) {
			sbInsert_int(sbError,
				     "error reading member %i of sub-structure: ",
				     xMember);
			return FAILURE;
		    }
		}

		// The call above to read_Node may cause the
		// pNodeContent_Sequence->aFieldValue to be
		// reallocated, in which case, we have to recompute
		// pVector.
		pVector = *(pNodeContent_Sequence->aFieldValue + (xField - 1));

		*(((NodeContent_Structure **) pVector) + xRow0) = pContent_structure;	/* insert content value in row */
		DBG(printf
		    ("created structure content at %p\n",
		     (void *) pContent_structure);
		    )
/*
						if( imp_structure_VerifyStructure( pTemplateFieldNode, pContent_structure, sbError ) == FAILURE ){
							printf("failed verification\n");
						} else {
							printf("passed verification\n");
						}
*/
	    }
	    break;
	case NODE_root:;
	    sbAppend_int(sbError, "invalid type [root] in field %d",
			 xField);
	    return FAILURE;
	default:
	    sbAppend_int(sbError, "undefined type %d", eFieldType);
	    sbAppend_int(sbError, " in field %d", xField);
	    return FAILURE;
	}
    }
    return SUCCESS;
}

void *clone_NodeTemplate(DAPNode * pnodeTemplate, StringBuffer * sbError)
{
    switch (pnodeTemplate->eType) {
    case NODE_Byte:;		/* bytes are stored directly in the vector with no xdr translation */
    case NODE_Int16:;
    case NODE_UInt16:;
    case NODE_Int32:;
    case NODE_UInt32:;
    case NODE_Float32:;
    case NODE_Float64:;
	return NULL;
    case NODE_String:;
    case NODE_URL:;
	{
	    NodeContent_String* pContent_string
		= (NodeContent_String*)malloc(sizeof(NodeContent_String));
	    if (pContent_string == NULL) {
		sbAppend(sbError,
			 "failed to allocate memory for cloned string");
		return NULL;
	    }
	    pContent_string->iLength = 0;
	    pContent_string->sContent = NULL;
	    return pContent_string;
	} break;

    case NODE_Array:;
	{
	    /* duplicate the template content for the array */
	    NodeContent_Array *pTemplateArrayContent =
		pnodeTemplate->nodeContent;
	    unsigned int xDimension = 0;
	    NodeContent_Array *pContent_array =
		malloc(sizeof(NodeContent_Array));
	    if (pContent_array == NULL) {
		sbAppend(sbError,
			 "failed to allocate memory for cloned array");
		return NULL;
	    }
	    pContent_array->eType = pTemplateArrayContent->eType;
	    pContent_array->ctDimensions =
		pTemplateArrayContent->ctDimensions;
	    pContent_array->iDimensionLengthArray1 =
		malloc(sizeof(unsigned int) *
		       (pContent_array->ctDimensions + 1));
	    if (pContent_array->iDimensionLengthArray1 == NULL) {
		sbAppend(sbError,
			 "failed to allocate memory for array dim size");
		return NULL;
	    }
	    pContent_array->sDimensionName1 =
		malloc(sizeof(char *) *
		       (pContent_array->ctDimensions + 1));
	    if (pContent_array->sDimensionName1 == NULL) {
		sbAppend(sbError,
			 "failed to allocate memory for array dim names");
		return NULL;
	    }
	// Make sure
	    pContent_array->iDimensionLengthArray1[0] = 0;
	    pTemplateArrayContent->sDimensionName1[0] = NULL;

	    for (; xDimension <= pContent_array->ctDimensions;
		 xDimension++) {
		pContent_array->iDimensionLengthArray1[xDimension] =
		    pTemplateArrayContent->
		    iDimensionLengthArray1[xDimension];
		pContent_array->sDimensionName1[xDimension] =
		    sCloneString(pTemplateArrayContent->
				 sDimensionName1[xDimension]);
	    }
	    // Original code forgot to fill in this field
	    pContent_array->sArrayName = sCloneString(pTemplateArrayContent->sArrayName);

	    pContent_array->array = NULL;
	    return pContent_array;
	}
    case NODE_Grid:;
    case NODE_Sequence:;
	break;
    case NODE_Structure:;
	// Apparently nested structures require presence of existing
	// NodeContent_Structure instance
	{ unsigned int ctmem, imem;
	    NodeContent_Structure* pContent_structure
	       = (NodeContent_Structure*)malloc(sizeof(NodeContent_Structure));
	    if (pContent_structure == NULL) {
		sbAppend(sbError,
			 "failed to allocate memory for cloned structure");
		return NULL;
	    }
	    ctmem = pnodeTemplate->iSubNodeCount;
	    pContent_structure->ctMembers = ctmem;
	    pContent_structure->array_eNodeType
		= (eNODE_TYPE*)malloc(sizeof(eNODE_TYPE)*ctmem);
	    pContent_structure->array_NodeContent
		= (void**)malloc(sizeof(void*)*ctmem);
	    for(imem=0;imem<pContent_structure->ctMembers;imem++) {
		DAPNode* subnode = pnodeTemplate->nodeSub[imem];
	        pContent_structure->array_eNodeType[imem] = subnode->eType;
	        pContent_structure->array_NodeContent[imem]
		    = clone_NodeTemplate(subnode,sbError);
	    }
	    return pContent_structure;
	} break;

    case NODE_root:;
    default:
	return NULL;
    }
    return NULL;
}

eRESULT clone_ArrayTemplate(DAPNode * pTemplateNode,
			    NodeContent_Array ** hContent_array,
			    StringBuffer * sbError)
{

    /* duplicate the template content for the array */
    NodeContent_Array *pTemplateArrayContent = pTemplateNode->nodeContent;
    unsigned int xDimension = 0;
    NodeContent_Array *pContent_array = malloc(sizeof(NodeContent_Array));
    if (pContent_array == NULL) {
	sbAppend(sbError, "failed to allocate memory for cloned array");
	return FAILURE;
    }
    pContent_array->eType = pTemplateArrayContent->eType;
    pContent_array->ctDimensions = pTemplateArrayContent->ctDimensions;
    pContent_array->iDimensionLengthArray1 =
	malloc(sizeof(unsigned int) * (pContent_array->ctDimensions + 1));
    if (pContent_array->iDimensionLengthArray1 == NULL) {
	sbAppend(sbError,
		 "failed to allocate memory for cloned array dim size");
	return FAILURE;
    }
    pContent_array->sDimensionName1 =
	malloc(sizeof(char *) * (pContent_array->ctDimensions + 1));
    if (pContent_array->sDimensionName1 == NULL) {
	sbAppend(sbError,
		 "failed to allocate memory for cloned array dim names");
	return FAILURE;
    }
    for (; xDimension <= pContent_array->ctDimensions; xDimension++) {
	pContent_array->iDimensionLengthArray1[xDimension] =
	    pTemplateArrayContent->iDimensionLengthArray1[xDimension];
	pContent_array->sDimensionName1[xDimension] =
	    sCloneString(pTemplateArrayContent->
			 sDimensionName1[xDimension]);
    }
    pContent_array->array = NULL;

    *hContent_array = pContent_array;

    return SUCCESS;
}
