/*
 * This file is part of the "OCAPI" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005 OPeNDAP, Inc.
 * Author: John Chamberlain <jchamberlain at opendap.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */


/*#include "config.h"*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

/*Turn this on so the DBG statements will print for debugging purposes--remove if do not want the printout*/
/*#define DODS_DEBUG*/


#include "global.h"

#define DATA_HEADER "\nData:\n"
#define BUFSIZE 512
#define DODSRC_SIZE 9
#define DODSRC "/.dodsrc"

/***************/
/**  Loading  **/
/***************/

eRESULT zFile_LoadStructure(const char *sFilename, DAPObject ** hStructure,
			    StringBuffer * sbError);
eRESULT zFile_LoadContent(const char *sFilename, char **hContent);
extern int parse_buffer(void *pBuffer);	/* defined in http.c */
extern eRESULT read_Node(DAPNode * nodeTemplate, void *nodeContent_Supplied, Buffer * pBuffer, StringBuffer * sbError);	/* defined in data.c */
extern eRESULT zParseDAS(DAPObject ** hStructure, Buffer * pBuffer, StringBuffer * sbError);	/* defined in parseutil.c */

/***************************/
/**  Node Management      **/
/***************************/

eRESULT not_implemented(char **sError);
eRESULT free_ArrayContent(NodeContent_Array * pArray,
			  StringBuffer * sbError);
eRESULT free_GridContent(NodeContent_Grid * pGrid, StringBuffer * sbError);
eRESULT free_SequenceContent(NodeContent_Sequence * pSequenceContent,
			     StringBuffer * sbError);
eRESULT free_StructureContent(NodeContent_Structure * pStructureContent,
			      StringBuffer * sbError);
eRESULT free_Attributes(DAPAttributeTable * attributes,
			StringBuffer * sbError);
DAPAttributeTable *attribute_NewTable(StringBuffer * sbError);
DAPAttribute *attribute_New(eATTRIBUTE_TYPE eType, char *sStackName,
			    StringBuffer * sbError);
eRESULT attribute_Add(DAPAttributeTable * pTable, char *sName,
		      DAPAttribute * pAttribute, StringBuffer * sbError);
eRESULT attribute_AddValue(DAPAttribute * pAttribute, void *pStackValue,
			   StringBuffer * sbError);
DAPAttributeTable *clone_Attributes(DAPAttributeTable * attributes,
				    StringBuffer * sbError);
DAPNode *nodeCreate(eNODE_TYPE type);
NodeContent_Array *arrayCreate();
int zIsInsideSequence(DAPNode * node);	/* returns 1 if true, 0 if false or invalid */

/* also, see implementation prototypes in global.h */

/***************************/
/**  Node Validation      **/
/***************************/

/* see prototypes in global.h */

/***************************/
/**       Printing        **/
/***************************/

eRESULT print_Node_toFile(DAPNode * node, int iIndent, FILE * stream,
			  StringBuffer * sbError);
eRESULT print_NodeContent_toFile(DAPNode * node, void *pNestedContent,
				 int iIndent, FILE * stream,
				 StringBuffer * sbError);
eRESULT print_Grid_toFile(NodeContent_Grid * pGrid, int iIndent,
			  FILE * stream, StringBuffer * sbError);
eRESULT print_Array_toFile(NodeContent_Array * pArray, int iIndent,
			   FILE * stream, StringBuffer * sbError);
eRESULT print_Sequence_toFile(DAPNode * pSequenceNode,
			      NodeContent_Sequence * pSequence,
			      int iIndent, FILE * stream,
			      StringBuffer * sbError);
eRESULT print_Structure_toFile(DAPNode * pStructureNode,
			       NodeContent_Structure * pStructure,
			       int iIndent, FILE * stream,
			       StringBuffer * sbError);
eRESULT print_DAS_toFile(DAPObject * pDAPObject, FILE * stream,
			 StringBuffer * sbError);
eRESULT print_DASNode_toFile(DAPNode * pStructureNode, int iIndent,
			     FILE * stream, StringBuffer * sbError);
eRESULT print_Attributes_toFile(DAPAttributeTable * attributes,
				int iIndent, FILE * stream);
eRESULT node_AddDAS(DAPNode * nodeDAS, DAPNode * nodeTarget,
		    StringBuffer * sbError);
eRESULT node_AddNode(DAPNode * pNode_super, DAPNode * pNode_sub,
		     StringBuffer * sbError);
char *render_STRUCTURE_TYPE(eSTRUCTURE_TYPE e);
char *render_NODE_TYPE(eNODE_TYPE e);
char *render_ARRAY_TYPE(eARRAY_TYPE e);
char *render_ATTRIBUTE_TYPE(eATTRIBUTE_TYPE e);
void indent(int i, FILE * stream);

/***************************/
/**       Accessors       **/
/***************************/

DAPNode *getNodeByName_Recursive(DAPNode * nodeParent,
				 StringList * bufferQualifiedName,
				 int xName1);

extern eRESULT zInitialize_http(ConnectionParameters ** hDefaultConnectionParameters, StringBuffer * sbError);	/* defined in http.c */
extern eRESULT zConnection_FetchWithParameters(ConnectionParameters * parameters, Buffer * pBuffer, StringBuffer * sbError);	/* defined in http.c */
extern eRESULT zConnection_Fetch(char *sURL, Buffer * pBuffer, ThreadSpecificStorage * pTSS, StringBuffer * sbError);	/* defined in http.c */
extern eRESULT zConnection_Load(char *sURL, DAPObject ** hStructure, ThreadSpecificStorage * pTSS, StringBuffer * sbError);	/* defined in http.c */
extern Buffer *bufferLoadFile(const char *sFilename);	/* defined in buffers.c */

void vFreeConnectionParameters(ConnectionParameters * parameters);

/***************************/
/**     Initialization    **/
/***************************/

eRESULT zInitializeOCAPI(ThreadSpecificStorage ** hTSS, char **sError) {
    char buf[BUFSIZE];
    char *env;
    int len;
#if 0
    FILE *out_file;
    FILE *in_file;
#endif
    ConnectionParameters *pConnectionParameters;
    StringBuffer *sbError = sbNew(NULL);
    ThreadSpecificStorage *pTSS = malloc(sizeof(ThreadSpecificStorage));
    if (pTSS == NULL) {
	sbAppend(sbError,
		"Failed to allocate memory for thread-specific storage.");
	*hTSS = NULL;
	return FAILURE;
    } else {
	*hTSS = pTSS;
    }

    if (zInitialize_http(&pConnectionParameters, sbError) == SUCCESS) {
	pTSS->pConnectionParameters = pConnectionParameters;

	env = getenv("HOME");
	if (env != NULL) {
	    len = strlen(env);
	    if (len >= BUFSIZE - DODSRC_SIZE) {
		if (sError != NULL) {
		    StringBuffer *sb = sbNew(
			    "length of home directory is too long");
		    *sError = sb->sContent;
		    free(sb);
		}
		return FAILURE;
	    }

	    strncpy(buf, env, BUFSIZE - 1);
	    buf[len] = '\0';
	    strncat(buf, DODSRC, BUFSIZE - 1);
	    buf[len + DODSRC_SIZE] = '\0';
#if 0
	    fprintf(stderr, "Your Home: %s\n", buf);
#endif
	    if (access(buf, R_OK) != 0) {
		if (write_dodsrc(buf, sError) == FAILURE) {
		    sbInsert_string(sbError, "Error getting %s: ", buf);
		    *sError = sbError->sContent;
		    free(sbError);
		    return FAILURE;
		}
	    }

	    if (read_dodsrc(pTSS, buf, sError) == FAILURE) {
		sbInsert_string(sbError, "Error parsing %s: ", buf);
		*sError = sbError->sContent;
		free(sbError);
		return FAILURE;
	    }
	}

	sbFree(sbError);
	return SUCCESS;
    } else {
	sbInsert(sbError, "Error initializing: ");
	/* TODO: Examine this to make sure that sContent still exists once the
	 * sbError structure is freed. 5/9/08 */
	*sError = sbError->sContent;
	free(sbError);
	return FAILURE;
    }
}

void vCleanupOCAPI(ThreadSpecificStorage * pTSS)
{
    if (pTSS == NULL) {
#if 0
	fprintf(stderr, "warning, null pointer supplied to cleanup");
#endif
	return;
    }
    if (pTSS->pConnectionParameters == NULL) {
#if 0
	fprintf(stderr,
		"warning, null connection parameters supplied to cleanup");
#endif
	return;
    }
    /*vCleanup needs to free TSS, added here. rd 06/18/08*/
    vFreeConnectionParameters(pTSS->pConnectionParameters);
    if (pTSS)
      free(pTSS);
    return;
}

/***************************/
/**     Informational     **/
/***************************/

char *getVersion(void)
{
    return VERSION_STRING;
}

char *getVersionDate(void)
{
    return VERSION_DATE;
}

char *getConfiguration(ThreadSpecificStorage * pTSS)
{
    ConnectionParameters *pConnectionParameters;
    StringBuffer *sbConfiguration;
    sbConfiguration = sbNew(NULL);
#ifdef _WIN32
    sbAppend(sbConfiguration, "platform: Win32\n");
#endif
    if (pTSS == NULL) {
	sbAppend(sbConfiguration,
		 "internal error, no thread-specific storage\n");
	return sbConfiguration->sContent;
    }
    if (pTSS->pConnectionParameters == NULL) {
	sbAppend(sbConfiguration,
		 "internal error, no configuration parameters exist\n");
	return sbConfiguration->sContent;
    } else {
	pConnectionParameters = pTSS->pConnectionParameters;
    }
    /* sbAppend_string( sbConfiguration, "LogLevel = %s\n", renderLogLevel( geLogLevel ) ); */
    sbAppend_int(sbConfiguration, "ConnectTimeout_seconds = %d\n",
		 pConnectionParameters->iConnectTimeout_seconds);
    sbAppend_int(sbConfiguration, "ReadTimeout_seconds = %d\n",
		 pConnectionParameters->iReadTimeout_seconds);
    sbAppend_string(sbConfiguration, "Protocol = \"%s\"\n",
		    pConnectionParameters->sProtocol);
    if (pConnectionParameters->pLoggingStream == LOGGING_None) {
	sbAppend(sbConfiguration, "Logging Destination = [none]\n");
    } else if (pConnectionParameters->pLoggingStream == stdout) {
	sbAppend(sbConfiguration, "Logging Destination = stdout\n");
    } else {
	/* %p is implementation specific; use %x or %lx depending
	 * on the size of a pointer (32- or 64-bit) jhrg 2/1/06
	 * This should really test to see if void * matches %lx, but
	 * I don't think code is cricital, so this is good enough. I
	 * just got tired of seeing the warnings. jhrg 5/25/06 */
#if SIZEOF_VOID_P == 4
	sbAppend_int(sbConfiguration,
		     "Logging Destination = file handle: %x\n",
		     (int) pConnectionParameters->pLoggingStream);
#else
	sbAppend_int(sbConfiguration,
		     "Logging Destination = file handle: %lx\n",
		     (long) pConnectionParameters->pLoggingStream);
#endif
    }
    sbAppend_string(sbConfiguration, "UserAgent = %s\n",
		    pConnectionParameters->sUserAgent);
    return sbConfiguration->sContent;
}

eRESULT configure(ThreadSpecificStorage * pTSS, /** user must supply */
		int iConnectTimeout_seconds, /** defaults to 30, -1 = no change, 0 = infinite */
		int iReadTimeout_seconds, /** defaults to 60, -1 = no change, 0 = infinite */
		char *sProtocol, /** defaults to "HTTP/1.0", NULL or "" = no change */
		FILE * pLoggingStream, /** defaults to 0 = none, -1 = no change */
		/* not used. jhrg 4/3/2007 eLOG_LEVEL    eLogLevel, *//** defaults to LL_Normal */
		char *sUserAgent, char **sError) {
	ConnectionParameters *pConnectionParameters;
	if (pTSS == NULL) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew("no thread-specific storage supplied");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    }
    if (pTSS->pConnectionParameters == NULL) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew(
	    "internal error, thread-specific storage lacked a connect parameters member");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    } else {
	pConnectionParameters = pTSS->pConnectionParameters;
    }
    if (iConnectTimeout_seconds < -1) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew_int("invalid connect timeout %d",
		    iConnectTimeout_seconds);
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    }
    if (iReadTimeout_seconds < -1) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew_int("invalid read timeout %d",
		    iReadTimeout_seconds);
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    }
    if (iConnectTimeout_seconds != -1)
	pConnectionParameters->iConnectTimeout_seconds
		= iConnectTimeout_seconds;
    if (iReadTimeout_seconds != -1)
	pConnectionParameters->iReadTimeout_seconds = iReadTimeout_seconds;
    if (sProtocol != NULL && strlen(sProtocol) != 0)
	pConnectionParameters->sProtocol = sProtocol;
    /* There is 32/64 bit issue in the following line; should not cast a
     * pointer to an int on a 64-bit machine. jhrg 2/1/06
     * Moved/changed cast. jhrg 5/25/06 */
    if ( /*(int) */pLoggingStream != (void *) -1)
	pConnectionParameters->pLoggingStream = pLoggingStream;
    if (sUserAgent != NULL)
	pConnectionParameters->sUserAgent = sUserAgent;
    return SUCCESS;
}

/*Added the following to configure connection to a proxy server*/
eRESULT configure_proxy(ThreadSpecificStorage * pTSS, char *sHost, int iPort,
	char *sUser, char *sPassword, char **sError) {
    ConnectionParameters *pConnectionParameters;
    if (pTSS == NULL) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew("no thread-specific storage supplied");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    }
    if (pTSS->pConnectionParameters == NULL) {
	if (sError != NULL) {
	    StringBuffer
		    *sb =
			    sbNew(
				    "internal error, thread-specific storage lacked a connect parameters member");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    } else {
	pConnectionParameters = pTSS->pConnectionParameters;
    }

    if (sHost == NULL && (iPort != -1 || sUser != NULL || sPassword != NULL)) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew("no host detected");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    }

    if (sHost != NULL) {
	pConnectionParameters->pstructProxy = malloc(sizeof(Proxy));
	if (pConnectionParameters->pstructProxy == NULL) {
	    if (sError != NULL) {
		StringBuffer *sb = sbNew("failed to allocate memory for proxy");
		*sError = sb->sContent;
		free(sb);
	    }
	    return FAILURE;
	}
	pConnectionParameters->pstructProxy->sHost = sHost;

	/** These parameters must be given as either specific values or the
	 *  expected defaults. jhrg 24/10/08
	 */
	pConnectionParameters->pstructProxy->iPort = iPort;
	if (sUser == NULL && sPassword != NULL) {
	    if (sError != NULL) {
		StringBuffer *sb = sbNew("no username given");
		*sError = sb->sContent;
		free(sb);
	    }
	    return FAILURE;
	}
	if (sUser != NULL && sPassword == NULL) {
	    if (sError != NULL) {
		StringBuffer *sb = sbNew("no password given");
		*sError = sb->sContent;
		free(sb);
	    }
	    return FAILURE;
	}
	if (sUser != NULL && sPassword != NULL) {
	    pConnectionParameters->pstructProxy->sUser = sUser;
	    pConnectionParameters->pstructProxy->sPassword = sPassword;
	}
	if (sUser == NULL && sPassword == NULL) {
	    pConnectionParameters->pstructProxy->sUser = sUser;
	    pConnectionParameters->pstructProxy->sPassword = sPassword;
	}
    }
    return SUCCESS;
}
/*Added the following to configure connection to a http server*/
eRESULT configure_httpAuth(ThreadSpecificStorage * pTSS, char *sUserC,
	    char *sPasswordC,
	    char **sError){
		ConnectionParameters *pConnectionParameters;
		if (pTSS == NULL) {
		    if (sError != NULL) {
			StringBuffer *sb = sbNew("no thread-specific storage supplied");
			*sError = sb->sContent;
			free(sb);
		    }
		    return FAILURE;
		}
		if (pTSS->pConnectionParameters == NULL) {
		    if (sError != NULL) {
			StringBuffer *sb = sbNew(
			"internal error, thread-specific storage lacked a connect parameters member");
			*sError = sb->sContent;
			free(sb);
		    }
		    return FAILURE;
		} else {
			pConnectionParameters = pTSS->pConnectionParameters;
		}
		if (sUserC == NULL && sPasswordC != NULL){
		    if (sError != NULL) {
			StringBuffer *sb = sbNew(
			"no username given");
			*sError = sb->sContent;
			free(sb);
			}
		    return FAILURE;
		}
		if (sUserC != NULL && sPasswordC == NULL){
		    if (sError != NULL) {
			StringBuffer *sb = sbNew(
			"no password given");
			*sError = sb->sContent;
			free(sb);
			}
		    return FAILURE;
		}
		if (sUserC != NULL && sPasswordC != NULL){
		    pConnectionParameters->sUserC = sUserC;
		    pConnectionParameters->sPasswordC = sPasswordC;
		}
		if (sUserC == NULL && sPasswordC == NULL){
		    pConnectionParameters->sUserC = sUserC;
		    pConnectionParameters->sPasswordC = sPasswordC;
		}

	return SUCCESS;
}

/*Added the following to configure the use of cookies*/
eRESULT configure_cookies(ThreadSpecificStorage * pTSS, char *cook,
	char **sError) {
    ConnectionParameters *pConnectionParameters;
    if (pTSS == NULL) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew("no thread-specific storage supplied");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    }
    if (pTSS->pConnectionParameters == NULL) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew(
	    "internal error, thread-specific storage lacked a connect parameters member");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    } else {
	pConnectionParameters = pTSS->pConnectionParameters;
    }

    pConnectionParameters->cook = cook;

    return SUCCESS;
}
/*Added the following to allow the user to shut off SSL verification.*/
eRESULT configure_ssl(ThreadSpecificStorage * pTSS, int sVerify, char **sError) {
    ConnectionParameters *pConnectionParameters;
    if (pTSS == NULL) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew("no thread-specific storage supplied");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    }
    if (pTSS->pConnectionParameters == NULL) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew(
	    "internal error, thread-specific storage lacked a connect parameters member");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    } else {
	pConnectionParameters = pTSS->pConnectionParameters;
    }

    pConnectionParameters->sVerify = sVerify;

    return SUCCESS;

}
/*Added the following to allow for compression.*/
eRESULT configure_compression(ThreadSpecificStorage * pTSS, char *sCompress,
	char **sError) {
    ConnectionParameters *pConnectionParameters;
    if (pTSS == NULL) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew("no thread-specific storage supplied");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    }
    if (pTSS->pConnectionParameters == NULL) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew(
	    "internal error, thread-specific storage lacked a connect parameters member");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    } else {
	pConnectionParameters = pTSS->pConnectionParameters;
    }

    pConnectionParameters->sCompress = sCompress;

    return SUCCESS;

}

/*Allows for a .dodsrc file to be read in and parsed in order to get authentication information*/
eRESULT read_dodsrc(ThreadSpecificStorage * pTSS, char *in_file_name, char **sError) {
    char *p;
    char more[1024];
    char *v;
    FILE *in_file;
    ConnectionParameters *pConnectionParameters;

    if (pTSS == NULL) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew("no thread-specific storage supplied");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    }
    if (pTSS->pConnectionParameters == NULL) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew("internal error, thread-specific storage lacked a connect parameters member");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    } else {
	pConnectionParameters = pTSS->pConnectionParameters;
    }

    in_file = fopen(in_file_name, "r"); /* Open the file to read it */
    if (in_file == NULL) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew("Could not open the .dodsrc file");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    }

    while (fgets(more, 1024, in_file) != NULL) {
	if (more[0] != '#') {
	    p = strchr(more, '=');
	    v = p + 1;
	    *p = '\0';

	    if (strcmp(more, "USE_CACHE") == 0) {
#if 0
		fprintf(stderr, "Does Not support Caching at this time\n");
#endif
	    } else if (strcmp(more, "MAX_CACHE_SIZE") == 0) {
#if 0
		fprintf(stderr, "Does Not support Caching at this time\n");
#endif
	    } else if (strcmp(more, "MAX_CACHED_OBJ") == 0) {
#if 0
		fprintf(stderr, "Does Not support at this time\n");
#endif
	    } else if (strcmp(more, "IGNORE_EXPIRES") == 0) {
#if 0
		fprintf(stderr, "Does Not support at this time\n");
#endif
	    } else if (strcmp(more, "CACHE_ROOT") == 0) {
#if 0
		fprintf(stderr, "Does Not support at this time\n");
#endif
	    } else if (strcmp(more, "DEFAULT_EXPIRES") == 0) {
#if 0
		fprintf(stderr, "Does Not support at this time\n");
#endif
	    } else if (strcmp(more, "ALWAYS_VALIDATE") == 0) {
#if 0
		fprintf(stderr, "Does Not support at this time\n");
#endif
	    } else if (strcmp(more, "DEFLATE") == 0) {
		int v_len = strlen(v);
		pTSS->pConnectionParameters->sCompress = malloc(sizeof(char) * v_len + 1);
		strncpy(pTSS->pConnectionParameters->sCompress, v, v_len);
		pTSS->pConnectionParameters->sCompress[v_len] = '\0';
	    } else if (strcmp(more, "VALIDATE_SSL") == 0) {
		pTSS->pConnectionParameters->sVerify = atoi(v);
	    } else if (strcmp(more, "PROXY_SERVER") == 0) {
			int s_len;
			int p_len;
		int v_len = strlen(v);
		pTSS->pConnectionParameters->pstructProxy->sHost = malloc(sizeof(char) * v_len +1);
		strncpy(pTSS->pConnectionParameters->pstructProxy->sHost, v, v_len);
		pTSS->pConnectionParameters->pstructProxy->sHost[v_len] = '\0';

		pTSS->pConnectionParameters->pstructProxy->iPort = atoi(v);

		s_len = strlen(v);
		pTSS->pConnectionParameters->pstructProxy->sUser = malloc(sizeof(char) * s_len +1);
		strncpy(pTSS->pConnectionParameters->pstructProxy->sUser, v, s_len);
		pTSS->pConnectionParameters->pstructProxy->sUser[s_len] = '\0';

		p_len = strlen(v);
		pTSS->pConnectionParameters->pstructProxy->sPassword = malloc(sizeof(char) * p_len +1);
		strncpy(pTSS->pConnectionParameters->pstructProxy->sPassword, v, p_len);
		pTSS->pConnectionParameters->pstructProxy->sPassword[p_len] = '\0';

	    } else if (strcmp(more, "NO_PROXY_FOR") == 0) {
#if 0
		fprintf(stderr, "Does Not support at this time\n");
#endif
	    } else if (strcmp(more, "AIS_DATABASE") == 0) {
#if 0
		fprintf(stderr, "Does Not support at this time\n");
#endif
	    }else if (strcmp(more, "COOKIE_JAR") == 0) {
		int v_len = strlen(v);
		pTSS->pConnectionParameters->cook = malloc(sizeof(char) * v_len +1);
		strncpy(pTSS->pConnectionParameters->cook, v, v_len);
		pTSS->pConnectionParameters->cook[v_len] = '\0';
	    }

	}
    }
    fclose(in_file);

    return SUCCESS;
}

/*Allows for a .dodsrc file to be created if one does not currently exist for default authentication
 *  values*/
eRESULT write_dodsrc(char *out_file_name, char **sError) {
    char *authent[] = { "#DODS client configuation file. See the DODS\n",
	    "#users guide for information.\n",
	    "#USE_CACHE=0\n",
	    "#Cache and object size are given in megabytes (20 ==> 20Mb).\n",
	    "#MAX_CACHE_SIZE=20\n",
	    "#MAX_CACHED_OBJ=5\n",
	    "#IGNORE_EXPIRES=0\n",
	    "#CACHE_ROOT=/Users/jimg/.dods_cache/\n",
	    "#DEFAULT_EXPIRES=86400\n",
	    "#ALWAYS_VALIDATE=0\n",
	    "# Request servers compress responses if possible?\n",
	    "# 1 (yes) or 0 (false).\n",
	    "# Should SSL certificates and hosts be validated? SSL\n",
	    "# will only work with signed certificates.\n",
	    "VALIDATE_SSL=0\n"
	    "DEFLATE=0\n",
	    "# Proxy configuration (optional parts in []s):\n",
	    "#PROXY_SERVER=http://[username:password@]host[:port]\n",
	    "#NO_PROXY_FOR=host|domain\n",
	    "# AIS_DATABASE=<file or url>\n",
	    "# The cookie jar is a file that holds cookies sent from\n",
	    "# servers such as single signon systems. Uncomment this\n",
	    "# option and provide a file name to activate this feature.\n",
	    "# If the value is a filename, it will be created in this\n",
	    "# directory; a full pathname can be used to force a specific\n",
	    "# location.\n",
	    "# COOKIE_JAR=.dods_cookies\n" };

    unsigned int i = 0;		/* Added 'unsigned' jhrg 11/20/08 */
    FILE *out_file = fopen(out_file_name, "w");
    if (out_file == NULL) {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew("cannot open output file");
	    *sError = sb->sContent;
	    free(sb);
	}

	return FAILURE;
    }

    for (i = 0; i < sizeof authent / sizeof authent[0]; i++)
	fputs(authent[i], out_file);

    fclose(out_file);

    return SUCCESS;
}

/*************************/
/** Generic HTTP Access **/
/*************************/
eRESULT parseURL(char *pURL, char **sUser, char **sURL, char **sPassword,
		 char **sError){

    char *pos;
    char *end;
    char *middle;
#if 0
    char *midpas;
#endif
    int up_len = 0;
    int mid_len = 0;
    int midpas_len = 0;
    char *sUP = 0;
    int url_len = 0;


    if (strchr(pURL, '@')) {
	pos = strstr(pURL, "://");
	if (pos)
	    pos += 3;
	middle = strstr(pos, ":");
	mid_len = middle - pos;
	*sUser = malloc(sizeof(char) * (mid_len + 1));
	strncpy(*sUser, pos, mid_len);
	(*sUser)[mid_len] = '\0';

	if (middle)
	    middle += 1;

	end = strstr(middle, "@");
	midpas_len = (end) - (middle);
	*sPassword = malloc(sizeof(char) * (midpas_len + 1));
	strncpy(*sPassword, middle, midpas_len);
	(*sPassword)[midpas_len] = '\0';

	end = strstr(pURL, "@");
	up_len = end - pos;
	sUP = malloc(sizeof(char) * (up_len + 1));
	strncpy(sUP, pos, up_len);
	sUP[up_len] = '\0';

	url_len = strlen(pURL) - up_len;

	*sURL = malloc(sizeof(char) * (url_len + 1));
	strncpy(*sURL, pURL, pos - pURL);
	strncpy(*sURL + (pos - pURL), end + 1, url_len - (pos - pURL));
#if 0
	fprintf(stderr, "URL without username and password: %s:%d\n", sURL, url_len );
	fprintf(stderr, "URL username and password: %s:%d\n", sUP, up_len);
	fprintf(stderr, "URL username: %s:%d\n", sUser, mid_len);
	fprintf(stderr, "URL password: %s:%d\n", sPassword, midpas_len);
#endif
	(*sURL)[url_len] = '\0';


	return SUCCESS;
    } else {
	if (sError != NULL) {
	    StringBuffer *sb = sbNew(
				     "error getting new url");
	    *sError = sb->sContent;
	    free(sb);
	}
	return FAILURE;
    }
}


eRESULT getContent_byURL(URL * url, char **content,
			 ThreadSpecificStorage * pTSS, char **sError)
{
    char *sURL = url->sCanonicalURL;
    return getContent_byRawURL(sURL, content, pTSS, sError);
}

eRESULT getContent_byRawURL(char *sURL, char **content,
			    ThreadSpecificStorage * pTSS, char **sError)
{
    Buffer *pBuffer = bufferNew(0);
    StringBuffer *sbError = sbNew(NULL);
    if (zConnection_Fetch(sURL, pBuffer, pTSS, sbError) == SUCCESS) {
	*content = pBuffer->pcharContent;
	free(pBuffer);
	return SUCCESS;
    } else {
	sbInsert_string(sbError, "Error getting %s: ", sURL);
	*sError = sbError->sContent;
	free(sbError);
	return FAILURE;
    }
}

eRESULT getContent_byParameters(ConnectionParameters * parameters,
				char **content, char **sError)
{
    Buffer *pBuffer = bufferNew(0);
    StringBuffer *sbError = sbNew(NULL);
    if (zConnection_FetchWithParameters(parameters, pBuffer, sbError) ==
	SUCCESS) {
	*content = pBuffer->pcharContent;
	free(pBuffer);
	return SUCCESS;
    } else {
	sbInsert(sbError, "Error fetching: ");
	*sError = sbError->sContent;
	free(sbError);
	return FAILURE;
    }
}


/************************/
/** Structure Handling **/
/************************/
eRESULT structure_Load_URL(URL * url, DAPObject ** structure,
			   ThreadSpecificStorage * pTSS, char **sError)
{
    StringBuffer *sbError = sbNew(NULL);
    char *sURL = url->sCanonicalURL;
    if (zConnection_Load(sURL, structure, pTSS, sbError) == SUCCESS) {
	sbFree(sbError);
	return SUCCESS;
    } else {
	sbInsert_string(sbError, "Error loading %s: ", sURL);
	*sError = sbCopyContent(sbError);
	sbFree(sbError);
	return FAILURE;
    }
}

eRESULT structure_Load_File(char *sFilePath, DAPObject ** structure,
			    /* not used jhrg 4/3/07 ThreadSpecificStorage *pTSS, */
			    char **sError)
{
    StringBuffer *sbError;
    sbError = sbNew(NULL);
    if (zFile_LoadStructure(sFilePath, structure, sbError) == SUCCESS) {
	sbFree(sbError);
	return SUCCESS;
    } else {
	sbInsert_string(sbError, "Error loading %s: ", sFilePath);
	*sError = sbCopyContent(sbError);
	sbFree(sbError);
	return FAILURE;
    }
}

eRESULT structure_AddDAS(DAPObject * DAS, DAPObject * target,
			 char **sError)
{
    eRESULT eResult;
    StringBuffer *sbError;

    if (sError == NULL)
	return FAILURE;

    if (DAS == NULL) {
	*sError = "no DAS supplied";
	return FAILURE;
    }
    if (target == NULL) {
	*sError = "no target supplied";
	return FAILURE;
    }
    if (DAS->eStructureType != STRUCTURE_DAS) {
	*sError = "supplied DAS pointer does not point to a DAS";
	return FAILURE;
    }
    if (target->eStructureType != STRUCTURE_DDS
	&& target->eStructureType != STRUCTURE_DataDDS) {
	*sError =
	    "supplied target pointer does not point to a DataDDS or DDS";
	return FAILURE;
    }
    sbError = sbNew(NULL);
    eResult = node_AddDAS(DAS->nodeRoot, target->nodeRoot, sbError);
    *sError = sbCopyContent(sbError);
    sbFree(sbError);
    return eResult;
}

eRESULT node_AddDAS(DAPNode * nodeDAS, DAPNode * nodeTarget,
		    StringBuffer * sbError)
{
    unsigned int xSubNode_DAS;
    unsigned int xSubNode_Target;

    for (xSubNode_DAS = 1; xSubNode_DAS <= nodeDAS->iSubNodeCount;
	 	xSubNode_DAS++) {
		DAPNode *pSubNodeDAS = *(nodeDAS->nodeSub + xSubNode_DAS - 1);

		if (pSubNodeDAS->sName == NULL)
		    continue;
		for (xSubNode_Target = 1;
		     xSubNode_Target <= nodeTarget->iSubNodeCount;
		     xSubNode_Target++) {
		    DAPNode *pSubNodeTarget =
			*(nodeTarget->nodeSub + xSubNode_Target - 1);
		    if (pSubNodeTarget->sName == NULL)
			continue;
		    if (strcmp(pSubNodeDAS->sName, pSubNodeTarget->sName) == 0) {

			/* clone any attributes onto the target */
			if (pSubNodeDAS->attributes != NULL) {
			    if (pSubNodeTarget->attributes != NULL) {	/* delete any matching pre-existing attributes the target has */
				if (free_Attributes
				    (pSubNodeTarget->attributes,
				     sbError) == SUCCESS) {
				    pSubNodeTarget->attributes = NULL;
				} else {
#if 0
				    fprintf(stderr,
					    "warning, failed to free pre-existing attributes before adding new attribute set: %s\n",
					    sbError->sContent);
#endif
				    sbClear(sbError);
				}
			    }
			    pSubNodeTarget->attributes =
				clone_Attributes(pSubNodeDAS->attributes, sbError);
			    if (pSubNodeTarget->attributes == NULL) {
				sbInsert_string(sbError,
						"failed to clone attributes for node %s: ",
						pSubNodeDAS->sName);
			    }
			}

			/* recursively add any attributes in the sub nodes of the matching nodes */
			if (node_AddDAS(pSubNodeDAS, pSubNodeTarget, sbError) ==
			    FAILURE) {
			    return FAILURE;
			}

			break;		/* because we should not have multiple nodes with the same name */
		    }
		}
    }
    return SUCCESS;
}

eRESULT structure_Free(DAPObject * structure, char **sError)
{
    StringBuffer *sbError = sbNew(NULL);
    if (!imp_structure_Free(structure, sbError)) {
	if (sError != NULL)
	    *sError = sbCopyContent(sbError);
	sbFree(sbError);
	return FAILURE;
    }
    sbFree(sbError);
    return SUCCESS;
}

eRESULT buffer_Free(Buffer *
		    pBuffer /* not used jhrg 4/3/07, char **sError */ )
{
    imp_buffer_Free(pBuffer);
    return SUCCESS;
}

/** Verifies that the tree is valid. The tree must contains all required elements for its
 *  type and must be internally consistent. Should be the last step before returning a
 *  parsed tree to the user.
 */
eRESULT structure_Verify(DAPObject * pStructure, char **sError)
{
    StringBuffer *sbError = sbNew(NULL);
    if (imp_structure_Verify(pStructure, TRUE, sbError) == FAILURE) {
	if (sError != NULL) {
	    *sError = sbError->sContent;
	    sbFree(sbError);
	}
	return FAILURE;
    }
    sbFree(sbError);
    return SUCCESS;
}

eRESULT structure_Print_ASCII_toFile(DAPObject * structure, FILE * stream,
				     char **sError)
{
    StringBuffer *sbError = sbNew(NULL);
    if (imp_structure_Print_ASCII_toFile(structure, stream, sbError) ==
	FAILURE) {
	if (sError != NULL) {
	    *sError = sbError->sContent;
	    sbFree(sbError);
	}
	return FAILURE;
    }
    sbFree(sbError);
    return SUCCESS;
}

eRESULT not_implemented(char **sError)
{
    char *sErrorMessage = "not implemented";
    *sError = malloc(strlen(sErrorMessage) + 1);
    strncpy(*sError, sErrorMessage, strlen(sErrorMessage) + 1);
    return FAILURE;
}

/**************************/
/**  Internal Interface  **/
/**************************/

eRESULT imp_structure_New(DAPObject ** hStructure, StringBuffer * sbError)
{
    DAPObject *pOS = malloc(sizeof(DAPObject));
    if (pOS == NULL) {
	sbAppend(sbError, "failed to allocate memory for new structure");
	return FAILURE;
    }
    if (hStructure == NULL) {
	sbAppend(sbError, "no handle provided");
	return FAILURE;
    }
    pOS->eStructureType = 0;	/* undefined -- todo create undefined value ? */
    pOS->iEstimatedSize = 0;
    pOS->nodeRoot = NULL;
    *hStructure = pOS;
    return SUCCESS;
}

eRESULT imp_structure_Free(DAPObject * structure, StringBuffer * sbError)
{
    if (!imp_structure_Verify(structure, FALSE, sbError)) {
	sbInsert(sbError, "structure invalid: ");
	return FAILURE;
    }
    if (!imp_node_Free(structure->nodeRoot, sbError))
	return FAILURE;
    free(structure);
    return SUCCESS;
}

eRESULT imp_buffer_Free(Buffer * pBuffer
			/* not used jhrg 4/3/07  StringBuffer *sbError */ )
{
    if (pBuffer == NULL)
	return SUCCESS;
    if (pBuffer->pcharContent == NULL)
	return SUCCESS;
    free(pBuffer->pcharContent);
    free(pBuffer);
    return SUCCESS;
}

eRESULT imp_node_Free(DAPNode * pNode, StringBuffer * sbError)
{
    if (pNode == NULL)
	    return SUCCESS;

    if (!imp_structure_VerifyNode(pNode, TRUE, sbError)) {
		sbInsert(sbError, "unable to free node because it is invalid: ");
		return FAILURE;
	}

    {				/* free the sub nodes of this node first */
		int xSubNode;
		for (xSubNode = pNode->iSubNodeCount; xSubNode > 0; xSubNode--) {
		    if (imp_node_Free(*(pNode->nodeSub + (xSubNode - 1)), sbError)) {
			    *(pNode->nodeSub + (xSubNode - 1)) = NULL;
		    }
		    else {
			    sbInsert_string(sbError, "%s: ", pNode->sName);
			    return FAILURE;
		    }

		    pNode->iSubNodeCount--;
		}

		if (pNode->nodeSub != NULL)
		    free(pNode->nodeSub);
    }

    /* free any content in the node */
    if (pNode->nodeContent != NULL) {
	NodeContent_Array *pArray;
	NodeContent_Grid *pGrid;
	switch (pNode->eType) {
	case NODE_root:	/* nothing node-specific needs to be freed -- content should be null */
	case NODE_Byte:
	case NODE_Int16:
	case NODE_UInt16:
	case NODE_Int32:
	case NODE_UInt32:
	case NODE_Float32:
	case NODE_Float64:
	case NODE_Structure:
	    break;
	case NODE_String:
	case NODE_URL:
	    {
		NodeContent_String *pString =
		    (NodeContent_String *) pNode->nodeContent;
		if (pString->sContent != NULL) {
		    free(pString->sContent);
		}
		break;
	    }
	case NODE_Sequence:
	    {
		NodeContent_Sequence *pSequenceContent =
		    (NodeContent_Sequence *) pNode->nodeContent;
		DBG(printf
		    ("calling main sequence free for node %p\n",
		     (void *) pNode);)
		    if (free_SequenceContent(pSequenceContent, sbError) ==
			FAILURE) {
		    sbInsert_string(sbError, "%s: ", pNode->sName);
		    return FAILURE;
		}
		break;
	    }
	case NODE_Array:
	    pArray = (NodeContent_Array *) pNode->nodeContent;
	    if (pArray != NULL) {
		    if (free_ArrayContent(pArray, sbError) == FAILURE) {
		        sbInsert(sbError, "could not free array: ");
		        sbInsert_string(sbError, "%s: ", pNode->sName);
		        return FAILURE;
		    }
	    }
	    break;
	case NODE_Grid:
	    pGrid = (NodeContent_Grid *) pNode->nodeContent;
	    if (pGrid != NULL) {
		if (free_GridContent(pGrid, sbError) == FAILURE) {
		    sbInsert(sbError, "could not free grid: ");
		    sbInsert_string(sbError, "%s: ", pNode->sName);
		    return FAILURE;
		}
	    }
	    break;

	default:
	    sbAppend_pointer(sbError, "node at %p", pNode);
	    sbAppend_int(sbError, "had invalid type %d", pNode->eType);
	    return FAILURE;
	}

		free(pNode->nodeContent);	/* no matter what the type of the content reference,
					   it's container will be freed correctly because
					   the heap manager knows how big it is */
		pNode->nodeContent = NULL;
    }

    /* free any attributes */
    if (free_Attributes(pNode->attributes, sbError) == FAILURE) {
	    sbInsert(sbError, "error freeing attribute table: ");
	    return FAILURE;
    }
    /*use the following to debug a memory leak in the parser
    fprintf(stderr, "Free Address of sName: %lx\n", (unsigned long int)pNode->sName);*/
    free(pNode->sName);
    pNode->sName = NULL;

    free(pNode);

    return SUCCESS;
}

eRESULT free_ArrayContent(NodeContent_Array * pArrayContent,
			  StringBuffer * sbError)
{
    unsigned int xDim;
    if (pArrayContent == NULL) {
	    sbAppend(sbError, "array does not exist");
	    return FAILURE;
    }
    /*Free pArrayContent*/
    for (xDim = 1; xDim <= pArrayContent->ctDimensions; xDim++) {
	    char *sDimensionName = pArrayContent->sDimensionName1[xDim];
	    if (sDimensionName != NULL) {
	        free(sDimensionName);
	        pArrayContent->sDimensionName1[xDim] = NULL;
	    }
    }

    free(pArrayContent->sDimensionName1);
    pArrayContent->sDimensionName1 = NULL;

    free(pArrayContent->iDimensionLengthArray1);
    pArrayContent->iDimensionLengthArray1 = NULL;

    free(pArrayContent->array);
    pArrayContent->array = NULL;
    /* use the following line to depug parser memory leak
	fprintf(stderr, " Free Address of pArrayContent->sArrayName: %lx\n", (unsigned long int)pArrayContent->sArrayName);*/

    free(pArrayContent->sArrayName);
    pArrayContent->sArrayName = NULL;

    return SUCCESS;
}

eRESULT free_GridContent(NodeContent_Grid * pGrid, StringBuffer * sbError)
{
    if (pGrid->values != NULL) {
		if (free_ArrayContent(pGrid->values, sbError) == FAILURE) {
		    sbInsert(sbError, "could not free grid values: ");
		    return FAILURE;
		}

		free(pGrid->values);
    }

    {				/* free the mappings */
		unsigned int xMapping;
		for (xMapping = 1; xMapping <= pGrid->ctMappings; xMapping++) {
		    if (pGrid->maps1[xMapping] != NULL) {
				if (free_ArrayContent(pGrid->maps1[xMapping], sbError) == FAILURE) {
				    sbInsert_int(sbError, "could not free grid mapping: ",
						 xMapping);
				    return FAILURE;
				}

				free(pGrid->maps1[xMapping]);
		    }
		}

		free(pGrid->maps1);
    }

    return SUCCESS;
}

eRESULT free_SequenceContent(NodeContent_Sequence * pSequenceContent,
			     StringBuffer * sbError)
{
    unsigned int xField;
    DBG(printf("freeing sequence content at %p\n", (void *) pSequenceContent);)

	if (pSequenceContent->aeFieldType == NULL) {
	sbAppend(sbError, "no field type list exists");
	return FAILURE;
    }

    for (xField = 1; xField <= pSequenceContent->ctFields; xField++) {	/* todo ripping through a blind double pointer array like this is risky -- it is key to add internal validation codes to the OCAPI */
	unsigned int xRow;
	void *pVector;
	eNODE_TYPE eFieldType;
	eFieldType = *(pSequenceContent->aeFieldType + (xField - 1));

	/* Added this 7/11/08 */
	if (pSequenceContent->aFieldValue == NULL)
		continue;

	/* free any content in the node */
	pVector = *(pSequenceContent->aFieldValue + (xField - 1));
	DBG(printf("vector at %p\n", pVector);)

	switch (eFieldType) {
	case NODE_Byte:;
	case NODE_Int16:;
	case NODE_UInt16:;
	case NODE_Int32:;
	case NODE_UInt32:;
	case NODE_Float32:;
	case NODE_Float64:;
	    break;
	case NODE_String:;
	/*free the url*/
	case NODE_URL:;
	    {
		NodeContent_String **string_vector =
		    (NodeContent_String **) pVector;
		for (xRow = 1; xRow <= pSequenceContent->ctRows; xRow++) {
		    NodeContent_String *pStringContent =
			*(string_vector + xRow - 1);
		    free(pStringContent->sContent);
		    free(pStringContent);
		    *(string_vector + xRow -1) = NULL;
		}
	    }
	    break;
	/*free the array*/
	case NODE_Array:;
	    {
		NodeContent_Array **array_vector =
		    (NodeContent_Array **) pVector;
		for (xRow = 1; xRow <= pSequenceContent->ctRows; xRow++) {
		    NodeContent_Array *pArrayContent =
			*(array_vector + xRow - 1);
		    if (free_ArrayContent(pArrayContent, sbError) ==
			FAILURE) {
			sbInsert_int(sbError, "row %d: ", xRow);
			sbInsert_int(sbError,
				     "unable to free array field %d, ",
				     xField);
			return FAILURE;
		    }
		    free(pArrayContent);
		    *(array_vector + xRow - 1) = NULL;
		}
	    }
	    break;
	/*free the grid*/
	case NODE_Grid:;
	    {
		NodeContent_Grid **grid_vector =
		    (NodeContent_Grid **) pVector;
		for (xRow = 1; xRow <= pSequenceContent->ctRows; xRow++) {
		    NodeContent_Grid *pGridContent =
			*(grid_vector + xRow - 1);
		    if (free_GridContent(pGridContent, sbError) == FAILURE) {
			sbInsert_int(sbError, "row %d: ", xRow);
			sbInsert_int(sbError,
				     "unable to free grid field %d, ",
				     xField);
			return FAILURE;
		    }
		    free(pGridContent);
		    *(grid_vector + xRow - 1) = NULL;
		}
	    }
	    break;
	/*free the sequence*/
	case NODE_Sequence:;
	    {
		NodeContent_Sequence **sequence_vector =
		    (NodeContent_Sequence **) pVector;
		for (xRow = 1; xRow <= pSequenceContent->ctRows; xRow++) {
		    NodeContent_Sequence *pSequenceContent =
			*(sequence_vector + xRow - 1);
		    DBG(printf("calling internal sequence free for row %i in vector %p\n",
			   xRow, pVector);)
			if (free_SequenceContent(pSequenceContent, sbError)
			    == FAILURE) {
			sbInsert_int(sbError, "row %d: ", xRow);
			sbInsert_int(sbError,
				     "unable to free sequence field %d, ",
				     xField);
			return FAILURE;
		    }
		    free(pSequenceContent);
		    *(sequence_vector + xRow - 1) = NULL;
		}
	    }
	    break;
	/*free the structure*/
	case NODE_Structure:;
	    {
		NodeContent_Structure **structure_vector =
		    (NodeContent_Structure **) pVector;
		for (xRow = 1; xRow <= pSequenceContent->ctRows; xRow++) {
		    NodeContent_Structure *pStructureContent =
			*(structure_vector + xRow - 1);
		    DBG(printf
			("calling internal structure free for row %i in vector %p\n",
			 xRow, pVector);)
			if (free_StructureContent(pStructureContent, sbError) == FAILURE) {
			sbInsert_int(sbError, "row %d: ", xRow);
			sbInsert_int(sbError,
				     "unable to free structure field %d, ",
				     xField);
			return FAILURE;
		    }
		    free(pStructureContent);
		    *(structure_vector + xRow - 1) = NULL;
		}
	    }
	    break;
	/*free the root*/
	case NODE_root:;
	default:
	    sbAppend_int(sbError,
			 "invalid field type [root] in sequence %d",
			 eFieldType);
	    return FAILURE;
	}
	free(pVector);		/* this is possible because in the case of primitive values,
	the entire field is allocated as a single, contiguous array on the heap */
	DBG(printf("freeing vector at %p\n", pVector);)
    }

    free(pSequenceContent->aeFieldType);
    /* Added test 7/11/08 */
    if (pSequenceContent->aFieldValue != NULL)
    	free(pSequenceContent->aFieldValue);
    return SUCCESS;
}

eRESULT free_StructureContent(NodeContent_Structure * pStructureContent,
			      StringBuffer * sbError)
{
    unsigned int xMember;
    DBG(printf
	("freeing structure content at %p\n", (void *) pStructureContent);)
	if (pStructureContent->array_eNodeType == NULL) {
	sbAppend(sbError, "no field type list exists");
	return FAILURE;
    }
    if (pStructureContent->array_NodeContent == NULL) {
	sbAppend(sbError, "no field value list exists");
	return FAILURE;
    }

    for (xMember = 1; xMember <= pStructureContent->ctMembers; xMember++) {
    	void *pNodeContent;
    	eNODE_TYPE eFieldType;
    	eFieldType = *(pStructureContent->array_eNodeType + (xMember - 1));
    	pNodeContent = *(pStructureContent->array_NodeContent + (xMember - 1));
    /* free any content in the node */
	if (pNodeContent != NULL) {
	    NodeContent_Array *pArray;
	    NodeContent_Grid *pGrid;
	    switch (eFieldType) {
	    case NODE_root:	/* nothing node-specific needs to be freed -- content should be null */
	    case NODE_Byte:
	    case NODE_Int16:
	    case NODE_UInt16:
	    case NODE_Int32:
	    case NODE_UInt32:
	    case NODE_Float32:
	    case NODE_Float64:
	    case NODE_Structure:
		break;
	    case NODE_String:;
	    /*free the url*/
	    case NODE_URL:;
		{
		    NodeContent_String *pString =
			(NodeContent_String *) pNodeContent;
		    free(pString->sContent);
		}
		break;
		/*free the sequence*/
	    case NODE_Sequence:;
		{
		    NodeContent_Sequence *pSequenceContent =
			(NodeContent_Sequence *) pNodeContent;
		    DBG(printf
			("calling sequence free for sequence content %p in structure content node %p\n",
			 (void *) pSequenceContent,
			 (void *) pStructureContent);)
			if (free_SequenceContent(pSequenceContent, sbError)
			    == FAILURE) {
			sbInsert_int(sbError, "structure member %i: ",
				     xMember);
			return FAILURE;
		    }
		}
		break;
		/*free the array*/
	    case NODE_Array:;
		{
		    pArray = (NodeContent_Array *) pNodeContent;
		    if (pArray != NULL) {
			if (free_ArrayContent(pArray, sbError) == FAILURE) {
			    sbInsert(sbError,
				     "could not free array in structure: ");
			    sbInsert_int(sbError, "member %i: ", xMember);
			    return FAILURE;
			}
		    }
		}
		break;
		/*free the grid*/
	    case NODE_Grid:;
		{
		    pGrid = (NodeContent_Grid *) pNodeContent;
		    if (pGrid != NULL) {
			if (free_GridContent(pGrid, sbError) == FAILURE) {
			    sbInsert(sbError, "could not free grid: ");
			    sbInsert_int(sbError, "member %i: ", xMember);
			    return FAILURE;
			}
		    }
		}
		break;
	    default:
		sbAppend_int(sbError, "structure member %i", xMember);
		sbAppend_int(sbError, " had invalid type %d", eFieldType);
		return FAILURE;
	    }
	    free(pNodeContent);	/* no matter what the type of the content reference,
				   it's container will be freed correctly because
				   the heap manager knows how big it is */
	    pNodeContent = NULL;
	}
    }
    free(pStructureContent->array_eNodeType);
    free(pStructureContent->array_NodeContent);
    return SUCCESS;
}

eRESULT free_Attributes(DAPAttributeTable * pTable, StringBuffer * sbError)
{
    unsigned int xAttribute;
    unsigned int ctAttributes;
    if (pTable == NULL)
	return SUCCESS;
    ctAttributes = pTable->ctAttributes;
    /* This seems wrong. Why 10,000? There's no hard limit in the DASP 2
       spec. */
    if (ctAttributes > 10000) {
	sbAppend_int(sbError,
		     "validation error, attribute count is excessive %i",
		     ctAttributes);
    }
    for (xAttribute = ctAttributes; xAttribute > 0; xAttribute--) {
	char *sName;
	DAPAttribute *attribute =
	    *(pTable->pAttributeList + xAttribute - 1);
	if (attribute != NULL) {
	    unsigned int xValue;
	    if (attribute->sName != NULL) {
		free(attribute->sName);
		attribute->sName = NULL;
	    }
	    for (xValue = 1; xValue <= attribute->ctValues; xValue++) {
		void *pValue = *(attribute->pValueList + xValue - 1);
		if (pValue != NULL) {
		    free(pValue);
		    *(attribute->pValueList + xValue - 1) = NULL;
		}
	    }
	    free(attribute->pValueList);
	}
	free(attribute);
	sName = *(pTable->pNameList + xAttribute - 1);
	if (sName != NULL) {
	    free(sName);
	    *(pTable->pNameList + xAttribute - 1) = NULL;
	}
    }
    free(pTable->pNameList);
    free(pTable->pAttributeList);
    free(pTable);
    return SUCCESS;
}

/* returns NULL on failure */
/*Copy the attributes from the DAPAttributeTable onto the target.
 *This is being called from node_AddDAS*/
DAPAttributeTable *clone_Attributes(DAPAttributeTable * attributes,
				    StringBuffer * sbError)
{
    DAPAttributeTable *atNew;
    unsigned int xAttribute;
    unsigned int xValue;
    if (attributes == NULL) {
	sbAppend(sbError, "no attribute table supplied");
	return NULL;
    }
    atNew = attribute_NewTable(sbError);
    if (atNew == NULL)
	return NULL;
    if (attributes->ctAttributes <= 0)
	return atNew;
    if (attributes->pNameList == NULL) {
	sbAppend(sbError,
		 "invalid attribute table, name list does not exist");
	return NULL;
    }
    if (attributes->pAttributeList == NULL) {
	sbAppend(sbError,
		 "invalid attribute table, attribute list does not exist");
	return NULL;
    }
    for (xAttribute = 1; xAttribute <= attributes->ctAttributes;
	 xAttribute++) {
	DAPAttribute *attribute;
	DAPAttribute *attribNew;
	char *sName;
	char *sClonedName;
	attribute = *(attributes->pAttributeList + xAttribute - 1);
	sName = *(attributes->pNameList + xAttribute - 1);
	attribNew = attribute_New(attribute->eType, sName, sbError);
	if (attribNew == NULL)
	    return NULL;
	for (xValue = 1; xValue <= attribute->ctValues; xValue++) {
	    if (attribute_AddValue
		(attribNew, *(attribute->pValueList + xValue - 1),
		 sbError) == FAILURE) {
		sbInsert_int(sbError, "failed to add value %i: ", xValue);
		sbInsert_string(sbError, "in attribute %s ", sName);
		return NULL;
	    }
	}
	/*Copy the name from the attribute table being cloned*/
	sClonedName = sCloneString(sName);
	printf("cloned name %s at %p\n", sName, sName);
	if (attribute_Add(atNew, sClonedName, attribNew, sbError) ==
	    FAILURE) {
	    sbInsert_string(sbError, "failed to add attribute named %s: ",
			    sName);
	    return NULL;
	}
    }
    return atNew;
}

/** Verifies that the tree is valid. The tree must contains all required elements for its
 *  type and must be internally consistent. Should be the last step before returning a
 *  parsed tree to the user.
 */
eRESULT imp_structure_Verify(DAPObject * pStructure, int zRequireParent,
			     StringBuffer * sbError)
{
    if (pStructure == NULL) {
	sbAppend(sbError, "OPeNDAP structure does not exist");
	return FAILURE;
    }
    if (pStructure->nodeRoot == NULL) {
	sbAppend(sbError, "no root node");
	return FAILURE;
    }
    if (pStructure->nodeRoot->eType != NODE_root) {
	sbAppend(sbError, "root node has wrong type");
	return FAILURE;
    }
    return imp_structure_VerifyNode(pStructure->nodeRoot, zRequireParent,
				    sbError);
}

/* If the node is unattached then no parent should be required, but if the node is in a complete
 * structure then a parent should be required. This is called from
 * imp_node_Free, imp_structure_Verify*/
eRESULT imp_structure_VerifyNode(DAPNode * pNode, int zRequireParent,
				 StringBuffer * sbError)
{
    unsigned int xSubNode;
    if (pNode == NULL) {
	sbAppend(sbError, "node does not exist");
	return FAILURE;
    }

    /*Start verifying the node cases*/
    switch (pNode->eType) {
    case NODE_root:;
    /*Check to see if node has a parent node*/
	if (pNode->nodeSuper != NULL) {
	    sbAppend(sbError, "root node has parent pointer");
	    return FAILURE;
	}
	/*Check to see if node has a name*/
	if (pNode->sName == NULL) {
	    printf("root node at %p\n", (void *) pNode);
	    sbAppend(sbError, "root node has no name");
	    return FAILURE;
	}
	/* recursively verify structure */
	for (xSubNode = 1; xSubNode <= pNode->iSubNodeCount; xSubNode++) {
	    if (imp_structure_VerifyNode(*(pNode->nodeSub + xSubNode - 1), zRequireParent,
		 sbError) == FAILURE)
		return FAILURE;
	}
	return SUCCESS;

    case NODE_Byte:;
    case NODE_Int16:;
    case NODE_UInt16:;
    case NODE_Int32:;
    case NODE_UInt32:;
    case NODE_Float32:;
    case NODE_Float64:;

	if (pNode->sName == NULL) {
	    sbAppend(sbError, "variable has no name");
	    return FAILURE;
	}
	if (pNode->iSubNodeCount != 0) {
	    sbAppend_int(sbError,
			 "invalid sub node count for a primitive type %d",
			 pNode->iSubNodeCount);
	    return FAILURE;
	}
	return SUCCESS;

    case NODE_String:;
    case NODE_URL:;
	{
	    return SUCCESS;
	}
    case NODE_Sequence:;
	if (imp_structure_VerifySequence
	    (pNode, (NodeContent_Sequence *) pNode->nodeContent,
	     zRequireParent, sbError) == FAILURE) {
	    sbInsert_string(sbError, "invalid sequence named \"%s\": ",
			    pNode->sName);
	    return FAILURE;
	}
	break;
    case NODE_Array:;
	if (imp_structure_VerifyArray
	    ((NodeContent_Array *) pNode->nodeContent,
	     sbError) == FAILURE) {
	    sbInsert_string(sbError, "invalid array named \"%s\": ",
			    pNode->sName);
	    return FAILURE;
	}
	break;
    case NODE_Grid:;
	if (imp_structure_VerifyGrid
	    ((NodeContent_Grid *) pNode->nodeContent,
	     sbError) == FAILURE) {
	    sbInsert_string(sbError, "invalid grid named \"%s\": ",
			    pNode->sName);
	    return FAILURE;
	}
	break;
    case NODE_Structure:;
	{
	    NodeContent_Structure *pNode_structure =
		(NodeContent_Structure *) pNode->nodeContent;
	    if (pNode_structure != NULL) {
		sbInsert_string(sbError,
				"structure node has content (no content allowed in structure nodes): %s",
				pNode->sName);
		return FAILURE;
	    }
	    /* recursively verify structure's children */
	    for (xSubNode = 1; xSubNode <= pNode->iSubNodeCount; xSubNode++) {
		if (imp_structure_VerifyNode
		    (*(pNode->nodeSub + xSubNode - 1), zRequireParent,
		     sbError) == FAILURE)
		    return FAILURE;
	    }
	}
	break;
    default:;
	sbAppend_int(sbError, "unknown type [%d] ", pNode->eType);
	sbAppend_string(sbError, "for node named \"%s\" in DDS",
			pNode->sName);
	return FAILURE;
    }
    return SUCCESS;
}
/*This is used to verify the array is valid, case NODE_ARRAY
 *This is called in imp_structure_VerifyNode,imp_structure_VerifySequence,
 *imp_structure_VerifyStructure.*/
eRESULT imp_structure_VerifyArray(NodeContent_Array * pContent_array,
				  StringBuffer * sbError)
{
    eARRAY_TYPE type;
    if (pContent_array == NULL) {
	sbAppend(sbError, "no content passed to verify array");
	return FAILURE;
    }
    type = pContent_array->eType;
    if (!(type == ARRAY_Byte ||
	  type == ARRAY_Int16 ||
	  type == ARRAY_UInt16 ||
	  type == ARRAY_Int32 ||
	  type == ARRAY_UInt32 ||
	  type == ARRAY_Float32 ||
	  type == ARRAY_Float64 ||
	  type == ARRAY_String ||
	  type == ARRAY_URL || type == ARRAY_Structure)) {
	sbAppend_int(sbError, "invalid array type %d", type);
	return FAILURE;
    }
    if (pContent_array->ctDimensions < 1) {
	sbAppend(sbError, "array has 0 dimensions");
	return FAILURE;
    }
    return SUCCESS;
}
/*This is used to verify the grid is valid, case NODE_GRID.
 *This is called in imp_structure_VerifyNode, imp_structure_VerifyStructure,
 *imp_structure_VerifySequence*/
eRESULT imp_structure_VerifyGrid(NodeContent_Grid * pContent_grid,
				 StringBuffer * sbError)
{
    NodeContent_Array *pContent_values;
    NodeContent_Array **pContent_maps;
    unsigned int ctMappingsSupplied;
    unsigned int xMapping;

    if (pContent_grid == NULL) {
	sbAppend(sbError, "grid does not exist");
	return FAILURE;
    }
    pContent_values = pContent_grid->values;
    pContent_maps = pContent_grid->maps1;

    if (pContent_values == NULL) {
	sbAppend(sbError, "no values");
	return FAILURE;
    }

    if (pContent_maps == NULL) {
	sbAppend(sbError, "no mappings");
	return FAILURE;
    }

    ctMappingsSupplied = pContent_grid->ctMappings;

    if (ctMappingsSupplied == pContent_grid->values->ctDimensions) {
	for (xMapping = 1; xMapping <= ctMappingsSupplied; xMapping++) {
	    NodeContent_Array *pContent_MapArray = pContent_maps[xMapping];

	    if (pContent_MapArray == NULL)
		break;
	    /*checking for the length of the dimension*/
	    if (pContent_MapArray->iDimensionLengthArray1 == NULL) {
		sbAppend_int(sbError,
			     "mapping array %d missing dimension lengths",
			     xMapping);
		return FAILURE;
	    }
	    /*check to see if mapping array has one dimension*/
	    if (pContent_MapArray->ctDimensions != 1) {
		sbAppend_int(sbError,
			     "mapping array %d does not have exactly one dimension",
			     xMapping);
		return FAILURE;
	    }
	    /*check to see if mapping array and value array have equal dimensions*/
	    if (pContent_MapArray->iDimensionLengthArray1[1] !=
		pContent_values->iDimensionLengthArray1[xMapping]) {
		sbAppend_int(sbError,
			     "mapping array %d does not match the corresponding dimension in the value array",
			     xMapping);
		return FAILURE;
	    }
	}
    }
    else {
    	/* allow a multi-dimensional mapping if it matches the dimensions of the values */
    	if (ctMappingsSupplied == 1) {
    		NodeContent_Array *pContent_MapArray = pContent_maps[1];

			if (pContent_MapArray == NULL) {
				sbAppend(sbError, "mapping array content missing");
				return FAILURE;
			}
			/*checking if mapping array and value array have the same number of dimensions
			 * for a multi-dimensional mapping*/
			if (pContent_MapArray->ctDimensions ==
				pContent_grid->values->ctDimensions) {
				unsigned int xDimension = 1;
				for (; xDimension <= pContent_grid->values->ctDimensions;
					xDimension++) {
					/*recursively checking whether or not the
					 * mapping array dimensions match the value array dimensions*/
					if (pContent_MapArray->iDimensionLengthArray1[xDimension] !=
					pContent_values->iDimensionLengthArray1[xDimension]) {
						sbAppend_int(sbError,
								"mapping array dimension %d has a different length than the corresponding dimension in the value array",
				     xDimension);
					return FAILURE;
					}
				}
			}
			else {
	    	sbAppend_int(sbError,
			     "invalid grid, count of mapping dimensions (%d) differs from number of dimensions in value array",
			     pContent_MapArray->ctDimensions);
			return FAILURE;
			}
    	}
    	else {
	    sbAppend(sbError,
		     "invalid grid, number of mappings differs from number of dimensions in value array");
	    return FAILURE;
	}
    }
    return SUCCESS;
}

/*This is called in imp_structure_VerifyNode, imp_structure_VerifyStructure, imp_structure_VerifySequence=>case NODE_Sequence*/
eRESULT imp_structure_VerifySequence(DAPNode * pSequenceNode, NodeContent_Sequence * pNodeContent_sequence,
				     int zRequireParent,StringBuffer * sbError)
{
    unsigned int ctSubNodes;
    unsigned int ctFields;
    unsigned int ctRows;
    unsigned int xField;
    unsigned int xRow;
    void *pContent_vector;

    DBG(printf("verifying sequence at %p with content at %p\n",
       (void *) pSequenceNode, (void *) pNodeContent_sequence);)

	if (pSequenceNode == NULL) {
	sbAppend(sbError, "no sequence node supplied");
	return FAILURE;
    }

    if (zRequireParent) {
    	/*verifing that the sequence has a parent node*/
    	if (pSequenceNode->nodeSuper == NULL) {
    		sbAppend(sbError, "sequence node has no parent");
    		return FAILURE;
			}

		/* content checking can only be done for an attached node */
		if (pNodeContent_sequence == NULL) {
			if (zIsInsideSequence(pSequenceNode)) {
			/* correct, no content for nested sequences */
			}
			else {
				sbAppend(sbError,"sequence is top-level, but no sequence content is supplied");
				return FAILURE;
			}
		}
			else {

			}
		}

    ctRows = pNodeContent_sequence->ctRows;
    ctFields = pNodeContent_sequence->ctFields;

    /* if a newly initialized node is verified this will happen TODO use different verification for raw nodes */
    if (ctFields == 0) {
	return SUCCESS;
    }
    /*verifying that sub nodes equal the number of fields*/
    ctSubNodes = pSequenceNode->iSubNodeCount;
    if (ctSubNodes != ctFields) {
    	DBG(printf("sequence at %p failed verification\n",
    		 (void *) pSequenceNode);)
	    sbAppend_int(sbError, "the number of subnodes in the sequence %d ",
			 ctSubNodes);
		sbAppend_int(sbError, "did not match the number of fields in its content %d",
		     ctFields);
	return FAILURE;
    }
    /* verify that the field types match, sub nodes match the field type */
    for (xField = 1; xField <= ctFields; xField++) {
	DAPNode *nodeSub = *(pSequenceNode->nodeSub + xField - 1);
		if (nodeSub->eType !=
			*(pNodeContent_sequence->aeFieldType + xField - 1)) {
			sbAppend_int(sbError, "the type of subnode %d which was ",
					ctSubNodes);
			sbAppend_int(sbError, "%d did not match ", nodeSub->eType);
			sbAppend_int(sbError, "the type of the content field which was %d",
			 *(pNodeContent_sequence->aeFieldType + xField - 1));
	    return FAILURE;
		}
    }
    /* check to see if there is data: no data -- for example could be a DDS */
    if (ctRows == 0 || ctFields == 0)
    return SUCCESS;

    if (pNodeContent_sequence->aFieldValue == NULL) {
	sbAppend(sbError, "sequence content array is missing");
	return FAILURE;
    }

    for (xField = 1; xField <= ctFields; xField++) {
		eNODE_TYPE eFieldType = *(pNodeContent_sequence->aeFieldType + xField - 1);
		/*validate field types*/
		if (eFieldType == NODE_root) {
			sbAppend_int(sbError,
				 "field %d contains a root node, invalid in a sequence",
				 xField);
			return FAILURE;
		}
		if (eFieldType == NODE_Byte ||
			eFieldType == NODE_Int16 ||
			eFieldType == NODE_UInt16 ||
			eFieldType == NODE_Int32 ||
			eFieldType == NODE_UInt32 ||
			eFieldType == NODE_Float32 || eFieldType == NODE_Float64) {
			continue;		/* no way to validate primitive value further */
		}
		/*verify the value of the vector*/
		pContent_vector = *(pNodeContent_sequence->aFieldValue + xField - 1);
		for (xRow = 1; xRow <= ctRows; xRow++) {
			switch (eFieldType) {
			case NODE_String:;
			case NODE_URL:; // Forgotten in the original code
			{
				NodeContent_String *pNode_String = *((NodeContent_String **) pContent_vector + xRow - 1);
				/* null values are allowed for now for strings */
				if (pNode_String == NULL) {
				continue;
				}
				break;
			}
			case NODE_Sequence:;
			/* todo add field/row numbers to error handling */
			{
				DAPNode *nodeSub = *(pSequenceNode->nodeSub + xField - 1);
				NodeContent_Sequence *pSequenceContent = *((NodeContent_Sequence **) pContent_vector +
						xRow - 1);
				DBG(printf("verifying sequence content vector at %p with row %u at %p\n",
						(void *) pContent_vector, xRow,(void *) pSequenceContent);)
				if (imp_structure_VerifySequence(nodeSub, pSequenceContent, zRequireParent,
					sbError) == FAILURE) {
					sbInsert_int(sbError, ", row %d: ", xRow);
					sbInsert_int(sbError, "bad sequence in field %d",
						 xField);
				return FAILURE;
				}
				break;
			}
			case NODE_Array:;
			{
				NodeContent_Array *pNode_Array = *((NodeContent_Array **) pContent_vector + xRow -
				  1);
				if (imp_structure_VerifyArray(pNode_Array, sbError) == FAILURE) {
					sbInsert_int(sbError, ", row %d: ", xRow);
					sbInsert_int(sbError, "bad array in field %d",
						 xField);
				return FAILURE;
				}
				break;
			}
			case NODE_Grid:;
			{
				NodeContent_Grid *pNode_Grid = *((NodeContent_Grid **) pContent_vector + xRow -
				  1);
				if (imp_structure_VerifyGrid(pNode_Grid, sbError) == FAILURE) {
					sbInsert_int(sbError, ", row %d: ", xRow);
					sbInsert_int(sbError, "bad grid in field %d",
						 xField);
				return FAILURE;
				}
				break;
			}
			case NODE_Structure:;
			{
				DAPNode *nodeSub = *(pSequenceNode->nodeSub + xField - 1);
				NodeContent_Structure *pNode_Structure = *((NodeContent_Structure **) pContent_vector +
				  xRow - 1);
				if (imp_structure_VerifyStructure(nodeSub, pNode_Structure, zRequireParent,
					sbError) == FAILURE) {
					sbInsert_int(sbError, ", row %d: ", xRow);
					sbInsert_int(sbError, "bad structure in field %d", xField);
				return FAILURE;
				}
				break;
			}
			default:;
			sbAppend_int(sbError, "field %d is of unknown type:", xField);
			sbAppend_int(sbError, " %d", eFieldType);
			return FAILURE;
			}
		}
    }
    return SUCCESS;
}

int zIsInsideSequence(DAPNode * node)
{
    if (node->nodeSuper == NULL)
	return 0;		/* false */
    if (node->nodeSuper->eType == NODE_Sequence)
	return 1;		/* true */
    return zIsInsideSequence(node->nodeSuper);
}

eRESULT imp_structure_VerifyStructure(DAPNode * nodeStructure,
				      NodeContent_Structure *
				      pNodeContent_structure,
				      int zRequireParent,
				      StringBuffer * sbError)
{
    DAPNode *pMemberNode;
    unsigned int xMember;
    if (DEBUG)
	printf("verifying structure at %p, with content at %p\n",
	       (void *) nodeStructure, (void *) pNodeContent_structure);
    if (pNodeContent_structure == NULL)
	return SUCCESS;
    if (nodeStructure == NULL) {
	sbAppend(sbError, "structure node was not supplied, was null");
	return FAILURE;
    }
    if (nodeStructure->eType != NODE_Structure) {
	sbAppend(sbError, "supplied node was not of type Structure");
	return FAILURE;
    }
    for (xMember = 1; xMember <= pNodeContent_structure->ctMembers;
	 xMember++) {
	eNODE_TYPE eFieldType =
	    *(pNodeContent_structure->array_eNodeType + xMember - 1);
	void *pContent =
	    *(pNodeContent_structure->array_NodeContent + xMember - 1);

	pMemberNode = *(nodeStructure->nodeSub + xMember - 1);
	switch (eFieldType) {
	case NODE_root:;
	    sbAppend_int(sbError, "member %d contains a root node",
			 xMember);
	    return FAILURE;
	case NODE_Byte:;
	case NODE_Int16:;
	case NODE_UInt16:;
	case NODE_Int32:;
	case NODE_UInt32:;
	case NODE_Float32:;
	case NODE_Float64:;
	    /* unable to validate further -- could be null pointer ? */
	    return SUCCESS;
	case NODE_String:;
	case NODE_URL:;
	    {
		NodeContent_String *pNode_String =
		    (NodeContent_String *) pContent;
		if (pNode_String == NULL) {
		    continue;	/* null values are allowed for now for strings */
		}
		return SUCCESS;
	    }
	case NODE_Sequence:;	/* todo add field/row numbers to error handling */
	    {
		NodeContent_Sequence *pNode_Sequence =
		    (NodeContent_Sequence *) pContent;
		return imp_structure_VerifySequence(pMemberNode,
						    pNode_Sequence,
						    zRequireParent,
						    sbError);
	    }
	case NODE_Array:;
	    {
		NodeContent_Array *pNode_ArrayContent =
		    (NodeContent_Array *) pContent;
		if (imp_structure_VerifyArray(pNode_ArrayContent, sbError)
		    == FAILURE) {
		    sbInsert_int(sbError, "array structure member %d: ",
				 xMember);
		    return FAILURE;
		}
		break;
	    }
	case NODE_Grid:;
	    {
		NodeContent_Grid *pNode_Grid =
		    (NodeContent_Grid *) pContent;
		return imp_structure_VerifyGrid(pNode_Grid, sbError);
	    }
	case NODE_Structure:;
	    {
		NodeContent_Structure *pNode_Structure =
		    (NodeContent_Structure *) pContent;
		return imp_structure_VerifyStructure(pMemberNode,
						     pNode_Structure,
						     zRequireParent,
						     sbError);
	    }
	default:;
	    sbAppend_int(sbError, "member %d is of unknown type", xMember);
	    return FAILURE;
	}
    }
    return SUCCESS;
}

eRESULT imp_structure_Print_ASCII_toFile(DAPObject * structure,
					 FILE * stream,
					 StringBuffer * sbError)
{
    /* if( !imp_structure_Verify( structure, sbError ) ){ sbInsert( sbError, "DAPObject invalid: "); return FAILURE; } */
    fputs(render_STRUCTURE_TYPE(structure->eStructureType), stream);
    if (structure->iEstimatedSize > 0) {
	fprintf(stream, " size: %d\n", structure->iEstimatedSize);
    } else {
	fputs(" size: ?\n", stream);
    }
    return print_Node_toFile(structure->nodeRoot, 0, stream, sbError);
}

eRESULT print_Node_toFile(DAPNode * node, int iIndent, FILE * stream,
			  StringBuffer * sbError)
{
    if (node == NULL) {
	indent(iIndent, stream);
	fputs("{null}\n", stream);
	return SUCCESS;
    }
    return print_NodeContent_toFile(node, NULL, iIndent, stream, sbError);
}
/*this is called in print_Node_toFile, print_Sequence_toFile, print_Structure_toFile*/
eRESULT print_NodeContent_toFile(DAPNode * node, void *pNestedContent,
				 int iIndent, FILE * stream,  StringBuffer * sbError)
{
    unsigned int xSubNode;
    NodeContent_Sequence *pSequence;
    NodeContent_Array *pArray;
    NodeContent_Grid *pGrid;
    NodeContent_String *pString;
    void *pNodeContent;

    if (node == NULL) {
    	indent(iIndent, stream);
    	fputs("{null}\n", stream);
    	return SUCCESS;
    }

/* fprintf( stream, "printing node %s with content %p and nested content %p\n", node->sName, node->nodeContent, pNestedContent ); */

    if (pNestedContent == NULL) {
    	pNodeContent = node->nodeContent;
    }
    else {
    	pNodeContent = pNestedContent;
    }

    /* title information */
    indent(iIndent, stream);
    fputs("{ ", stream);
	if (node->sName == NULL)
		fputs("[no name]", stream);
    else
    	fputs(node->sName, stream);

    /* type information */
    fputs("  * ", stream);
    fputs(render_NODE_TYPE(node->eType), stream);

    /* primitive content goes right on the title line along with type */
    if (pNodeContent == NULL) {
    	fputs(" ", stream);
    }
    else {
	switch (node->eType) {
	case NODE_Byte:
	    fprintf(stream, "  value: %u ", *(DBYTE *) pNodeContent); /* DBYTE is an unsigned char */
	    break;
	case NODE_Int16:
	    fprintf(stream, "  value: %d ", *(DINT16 *) pNodeContent);
	    break;
	case NODE_UInt16:
	    fprintf(stream, "  value: %u ", *(DUINT16 *) pNodeContent);
	    break;
	case NODE_Int32:
	    fprintf(stream, "  value: %d ", *(DINT32 *) pNodeContent);
	    break;
	case NODE_UInt32:
	    fprintf(stream, "  value: %u ", *(DUINT32 *) pNodeContent);
	    break;
	case NODE_Float32:
	    fprintf(stream, "  value: %f ", *(DFLOAT32 *) pNodeContent);
	    break;
	case NODE_Float64:
	    fprintf(stream, "  value: %f ", *(DFLOAT64 *) pNodeContent);
	    break;
	    /* It turns out that fprintf handles a null value for %s
	     * differently on different operating systems. Provide
	     * consistent output for NULL.*/
	case NODE_String:
	case NODE_URL:
	    pString = (NodeContent_String*) pNodeContent;
	    fprintf(stream, "  value: %s ",
		    pString->sContent != NULL ? pString->sContent : "(null)");
	    break;

	    /* other types have only internal content that is put out in the sub routines */
	case NODE_root:
	case NODE_Structure:
	case NODE_Sequence:
	case NODE_Array:
	case NODE_Grid:
	default:
	    break;
	}
    }

    if (node->nEstimatedSize > 0) {
    	fprintf(stream, "  size: %lu", node->nEstimatedSize);
    }

    /* attributes */
    if (print_Attributes_toFile(node->attributes, iIndent + 1, stream)
	    == FAILURE) {
    }

    /* content */
    if (pNodeContent != NULL) {
	switch (node->eType) {
	case NODE_Sequence:
	    ;
	    pSequence = (NodeContent_Sequence *) pNodeContent;
	    if (print_Sequence_toFile(node, pSequence, iIndent + 1, stream,
		    sbError) == FAILURE) {
		fprintf(stream, "[error in sequence: %s]\n", sbError->sContent);
		sbClear(sbError);
	    }
	    indent(iIndent, stream);
	    break;
	case NODE_Array:
	    ;
	    pArray = (NodeContent_Array *) pNodeContent;
	    if (print_Array_toFile(pArray, iIndent + 1, stream, sbError)
		    == FAILURE) {
		fprintf(stream, "[error in array: %s]\n", sbError->sContent);
		sbClear(sbError);
	    }
	    break;
	case NODE_Grid:
	    ;
	    pGrid = (NodeContent_Grid *) pNodeContent;
	    if (print_Grid_toFile(pGrid, iIndent + 1, stream, sbError)
		    == FAILURE) {
		fprintf(stream, "[error in grid: %s]\n", sbError->sContent);
		sbClear(sbError);
	    }
	    indent(iIndent, stream);
	    break;
	default:
	    /* the other content types are printed above */
	    break;
	}
    }
    /* if there is no content for the sequence then just print its structure */
    else if (node->eType == NODE_Sequence) {
	fputs("\n", stream);
	for (xSubNode = 1; xSubNode <= node->iSubNodeCount; xSubNode++) {
	    if (!print_NodeContent_toFile(*(node->nodeSub + (xSubNode - 1)),
		    pNodeContent, iIndent + 1, stream, sbError)) {
		indent(iIndent, stream);
		fprintf(
			stream,
			"??? Error printing sequence structure sub-node %i: %s\n",
			xSubNode, sbError->sContent);
		return FAILURE;
	    }
	}
	indent(iIndent, stream);
    }

    /* recursively process any sub nodes of structures */
    if (node->eType == NODE_root || node->eType == NODE_Structure) {
	if (node->eType == NODE_root) {
	    if (pNestedContent != NULL) {
		sbAppend(sbError, "root node has content");
		return FAILURE;
	    }
	    fputs("\n", stream);
	    for (xSubNode = 1; xSubNode <= node->iSubNodeCount; xSubNode++) {
		DAPNode *pSubNode;
		pSubNode = *(node->nodeSub + (xSubNode - 1));
		if (!print_NodeContent_toFile(pSubNode, NULL, iIndent + 1,
			stream, sbError)) {
		    indent(iIndent, stream);
		    fprintf(stream,
			    "[Error printing root's sub-node %i: %s]\n",
			    xSubNode, sbError->sContent);
		    return FAILURE;
		}
	    }
	} else {
	    if (pNestedContent == NULL) {
		for (xSubNode = 1; xSubNode <= node->iSubNodeCount; xSubNode++) {
		    DAPNode *pSubNode;
		    pSubNode = *(node->nodeSub + (xSubNode - 1));
		    if (!print_NodeContent_toFile(pSubNode, NULL, iIndent + 1,
			    stream, sbError)) {
			indent(iIndent, stream);
			fprintf(stream,
				"[Error printing structure sub-node %i: %s]\n",
				xSubNode, sbError->sContent);
			return FAILURE;
		    }
		}
	    } else {
		NodeContent_Structure *pStructureContent =
			(NodeContent_Structure *) pNestedContent;
		if (pStructureContent->ctMembers != node->iSubNodeCount) {
		    sbAppend_int(
			    sbError,
			    "structure content member count %i does not match ",
			    pStructureContent->ctMembers);
		    sbAppend_int(sbError, "template sub-node count %i",
			    node->iSubNodeCount);
		    return FAILURE;
		}
		fputs("\n", stream);
		for (xSubNode = 1; xSubNode <= node->iSubNodeCount; xSubNode++) {
		    void *pSubNodeContent;
		    DAPNode *pSubNode;
		    NodeContent_Structure *pStructureContent =
			    (NodeContent_Structure *) pNestedContent;
		    pSubNode = *(node->nodeSub + xSubNode - 1);

		    if (*(pStructureContent->array_eNodeType + xSubNode - 1)
			    != pSubNode->eType) {
			sbAppend_int(
				sbError,
				"structure content member type %i does not match ",
				*(pStructureContent->array_eNodeType + xSubNode
					- 1));
			sbAppend_int(sbError, "template sub-node type %i",
				pSubNode->eType);
			return FAILURE;
		    }
		    pSubNodeContent = *(pStructureContent->array_NodeContent
			    + xSubNode - 1);
		    if (!print_NodeContent_toFile(pSubNode, pSubNodeContent,
			    iIndent + 1, stream, sbError)) {
			indent(iIndent, stream);
			fprintf(stream,
				"[Error printing structure sub-node %i: %s]\n",
				xSubNode, sbError->sContent);
			return FAILURE;
		    }
		}
	    }
	}
	indent(iIndent, stream);
    }

    fputs("} - ", stream);
    if (node->sName == NULL)
	fputs("[no name]", stream);
    else
	fputs(node->sName, stream);
    fputs("\n", stream);

    return SUCCESS;
}
/* Print the Grid to a file. This is called by
 * print_NodeContent_toFile, print_Sequence_toFile,
 * and the data print out section of print_Sequence_toFile */
eRESULT print_Grid_toFile(NodeContent_Grid * pGrid, int iIndent,
			  FILE * stream, StringBuffer * sbError)
{
    unsigned int xMapping;
    if (pGrid == NULL) {
	sbAppend(sbError, "  [error: grid missing]");
	return FAILURE;
    }
    fputs("\n", stream);
    indent(iIndent, stream);
    fprintf(stream, "--- Value Array: ");
    if (pGrid->values == NULL) {
	fprintf(stream, "[error: array missing]\n");
    } else {
	fprintf(stream, " %s", pGrid->values->sArrayName);
	if (print_Array_toFile(pGrid->values, iIndent, stream, sbError) ==
	    FAILURE) {
	    fprintf(stream, "[error in array: %s]\n", sbError->sContent);
	    sbClear(sbError);
	}
	fputs("\n", stream);
    }
    if (pGrid->ctMappings == 0) {
	fprintf(stream, "--- No Mapping Arrays Present\n");
    }
    for (xMapping = 1; xMapping <= pGrid->ctMappings; xMapping++) {
	indent(iIndent, stream);
	if (pGrid->maps1[xMapping] == NULL) {
	    fprintf(stream, "--- [error: array missing]\n");
	} else {
	    fprintf(stream, "--- Mapping Array (%d): ", xMapping);
	    fprintf(stream, " %s", pGrid->maps1[xMapping]->sArrayName);
	    if (!print_Array_toFile
		(pGrid->maps1[xMapping], iIndent, stream, sbError)) {
		fprintf(stream, "[error: %s]\n", sbError->sContent);
		sbClear(sbError);
	    }
	    fputs("\n", stream);
	}
    }
    fputs("\n", stream);
    return SUCCESS;
}

/* This is called in print_NodeContent_toFile, print_Grid_toFile, print_Sequence_toFile. */
/* This function needs to be extended to 'render' arrays of constructor types
 * (Structure, Sequence, and Grid). jhrg 12/02/08
 */
eRESULT print_Array_toFile(NodeContent_Array * pArray, int iIndent,
			   FILE * stream, StringBuffer * sbError)
{
    unsigned int xDim;

    /*  HEADER  */
    if (pArray == NULL) {
	sbAppend(sbError, "  [error: array missing]");
	return FAILURE;
    }

    fprintf(stream, "  dimensions: ");
    for (xDim = 1; xDim <= pArray->ctDimensions; xDim++) {
    	if (pArray->sDimensionName1[xDim] == NULL) {
    		fprintf(stream, "[%d]", pArray->iDimensionLengthArray1[xDim]);
    	}
    	else {
    		fprintf(stream, "[%s = ", pArray->sDimensionName1[xDim]);
    		fprintf(stream, "%d]", pArray->iDimensionLengthArray1[xDim]);
    	}
    }

    /* When the code and/or data structures are modified to handle arrays
     * of structures, et c., this can be used to call whatever function will
     * be used for those constructor types. jhrg 12/0/2/08
     */
/*    switch (pArray->eType) {
    case ARRAY_Byte:
    case ARRAY_Int16:
    case ARRAY_UInt16:
    case ARRAY_Int32:
    case ARRAY_UInt32:
    case ARRAY_Float32:
    case ARRAY_Float64:
    case ARRAY_String:
    case ARRAY_URL:
	fprintf(stream, "  value type: %s", render_ARRAY_TYPE(pArray->eType));
	break;
    case ARRAY_Structure:
    case ARRAY_Sequence:
    case ARRAY_Grid:
	fprintf(stream, "  value type: %s", render_ARRAY_TYPE(pArray->eType));
	break;
    default:
	fprintf(stream, "  value type: %s", render_ARRAY_TYPE(pArray->eType));
	break;
    }
*/
    fprintf(stream, "  value type: %s", render_ARRAY_TYPE(pArray->eType));

    if (pArray->array == NULL) {
    	fputs(" ", stream);	/* there should be a space between the end of the array def and the close bracket */
    	return SUCCESS;		/* no values */
    }

    /*  BODY  */

    /* values - each row is prefixed by its dimension address followed by colon */
    /* values are space-separated */
    {
	unsigned int *iDimCounter = malloc((pArray->ctDimensions + 1) * sizeof(unsigned int));	/* one-based */
	unsigned int xRow;
	unsigned int xColumn;
	unsigned int xArray;
	unsigned int ctRows = 1;
	unsigned int ctRowDimensions = pArray->ctDimensions - 1;
	unsigned int iColumnLength = pArray->iDimensionLengthArray1[pArray->ctDimensions];

	for (xDim = 1; xDim <= ctRowDimensions; xDim++)
	    iDimCounter[xDim] = 1;
	for (xDim = 1; xDim <= ctRowDimensions; xDim++)
	    ctRows = ctRows * pArray->iDimensionLengthArray1[xDim];
	for (xRow = 1; xRow <= ctRows; xRow++) {
	    unsigned int iRowOffset;
	    xArray = 1;
	    fputs("\n", stream);
	    indent(iIndent, stream);

	    for (xDim = 1; xDim <= ctRowDimensions; xDim++) {
	    	char *sFormat;
	    	unsigned int iDimSize = pArray->iDimensionLengthArray1[xDim];
	    	if (iDimSize < 10) {
	    		sFormat = "[%d]";
	    	}
	    	else if (iDimSize < 100) {
	    		sFormat = "[%02d]";
	    	}
	    	else if (iDimSize < 1000) {
	    		sFormat = "[%03d]";
	    	} else if (iDimSize < 10000) {
	    		sFormat = "[%04d]";
	    	} else if (iDimSize < 100000) {
	    		sFormat = "[%05d]";
	    	} else {
	    		sFormat = "[%06d]";
	    	}
	    	fprintf(stream, sFormat, iDimCounter[xDim]);
	    	xArray = xArray * iDimCounter[xDim];
	    }

	    iRowOffset = (xArray - 1) * iColumnLength;
	    fputs("[-]:  ", stream);	/* the row is the last dimension */

	    switch (pArray->eType) {
		/* Fix these format types. jhrg 4/6/07 */
	    case ARRAY_Byte:
		for (xColumn = 0; xColumn < iColumnLength; xColumn++)
		    fprintf(stream, "%02x ", *((DBYTE *) (pArray->array) + iRowOffset +
			      xColumn));
		break;
	    case ARRAY_Int16:
		for (xColumn = 0; xColumn < iColumnLength; xColumn++)
		    fprintf(stream, "%d ", *((DINT16 *) (pArray->array) + iRowOffset +
			      xColumn));
		break;
	    case ARRAY_UInt16:
		for (xColumn = 0; xColumn < iColumnLength; xColumn++)
		    fprintf(stream, "%u ", *((DUINT16 *) (pArray->array) + iRowOffset +
			      xColumn));
		break;
	    case ARRAY_Int32:
		for (xColumn = 0; xColumn < iColumnLength; xColumn++)
		    fprintf(stream, "%d ", *((DINT32 *) (pArray->array) + iRowOffset +
			      xColumn));
		break;
	    case ARRAY_UInt32:
		for (xColumn = 0; xColumn < iColumnLength; xColumn++)
		    fprintf(stream, "%u ", *((DUINT32 *) (pArray->array) + iRowOffset +
			      xColumn));
		break;
	    case ARRAY_Float32:
		for (xColumn = 0; xColumn < iColumnLength; xColumn++)
		    fprintf(stream, "%f ", *((DFLOAT32 *) (pArray->array) + iRowOffset +
			      xColumn));
		break;
	    case ARRAY_Float64:
		for (xColumn = 0; xColumn < iColumnLength; xColumn++)
		    fprintf(stream, "%f ", *((DFLOAT64 *) (pArray->array) + iRowOffset +
			      xColumn));
		break;
	    case ARRAY_String:
		for (xColumn = 0; xColumn < iColumnLength; xColumn++) {
		    NodeContent_String *pStringContent = *((NodeContent_String **) (pArray->array) +
			  iRowOffset + xColumn);
		    fprintf(stream, "\"%s\" ", pStringContent->sContent);
		}
		break;
	    default:
		/*fprintf(stream, "Unimplemented type");*/
		break;
		/* no output -- todo */
	    }
	    xDim = ctRowDimensions;
	    while (xDim > 0) {
		DBG(fprintf(stderr, "xDim: %u\n", xDim));

		if (iDimCounter[xDim] == pArray->iDimensionLengthArray1[xDim]) {
		    iDimCounter[xDim] = 1;
		    xDim--;
		}
		else {
		    iDimCounter[xDim]++;
		    break;
		}
	    }
	}
	free(iDimCounter);
    }
    return SUCCESS;
}
/*Refer to data.c about line 760 that defines how sequences are handled within the Ocapi.
 * This is a great resource for understanding the following function.
 * this is called in print_NodeContent_toFile*/
eRESULT print_Sequence_toFile(DAPNode * pSequenceNode, NodeContent_Sequence * pNestedContent,
			      int iIndent, FILE * stream, StringBuffer * sbError)
{
    unsigned int xField;
    unsigned int xRow;
    unsigned int ctRows;
    unsigned int ctFields_Template = pSequenceNode->iSubNodeCount;
    NodeContent_Sequence *pSequenceContent;

    if (pNestedContent == NULL) {
    	pSequenceContent = pSequenceNode->nodeContent;
    }
    else {
    	pSequenceContent = pNestedContent;
    }

    /* validate that sequence content (if supplied) matches up with the template */
    if (pSequenceContent != NULL) {
    	if (pSequenceContent->ctFields != ctFields_Template) {
    		sbAppend_int(sbError, "validation error, template field count %i, ",
    			ctFields_Template);
    		sbAppend_int(sbError, "does not match content field count %i, ",
    			pSequenceContent->ctFields);
	    return FAILURE;
    	}
    }

    if (DEBUG)
    	printf("printing sequence at %p with content at %p\n",
    			(void *) pSequenceNode, (void *) pSequenceContent);

    /* header */
    fprintf(stream, "  fields: %d", ctFields_Template);
    if (pSequenceContent->ctRows > 0)
    	fprintf(stream, "  rows: %d", pSequenceContent->ctRows);
		fputs("\n", stream);

    for (xField = 1; xField <= ctFields_Template; xField++) {	/* todo ripping through a blind double pointer array like this is risky -- it is key to add internal validation codes to the OCAPI */
		eNODE_TYPE eFieldType_node;
		eNODE_TYPE eFieldType_content;
		void *pValueVector;
		DAPNode *nodeField;
		nodeField = *(pSequenceNode->nodeSub + xField - 1);
		eFieldType_node = nodeField->eType;

		/* field information print out */
		indent(iIndent, stream);
		fprintf(stream, "field %u: %s  type: %s", xField, nodeField->sName,
			render_NODE_TYPE(eFieldType_node));

		/* data printout */
		if (pSequenceContent == NULL || pSequenceContent->aFieldValue == NULL) {	/* then this is a DDS, no data to print out */
			switch (eFieldType_node) {
			case NODE_Array:;
			{
				NodeContent_Array *pArrayContent = nodeField->nodeContent;
				if (!print_Array_toFile(pArrayContent, iIndent + 1, stream, sbError)) {
					fprintf(stream, "[error: %s]\n", sbError->sContent);
					sbClear(sbError);
				}
				break;
			}
			case NODE_Grid:;
			{
				NodeContent_Grid *pGridContent = nodeField->nodeContent;
				if (!print_Grid_toFile(pGridContent, iIndent + 1, stream, sbError)) {
					fprintf(stream, "[error: %s]\n", sbError->sContent);
					sbClear(sbError);
				}
				break;
			}
			case NODE_Sequence:;
			{
				if (!print_Sequence_toFile(nodeField, NULL, iIndent + 1, stream, sbError)) {
				fprintf(stream, "[error: %s]\n", sbError->sContent);
				sbClear(sbError);
				}
				break;
			}
			case NODE_Structure:;
			{
				fputs("\n", stream);	/* TODO newline placement is not consistent */
				if (!print_NodeContent_toFile(nodeField, NULL, iIndent + 1, stream, sbError)) {
					fprintf(stream, "[error: %s]\n", sbError->sContent);
					sbClear(sbError);
				}
				break;
			}
			default:
			fputs("\n", stream);	/* other types require no further information printed for DDS */
			}
			continue;
		}

		/* -------------------------
		 * DATA PRINTOUT
		 * ------------------------- */

		/* validate that the data type for the template matches that of the content structure */
		eFieldType_content = *(pSequenceContent->aeFieldType + xField - 1);
		if (eFieldType_content != eFieldType_node) {
			sbAppend_int(sbError, "validation error, in field %i, ", xField);
			sbAppend_string(sbError, "content field type %s, ", render_NODE_TYPE(eFieldType_content));
			sbAppend_string(sbError, "does not match template field type %s",
					render_NODE_TYPE(eFieldType_node));
			return FAILURE;
		}

		ctRows = pSequenceContent->ctRows;
		fputs("  values: ", stream);
		pValueVector = *(pSequenceContent->aFieldValue + xField - 1);

		if (pValueVector == NULL) {
			fputs(" [error: field vector is null]\n", stream);
			continue;
		}

		if (DEBUG)
			printf("field %i has value vector at %p\n", xField, pValueVector);

		switch (eFieldType_node) {
			case NODE_Byte:;
				{
				DBYTE *byte_vector = (DBYTE *) pValueVector;
				for (xRow = 1; xRow <= ctRows; xRow++)
					fprintf(stream, " (%i) %u", xRow, *(byte_vector + xRow - 1));
					fputs("\n", stream);
				}
				break;
			case NODE_Int16:;
				{
				DINT16 *int16_vector = (DINT16 *) pValueVector;
				for (xRow = 1; xRow <= ctRows; xRow++)
					fprintf(stream, " (%i) %d", xRow, *(int16_vector + xRow - 1));
					fputs("\n", stream);
				}
				break;
			case NODE_UInt16:;
				{
				DUINT16 *uint16_vector = (DUINT16 *) pValueVector;
				for (xRow = 1; xRow <= ctRows; xRow++)
					fprintf(stream, " (%i) %u", xRow, *(uint16_vector + xRow - 1));
					fputs("\n", stream);
				}
				break;
			case NODE_Int32:;
				{
				DINT32 *int32_vector = (DINT32 *) pValueVector;
				for (xRow = 1; xRow <= ctRows; xRow++)
					fprintf(stream, " (%i) %d", xRow, *(int32_vector + xRow - 1));
					fputs("\n", stream);
				}
				break;
			case NODE_UInt32:;
				{
				DUINT32 *uint32_vector = (DUINT32 *) pValueVector;
				for (xRow = 1; xRow <= ctRows; xRow++)
					fprintf(stream, " (%i) %u", xRow, *(uint32_vector + xRow - 1));
					fputs("\n", stream);
				}
				break;
			case NODE_Float32:;
				{
				DFLOAT32 *float32_vector = (DFLOAT32 *) pValueVector;
				for (xRow = 1; xRow <= ctRows; xRow++)
					fprintf(stream, " (%i) %f", xRow, *(float32_vector + xRow - 1));
					fputs("\n", stream);
				}
				break;
			case NODE_Float64:;
				{
				DFLOAT64 *float64_vector = (DFLOAT64 *) pValueVector;
				for (xRow = 1; xRow <= ctRows; xRow++)
					fprintf(stream, " (%i) %f", xRow, *(float64_vector + xRow - 1));
					fputs("\n", stream);
				}
				break;
			case NODE_String:;
			case NODE_URL:;
				{
				NodeContent_String **string_vector = (NodeContent_String **) pValueVector;
				fputs("\n", stream);
				for (xRow = 1; xRow <= ctRows; xRow++) {
					NodeContent_String *pStringContent = *(string_vector + xRow - 1);
					indent(iIndent + 1, stream);
					fprintf(stream, "row %d: \"", xRow);
					if (pStringContent == NULL) {
						fprintf(stream, "[error: row %i string value has null content]",
						xRow);
					}
					else {
					fprintf(stream, "%s", pStringContent->sContent);
					}
					fputs("\"\n", stream);
				}
				}
				break;
			case NODE_Array:;
				{
				NodeContent_Array **array_vector = (NodeContent_Array **) pValueVector;
				fputs("\n", stream);
				for (xRow = 1; xRow <= ctRows; xRow++) {
					NodeContent_Array *pContent_Array = *(array_vector + xRow - 1);
					indent(iIndent + 1, stream);
					fprintf(stream, "row %d:\n", xRow);
					if (!print_Array_toFile(pContent_Array, iIndent + 2, stream, sbError)) {
						fprintf(stream, "[error: %s]\n", sbError->sContent);
						sbClear(sbError);
					}
				}
				}
				break;
			case NODE_Grid:;
				{
				NodeContent_Grid **grid_vector = (NodeContent_Grid **) pValueVector;
				fputs("\n", stream);
				for (xRow = 1; xRow <= ctRows; xRow++) {
					NodeContent_Grid *pContent_Grid = *(grid_vector + xRow - 1);
					indent(iIndent + 1, stream);
					fprintf(stream, "row %d:\n", xRow);
					if (!print_Grid_toFile(pContent_Grid, iIndent + 2, stream, sbError)) {
						fprintf(stream, "[error: %s]\n", sbError->sContent);
						sbClear(sbError);
					}
				}
				}
				break;
			case NODE_Sequence:;
				{
				NodeContent_Sequence **sequence_vector = (NodeContent_Sequence **) pValueVector;
				if (DEBUG)
					printf("printing sequence vector %p\n", (void *) sequence_vector);
					fputs("\n", stream);
				for (xRow = 1; xRow <= ctRows; xRow++) {
					NodeContent_Sequence *pContent_Sequence = *(sequence_vector + xRow - 1);
					if (DEBUG)
						printf("printing sequence %p in row %u\n", (void *) pContent_Sequence, xRow);
						indent(iIndent + 1, stream);
						fprintf(stream, "row %d: ", xRow);
					if (!print_Sequence_toFile(nodeField, pContent_Sequence, iIndent + 2, stream,
					 	sbError)) {
						fprintf(stream, "[error: %s]\n", sbError->sContent);
						sbClear(sbError);
					}
				}
				}
				break;
			case NODE_Structure:;
				{
				NodeContent_Structure **structure_vector = (NodeContent_Structure **) pValueVector;
				fputs("\n", stream);
				for (xRow = 1; xRow <= ctRows; xRow++) {
					NodeContent_Structure *pContent_Structure = *(structure_vector + xRow - 1);
					indent(iIndent + 1, stream);
					/* Removed pContent_Structure below. jhrg 8/14/06 */
					fprintf(stream, "row %d:\n", xRow);
					if (print_Structure_toFile(nodeField, pContent_Structure, iIndent + 2,
						stream, sbError) == FAILURE) {
						indent(iIndent, stream);
						fprintf(stream, "Error printing structure: %s]\n", sbError->sContent);
						sbClear(sbError);
					}
				}
				}
				break;
			case NODE_root:;
			default:
			fputs("\n", stream);
			indent(iIndent, stream);
			sbAppend_int(sbError, "invalid field type [root] in sequence %d", eFieldType_node);
			return FAILURE;
		}
    }
    return SUCCESS;
}

eRESULT print_Structure_toFile(DAPNode * pStructureNode,
			       NodeContent_Structure * pStructureContent,
			       int iIndent, FILE * stream,
			       StringBuffer * sbError)
{
    return print_NodeContent_toFile(pStructureNode, pStructureContent,
				    iIndent, stream, sbError);
}

eRESULT print_DAS_toFile(DAPObject * pDAPObject, FILE * stream,
			 StringBuffer * sbError)
{
    if (pDAPObject == NULL) {
	sbAppend(sbError, "no DAS object supplied");
	return FAILURE;
    }
    if (stream == NULL)
	stream = stdout;
    if (pDAPObject->eStructureType != STRUCTURE_DAS) {
	sbAppend(sbError, "supplied object is not a DAS");
	return FAILURE;
    }
    return print_DASNode_toFile(pDAPObject->nodeRoot, 0, stream, sbError);
}

eRESULT print_DASNode_toFile(DAPNode * pStructureNode, int iIndent,
			     FILE * stream, StringBuffer * sbError)
{
    int xIndent;
    if (pStructureNode == NULL) {
	sbAppend(sbError, "no node supplied");
	return FAILURE;
    }
    for (xIndent = 1; xIndent <= iIndent; xIndent++) {
	fprintf(stdout, "\t");
    }
    fprintf(stream, "%s (%s) {\n", pStructureNode->sName,
	    render_NODE_TYPE(pStructureNode->eType));
    if (print_Attributes_toFile
	(pStructureNode->attributes, iIndent + 1,
	 stream) == SUCCESS) {
	unsigned int xSubNode;
	for (xSubNode = 1; xSubNode <= pStructureNode->iSubNodeCount;
	     xSubNode++) {
	    DAPNode *pSubNode = *(pStructureNode->nodeSub + xSubNode - 1);
	    if (pSubNode == NULL) {
		for (xIndent = 1; xIndent <= iIndent + 1; xIndent++) {
		    fprintf(stream, "\t");
		}
		fprintf(stream, "error, subnode %i is unexpectedly null\n",
			xSubNode);
	    } else {
		print_DASNode_toFile(pSubNode, iIndent + 1, stream,
				     sbError);
	    }
	}
    } else {
	/* This block will never be reached since
	   print_Afttributes_toFile() never returns an error. jhrg
	   4/3/07 */
	fprintf(stream, "\tError printing attributes: %s",
		sbError->sContent);
	sbClear(sbError);
    }
    fprintf(stream, "}\n");
    return SUCCESS;
}

eRESULT print_Attributes_toFile(DAPAttributeTable * pTable, int iIndent,
				FILE * stream)
{
    unsigned int xAttribute;
    if (pTable == NULL || pTable->ctAttributes == 0)
	return SUCCESS;
    for (xAttribute = 1; xAttribute <= pTable->ctAttributes; xAttribute++) {
	DAPAttribute *attribute =
	    *(pTable->pAttributeList + xAttribute - 1);
	if (attribute != NULL) {
	    fprintf(stream, "\n");
	    indent(iIndent, stream);
	    fprintf(stream, "[%s:", attribute->sName);
	    if (attribute->pValueList == NULL)
		fprintf(stream, " null]");
	    else
		switch (attribute->eType) {
		case ATTRIBUTE_alias:;
		    fputc('=', stream);	/* fall through and print string */
		case ATTRIBUTE_String:;
		case ATTRIBUTE_URL:;
		    {
			unsigned int xString;
			fprintf(stream, " \"%s\"",
				(char *) *(attribute->pValueList));
			for (xString = 2; xString <= attribute->ctValues;
			     xString++) {
			    fprintf(stream, ", \"%s\"",
				    (char *) *(attribute->pValueList +
					       xString - 1));
			}
			fprintf(stream, "]");
		    }
		    break;
		    /* no longer used
		       case ATTRIBUTE_container: ;
		       fputs("\n", stream);
		       if( !print_Attributes_toFile( attribute->value, iIndent + 1, stream, sbError ) ){
		       fputs("error printing attribute table: ", stream);
		       fputs(sbError->sContent, stream);
		       sbClear(sbError);
		       fputs("]\n", stream);
		       }
		       break;
		     */
		case ATTRIBUTE_Byte:;
		    {
			unsigned int xValue;
			fprintf(stream, " %i", *(DBYTE *) * (attribute->pValueList));	/* todo check for null */
			for (xValue = 2; xValue <= attribute->ctValues;
			     xValue++) {
			    fprintf(stream, ", \"%i\"",
				    *(DBYTE *) * (attribute->pValueList +
						  xValue - 1));
			}
			fprintf(stream, "]");
		    }
		    break;
		case ATTRIBUTE_Int16:;
		    {
			unsigned int xValue;
			fprintf(stream, " %i", *(DINT16 *) * (attribute->pValueList));	/* todo check for null */
			for (xValue = 2; xValue <= attribute->ctValues;
			     xValue++) {
			    fprintf(stream, ", \"%i\"",
				    *(DINT16 *) * (attribute->pValueList +
						   xValue - 1));
			}
			fprintf(stream, "]");
		    }
		    break;
		case ATTRIBUTE_UInt16:;
		    {
			unsigned int xValue;
			fprintf(stream, " %i", *(DUINT16 *) * (attribute->pValueList));	/* todo check for null */
			for (xValue = 2; xValue <= attribute->ctValues;
			     xValue++) {
			    fprintf(stream, ", \"%i\"",
				    *(DUINT16 *) * (attribute->pValueList +
						    xValue - 1));
			}
			fprintf(stream, "]");
		    }
		    break;
		case ATTRIBUTE_Int32:;
		    {
			unsigned int xValue;
			fprintf(stream, " %i", *(DINT32 *) * (attribute->pValueList));	/* todo check for null */
			for (xValue = 2; xValue <= attribute->ctValues;
			     xValue++) {
			    fprintf(stream, ", \"%i\"",
				    *(DINT32 *) * (attribute->pValueList +
						   xValue - 1));
			}
			fprintf(stream, "]");
		    }
		    break;
		case ATTRIBUTE_UInt32:;
		    {
			unsigned int xValue;
			fprintf(stream, " %i", *(DUINT32 *) * (attribute->pValueList));	/* todo check for null */
			for (xValue = 2; xValue <= attribute->ctValues;
			     xValue++) {
			    fprintf(stream, ", \"%i\"",
				    *(DUINT32 *) * (attribute->pValueList +
						    xValue - 1));
			}
			fprintf(stream, "]");
		    }
		    break;
		case ATTRIBUTE_Float32:;
		    {
			unsigned int xValue;
			fprintf(stream, " %g", *(DFLOAT32 *) * (attribute->pValueList));	/* todo check for null */
			for (xValue = 2; xValue <= attribute->ctValues;
			     xValue++) {
			    fprintf(stream, ", \"%g\"",
				    *(DFLOAT32 *) *
				    (attribute->pValueList + xValue - 1));
			}
			fprintf(stream, "]");
		    }
		    break;
		case ATTRIBUTE_Float64:;
		    {
			unsigned int xValue;
			fprintf(stream, " %g", *(DFLOAT64 *) * (attribute->pValueList));	/* todo check for null */
			for (xValue = 2; xValue <= attribute->ctValues;
			     xValue++) {
			    fprintf(stream, ", \"%g\"",
				    *(DFLOAT64 *) *
				    (attribute->pValueList + xValue - 1));
			}
			fprintf(stream, "]");
		    }
		    break;
		case ATTRIBUTE_unknown:;
		    fputs(" ---unknown attribute type---", stream);
		    break;
		default:
		    fputs(" ---undefined attribute type---", stream);
		    break;
		}
	}
    }
    fputs("\n", stream);
    return SUCCESS;
}

/************************/
/**  Internal Utility  **/
/************************/

char *render_STRUCTURE_TYPE(eSTRUCTURE_TYPE e)
{
    switch (e) {
    case STRUCTURE_DataDDS:
	return "DataDDS";
    case STRUCTURE_DDX:
	return "DDX";
    case STRUCTURE_DDS:
	return "DDS";
    case STRUCTURE_DAS:
	return "DAS";
    default:
	return "[undefined]";
    }
}

char *render_NODE_TYPE(eNODE_TYPE e)
{
    switch (e) {
    case NODE_root:
	return "root";
    case NODE_Byte:
	return "Byte";
    case NODE_Int16:
	return "Int16";
    case NODE_UInt16:
	return "UInt16";
    case NODE_Int32:
	return "Int32";
    case NODE_UInt32:
	return "UInt32";
    case NODE_Float32:
	return "Float32";
    case NODE_Float64:
	return "Float64";
    case NODE_String:
	return "String";
    case NODE_URL:
	return "URL";
    case NODE_Sequence:
	return "Sequence";
    case NODE_Array:
	return "Array";
    case NODE_Grid:
	return "Grid";
    case NODE_Structure:
	return "Structure";
    default:
	return "[undefined]";
    }
}

char *render_ATTRIBUTE_TYPE(eATTRIBUTE_TYPE e)
{
    switch (e) {
    case ATTRIBUTE_unknown:
	return "[unknown]";
    case ATTRIBUTE_alias:
	return "alias";
	/* no longer used case ATTRIBUTE_container: return "container"; */
    case ATTRIBUTE_Byte:
	return "Byte";
    case ATTRIBUTE_Int16:
	return "Int16";
    case ATTRIBUTE_UInt16:
	return "UInt16";
    case ATTRIBUTE_Int32:
	return "Int32";
    case ATTRIBUTE_UInt32:
	return "UInt32";
    case ATTRIBUTE_Float32:
	return "Float32";
    case ATTRIBUTE_Float64:
	return "Float64";
    case ATTRIBUTE_String:
	return "String";
    case ATTRIBUTE_URL:
	return "URL";
    default:
	return "[undefined]";
    }
}
/* Provides Array types*/
char *render_ARRAY_TYPE(eARRAY_TYPE e)
{
    switch (e) {
    case ARRAY_Byte:
	return "Byte";
    case ARRAY_Int16:
	return "Int16";
    case ARRAY_UInt16:
	return "UInt16";
    case ARRAY_Int32:
	return "Int32";
    case ARRAY_UInt32:
	return "UInt32";
    case ARRAY_Float32:
	return "Float32";
    case ARRAY_Float64:
	return "Float64";
    case ARRAY_String:
	return "String";
    case ARRAY_URL:
	return "URL";
    case ARRAY_Structure:
	return "Structure";
    case ARRAY_Sequence:
	return "Sequence";
    case ARRAY_Grid:
	return "Grid";
    default:
	return "[undefined (in ocapi.c:render_ARRAY_TYPE)]";
    }
}

void indent(int i, FILE * stream)
{
    int xTab;
    for (xTab = 1; xTab <= i; xTab++)
	fputc('\t', stream);
}

DAPNode *nodeCreate(eNODE_TYPE type)
{
    DAPNode *pNode = (DAPNode *) malloc(sizeof(DAPNode));
    if (pNode == NULL)
	return NULL;
    pNode->eType = type;
    pNode->nodeSuper = NULL;
    pNode->nodeSub = NULL;
    pNode->iSubNodeCount = 0;
    pNode->sName = NULL;
    pNode->nEstimatedSize = 0;
    pNode->attributes = NULL;
    pNode->nodeContent = NULL;

    switch (type) {
    case NODE_root:
    case NODE_Byte:
    case NODE_Int16:
    case NODE_UInt16:
    case NODE_Int32:
    case NODE_UInt32:
    case NODE_Float32:
    case NODE_Float64:
    case NODE_List:
    case NODE_Structure:
	pNode->nodeContent = NULL;
	break;

    case NODE_Sequence:;
	{
	    NodeContent_Sequence *pContent_sequence;
	    pContent_sequence = (NodeContent_Sequence *)
		malloc(sizeof(NodeContent_Sequence));
	    pContent_sequence->aeFieldType = NULL;
	    pContent_sequence->aFieldValue = NULL;
	    pContent_sequence->ctFields = 0;
	    pContent_sequence->ctRows = 0;
	    pNode->nodeContent = pContent_sequence;
	    break;
	}
    case NODE_String:;
    case NODE_URL:;
	{
	    NodeContent_String *pContent_string =
		(NodeContent_String *) malloc(sizeof(NodeContent_String));
	    pContent_string->iLength = 0;
	    pContent_string->sContent = NULL;
	    pNode->nodeContent = pContent_string;
	    break;
	}
    case NODE_Array:;
	{
	    pNode->nodeContent = arrayCreate();
	    if (pNode->nodeContent == NULL)
		return NULL;
	    break;
	}
    case NODE_Grid:;
	{
	    NodeContent_Grid *pContent_grid =
		malloc(sizeof(NodeContent_Grid));
	    if (pContent_grid == NULL)
		return NULL;
	    pContent_grid->ctMappings = 0;
	    pContent_grid->values = NULL;
	    pContent_grid->maps1 = NULL;
	    pNode->nodeContent = pContent_grid;
	    break;
	}
    }
    return pNode;
}

NodeContent_Array *arrayCreate()
{
    NodeContent_Array *pContent_array =
	(NodeContent_Array *) malloc(sizeof(NodeContent_Array));
    if (pContent_array == NULL)
	return NULL;
    pContent_array->eType = ARRAY_Int32;
    pContent_array->ctDimensions = 0;
    pContent_array->array = NULL;
    pContent_array->iDimensionLengthArray1 = NULL;
    pContent_array->sDimensionName1 = NULL;
    return pContent_array;
}

/** Adds a variable node to another of the OPeNDAP structure
 *  Note that capacity is increased in increments of 16 so whenever a node is added to a leaf
 *  then space for 16 pointers is added and once the node count reaches 17 another 16 is added
 *  and so on.
 */
eRESULT node_AddNode(DAPNode * pNode_super, DAPNode * pNode_sub,
		     StringBuffer * sbError)
{
    int iNewSize;
    DAPNode **hSubNodes;
#if 0
    unsigned int xExistingSubNodes;
    int xUnusedNodeSlots;
#endif
    if (pNode_super == NULL) {
	sbAppend(sbError, "internal error, super node is null");
	return FAILURE;
    }
    if (pNode_sub == NULL) {
	sbAppend(sbError, "internal error, sub node is null");
	return FAILURE;
    }
    if (			/* unsigned jhrg 4/3/07 pNode_super->iSubNodeCount < 0 || */
	   pNode_super->iSubNodeCount > 100000) {
	sbAppend(sbError, "internal error, invalid sub node count");
	return FAILURE;
    }
    if (pNode_super->iSubNodeCount % 16 == 0) {	/* need to add storage for 16 more nodes */
	iNewSize = pNode_super->iSubNodeCount + 16;
	hSubNodes = malloc(sizeof(DAPNode *) * iNewSize);
	if (hSubNodes == NULL) {
	    sbAppend(sbError, "Failed to allocate memory for sub node(s)");
	    return FAILURE;
	}

	memcpy(hSubNodes, pNode_super->nodeSub, pNode_super->iSubNodeCount);
	memset(hSubNodes + pNode_super->iSubNodeCount, 0, iNewSize - pNode_super->iSubNodeCount * sizeof(hSubNodes));

	if (pNode_super->nodeSub == NULL) {
	    /* node is being added to childless parent node */
	} else {
	    free(pNode_super->nodeSub);	/* free the old sub node buffer */
	}
	pNode_super->nodeSub = hSubNodes;
    }
    pNode_super->iSubNodeCount++;
    *(pNode_super->nodeSub + pNode_super->iSubNodeCount - 1) = pNode_sub;
    pNode_sub->nodeSuper = pNode_super;

    /* if the sub node is a sequence, its content array must be initialized since we
     * now know how many sub-nodes/fields it has and their type */
    if (pNode_sub->eType == NODE_Sequence) {
	unsigned int xField;
	NodeContent_Sequence *pNodeContent = pNode_sub->nodeContent;
	pNodeContent->ctFields = pNode_sub->iSubNodeCount;
	pNodeContent->aeFieldType =
	    malloc(sizeof(int *) * pNodeContent->ctFields);
	if (pNodeContent->aeFieldType == NULL) {
	    sbAppend(sbError,
		     "Failed to allocate memory for sequence field types.");
	    return FAILURE;
	}
	for (xField = 0; xField < pNodeContent->ctFields; xField++) {
	    DAPNode *pNodeField = *(pNode_sub->nodeSub + xField);
	    *(pNodeContent->aeFieldType + xField) = pNodeField->eType;
	}
    }
    return SUCCESS;
}

DAPAttributeTable *attribute_NewTable(StringBuffer * sbError)
{
    DAPAttributeTable *pTable = malloc(sizeof(DAPAttributeTable));
    if (pTable == NULL) {
	sbAppend(sbError,
		 "failed to allocate memory for new attribute table");
	return NULL;
    }
    pTable->ctAttributes = 0;
    pTable->pNameList = NULL;
    pTable->pAttributeList = NULL;
    return pTable;
}

/** Adds an attribute to an attribute table; manages memory allocation for the dynamic lists.
 *  Note that capacity is increased in increments of 16 so whenever an attribute is added
 *  then space for 16 pointers is added and once the node count reaches 17 another 16 is added
 *  and so on.
 *
 *  sName and pAttribute must be on the heap
 */
eRESULT attribute_Add(DAPAttributeTable * pTable, char *sName,
		      DAPAttribute * pAttribute, StringBuffer * sbError)
{
    int iNewSize;
    char **pNameList;
    DAPAttribute **pAttributeList;
#if 0
    unsigned int xExistingAttributes;
    int xUnusedAttributeSlots;
#endif

    if (pTable == NULL) {
	sbAppend(sbError, "internal error, attribute table is null");
	return FAILURE;
    }
    if (sName == NULL) {
	sName = sCloneString("[unnamed]");
    }
    if (pAttribute == NULL) {
	sbAppend(sbError, "internal error, attribute is null");
	return FAILURE;
    }
    if ( /* pTable->ctAttributes < 0 || unsigned. jhrg 4/2/2007 */ pTable->
	ctAttributes > 10000) {
	sbAppend_int(sbError,
		     "internal error, invalid attribute count (%i)",
		     pTable->ctAttributes);
	return FAILURE;
    }
    if (pTable->ctAttributes % 16 == 0) {	/* need to add storage for 16 more attributes */
	iNewSize = pTable->ctAttributes + 16;
	pNameList = malloc(sizeof(char *) * iNewSize);
	if (pNameList == NULL) {
	    sbAppend(sbError,
		     "failed to allocate memory for attribute name list");
	    return FAILURE;
	}
	pAttributeList = malloc(sizeof(DAPAttribute *) * iNewSize);
	if (pAttributeList == NULL) {
	    sbAppend(sbError,
		     "failed to allocate memory for attribute list");
	    return FAILURE;
	}
	if (pTable->pNameList != NULL) {	/* clone and free the old name list */

		memcpy(pNameList, pTable->pNameList, pTable->ctAttributes);
		memcpy(pAttributeList, pTable->pAttributeList, pTable->ctAttributes);
	    memset(pNameList + pTable->ctAttributes, 0, iNewSize - pTable->ctAttributes * sizeof(pNameList));
	    memset(pAttributeList + pTable->ctAttributes, 0, iNewSize - pTable->ctAttributes * sizeof(pAttributeList));
	    free(pTable->pNameList);	/* free the old name list */
	    free(pTable->pAttributeList);	/* free the old attribute list */
	}
	pTable->pNameList = pNameList;
	pTable->pAttributeList = pAttributeList;
    }
    pTable->ctAttributes++;
    *(pTable->pNameList + pTable->ctAttributes - 1) = sName;
    *(pTable->pAttributeList + pTable->ctAttributes - 1) = pAttribute;


    return SUCCESS;
}

/* copies name (assumes it is on the stack)
 * Is called from clone_Attributes*/
DAPAttribute *attribute_New(eATTRIBUTE_TYPE eType, char *sStackName,
			    StringBuffer * sbError)
{
    DAPAttribute *attribNew;
    attribNew = malloc(sizeof(DAPAttribute));
    if (attribNew == NULL) {
	sbAppend(sbError, "failed to allocate memory for new attribute");
	return FAILURE;
    }
    attribNew->eType = eType;
    attribNew->sName = malloc(strlen(sStackName) + 1);
    {
	unsigned int xChar;
	for (xChar = 0; xChar < strlen(sStackName); xChar++)
	    *(attribNew->sName + xChar) = *(sStackName + xChar);
	*(attribNew->sName + xChar) = 0;	/* null terminator for string */
    }
    attribNew->ctValues = 0;
    attribNew->pValueList = NULL;
    return attribNew;
}

/** Adds a value to an attribute; manages memory allocation for the dynamic lists.
 *  Note that capacity is increased in increments of 16 so whenever an attribute is added
 *  then space for 16 pointers is added and once the node count reaches 17 another 16 is added
 *  and so on.
 *  Called from clone_Attributes.
 */
eRESULT attribute_AddValue(DAPAttribute * pAttribute, void *pStackValue,
			   StringBuffer * sbError)
{
    void *pHeapValue;
    int iNewSize;
    void **pValueList;
#if 0
    unsigned int xExistingValues;
    int xUnusedValueSlots;
#endif

    if (pAttribute == NULL) {
	sbAppend(sbError, "internal error, attribute for value is null");
	return FAILURE;
    }
    if (pStackValue == NULL) {
	sbAppend(sbError, "internal error, value is null");
	return FAILURE;
    }
    /* pAttribute->ctValues < 0 || unsigned. jhrg 4/2/2007 */
    if (pAttribute->ctValues > 1000) {
	sbAppend_int(sbError, "internal error, invalid value count (%i)",
		     pAttribute->ctValues);
	return FAILURE;
    }
    /* need to add storage for 16 more attributes */
    if (pAttribute->ctValues % 16 == 0) {
	iNewSize = pAttribute->ctValues + 16;
	pValueList = malloc(sizeof(void *) * iNewSize);
	if (pValueList == NULL) {
	    sbAppend(sbError,
		     "failed to allocate memory for attribute value list");
	    return FAILURE;
	}
	if (pAttribute->pValueList != NULL) {
	/* copy existing values to new memory */
		memcpy(pValueList, pAttribute->pValueList, pAttribute->ctValues);
	    /* zero out all the new space */
	    memset (pValueList + pAttribute->ctValues, 0, iNewSize - pAttribute->ctValues * sizeof(pValueList));
	    /* free the old value list */
	    free(pAttribute->pValueList);
	}
	pAttribute->pValueList = pValueList;
    }
    /*Allocate memory for the new attribute types*/
    switch (pAttribute->eType) {
    case ATTRIBUTE_alias:;
    case ATTRIBUTE_String:;
    case ATTRIBUTE_URL:;
	{
	    unsigned int xChar;
	    char *sValue = (char *) pStackValue;
	    pHeapValue = malloc(strlen(sValue) + 1);
	    if (pHeapValue == NULL) {
		sbAppend(sbError,
			 "failed to allocate memory for new attribute string value");
		return FAILURE;
	    }
	    for (xChar = 0; xChar < strlen(sValue); xChar++)
		*((char *) pHeapValue + xChar) = *(sValue + xChar);
	    *((char *) pHeapValue + xChar) = 0;	/* null terminator for string */
	}
	break;
    case ATTRIBUTE_Byte:;
	pHeapValue = malloc(1);
	if (pHeapValue == NULL) {
	    sbAppend(sbError,
		     "failed to allocate memory for new byte value");
	    return FAILURE;
	}
	*((char *) pHeapValue) = *((char *) pStackValue);
	break;
    case ATTRIBUTE_Int16:;
    case ATTRIBUTE_UInt16:;
	pHeapValue = malloc(2);
	if (pHeapValue == NULL) {
	    sbAppend(sbError,
		     "failed to allocate memory for new short value");
	    return FAILURE;
	}
	*((char *) pHeapValue) = *((char *) pStackValue);
	*((char *) pHeapValue + 1) = *((char *) pStackValue + 1);
	break;
    case ATTRIBUTE_Int32:;
	pHeapValue = malloc(sizeof(int));
	if (pHeapValue == NULL) {
	    sbAppend(sbError,
		     "failed to allocate memory for new int/float value");
	    return FAILURE;
	}
	*((int *) pHeapValue) = *((int *) pStackValue);
	break;
    case ATTRIBUTE_UInt32:;
	pHeapValue = malloc(sizeof(unsigned int));
	if (pHeapValue == NULL) {
	    sbAppend(sbError,
		     "failed to allocate memory for new unsigned int value");
	    return FAILURE;
	}
	*((unsigned int *) pHeapValue) = *((unsigned int *) pStackValue);
	break;
    case ATTRIBUTE_Float32:;
	pHeapValue = malloc(sizeof(float));
	if (pHeapValue == NULL) {
	    sbAppend(sbError,
		     "failed to allocate memory for new int/float value");
	    return FAILURE;
	}
	*((float *) pHeapValue) = *((float *) pStackValue);
	break;
    case ATTRIBUTE_Float64:;
	pHeapValue = malloc(sizeof(double));
	if (pHeapValue == NULL) {
	    sbAppend(sbError,
		     "failed to allocate memory for new int/float value");
	    return FAILURE;
	}
	*((double *) pHeapValue) = *((double *) pStackValue);
	break;
    case ATTRIBUTE_unknown:;
    default:
	sbAppend(sbError, "unknown attribute value type");
	return FAILURE;
    }
    pAttribute->ctValues++;
    *(pAttribute->pValueList + pAttribute->ctValues - 1) = pHeapValue;
    return SUCCESS;
}

DAPNode *getNodeByName(DAPObject * structure, char *sQualifiedNodeName)
{
    StringList *name_list;
    if (structure == NULL || sQualifiedNodeName == NULL)
	return NULL;		/* obviously it would be good to test for validity */
    name_list = parseQualifiedName(sQualifiedNodeName);
    if (name_list == NULL) {
#if 0
	fprintf(stderr,
		"internal error, attempt to parse [%s] unexpectedly returned null\n",
		sQualifiedNodeName);
#endif
	return NULL;
    }
    if (name_list->ctStrings == 0) {
#if 0
	fprintf(stderr,
		"internal error or bad input, attempt to parse [%s] returned no strings\n",
		sQualifiedNodeName);
#endif
	return NULL;
    }
    return getNodeByName_Recursive(structure->nodeRoot, name_list, 1);
}

DAPNode *getNodeByName_Recursive(DAPNode * nodeParent,
				 StringList * bufferQualifiedName,
				 int xName1)
{
    unsigned int xSubNode;
    char *sNameSought = bufferQualifiedName->asStrings[xName1 - 1];	/* asArguments is 0-based */
    if (sNameSought == NULL) {
#if 0
	fprintf(stderr,
		"internal error, name argument %i is unexpectedly null\n",
		xName1);
#endif
	return NULL;
    }
    for (xSubNode = 1; xSubNode <= nodeParent->iSubNodeCount; xSubNode++) {
	DAPNode *nodeCurrent;
	char *sNameCurrent;
	nodeCurrent = *(nodeParent->nodeSub + xSubNode - 1);
	sNameCurrent = nodeCurrent->sName;
	if (sNameCurrent == NULL)
	    continue;
	if (strcmp(sNameSought, sNameCurrent) == 0) {
	    if (xName1 == bufferQualifiedName->ctStrings)
		return nodeCurrent;	/* found it */
	    return getNodeByName_Recursive(nodeCurrent, bufferQualifiedName, xName1 + 1);	/* we have a partial match keep drilling down */
	} else {
	    if (bufferQualifiedName->ctStrings == 1) {	/* if a simple name was supplied search the whole tree */
		DAPNode *nodeSimpleSearchResult =
		    getNodeByName_Recursive(nodeCurrent,
					    bufferQualifiedName, 1);
		if (nodeSimpleSearchResult == NULL) {
		    /* continue looking in more nodes */
		} else {
		    return nodeSimpleSearchResult;	/* found it */
		}
	    } else {
		/* continue on to the next sub since we need an exact qualified match */
	    }
	}
    }
    return NULL;		/* the name was not found */
}

void *getSequenceVector(DAPNode * nodeSequenceField, char **sError)
{
    DAPNode *nodeParent;
    NodeContent_Sequence *contentParentSequence;
    unsigned int xField;
    if (nodeSequenceField == NULL) {
	*sError = "supplied sequence field was null";
	return NULL;
    }
    nodeParent = nodeSequenceField->nodeSuper;
    if (nodeParent == NULL) {
	*sError =
	    "supplied node is not a node representing a field inside a sequence, parent is null";
	return NULL;
    }
    if (nodeParent->eType != NODE_Sequence) {
	*sError =
	    "supplied node is not a node representing a field inside a sequence, parent is not a sequence";
	return NULL;
    }
    contentParentSequence = nodeParent->nodeContent;
    if (contentParentSequence == NULL) {
	*sError =
	    "supplied node's parent sequence has no content; it may be a dds or may be a nested sequence";
	return NULL;
    }
    for (xField = 0; xField < nodeParent->iSubNodeCount; xField++) {
	if (*(nodeParent->nodeSub + xField) == nodeSequenceField) {
	    if (*(contentParentSequence->aeFieldType + xField) !=
		nodeSequenceField->eType) {
		*sError =
		    "validation error, content field type does not match node type";
		return NULL;
	    }
	    return contentParentSequence->aFieldValue + xField;
	}
    }
    *sError =
	"invalid structure, sequence field was not found in parent node";
    return NULL;
}

int getRowCount(DAPNode * nodeSequence, char **sError)
{
    NodeContent_Sequence *pSequenceContent;
    if (nodeSequence == NULL) {
	*sError = "supplied sequence field was null";
	return -1;
    }
    if (nodeSequence->eType != NODE_Sequence) {
	*sError = "supplied node is not a node representing a sequence";
	return -1;
    }
    if (nodeSequence->nodeContent == NULL) {
	*sError =
	    "supplied node has no content; it may be a dds or may be a nested sequence";
	return -1;
    }
    pSequenceContent = nodeSequence->nodeContent;
    return pSequenceContent->ctRows;
}

int getFieldCount(DAPNode * nodeSequence, char **sError)
{
    NodeContent_Sequence *pSequenceContent;
    if (nodeSequence == NULL) {
	*sError = "supplied sequence field was null";
	return -1;
    }
    if (nodeSequence->eType != NODE_Sequence) {
	*sError =
	    "supplied node is not a node representing a sequence, parent is not a sequence";
	return -1;
    }
    if (nodeSequence->nodeContent == NULL) {
	*sError =
	    "supplied node has no content; it may be a dds or may be a nested sequence";
	return -1;
    }
    pSequenceContent = nodeSequence->nodeContent;
    return pSequenceContent->ctFields;
}

void *getSequenceVector_Nested(DAPNode * nodeSequenceField,
			       NodeContent_Sequence *
			       contentParentSequence, char **sError)
{
    DAPNode *nodeParent;
    unsigned int xField;
    if (nodeSequenceField == NULL) {
	*sError = "supplied sequence field was null";
	return NULL;
    }
    nodeParent = nodeSequenceField->nodeSuper;
    if (nodeParent == NULL) {
	*sError =
	    "supplied node is not a node representing a field inside a sequence, parent is null";
	return NULL;
    }
    if (nodeParent->eType != NODE_Sequence) {
	*sError =
	    "supplied node is not a node representing a field inside a sequence, parent is not a sequence";
	return NULL;
    }
    if (contentParentSequence == NULL) {
	*sError = "the content for the parent sequence was null";
	return NULL;
    }
    for (xField = 0; xField < nodeParent->iSubNodeCount; xField++) {
	if (nodeSequenceField == *(nodeParent->nodeSub + xField))
	    break;		/* found the field number */
	if (xField == nodeParent->iSubNodeCount - 1) {
	    *sError =
		"invalid structure, could not find sequence field inside of its parent";
	    return NULL;
	}
    }
    if (nodeSequenceField->eType !=
	*(contentParentSequence->aeFieldType + xField)) {
	*sError =
	    "parameter inconsistence, the field node type did not match the content type";
	return NULL;
    }
    return contentParentSequence->aFieldValue + xField;
}

/* field/row is expected to be 1-based */
void *getSequenceValue(DAPNode * nodeSequence, unsigned int field1,
		       unsigned int row1, char **sError)
{
    unsigned int ctFields;
    unsigned int ctRows;
    void **pFieldVector;
    eNODE_TYPE eNodeType;
    NodeContent_Sequence *pSequenceContent;
    if (nodeSequence == NULL) {
	*sError = "supplied sequence was null";
	return NULL;
    }
    pSequenceContent = (NodeContent_Sequence *) nodeSequence->nodeContent;
    if (pSequenceContent == NULL) {
	*sError = "sequence has no content, might be a DDS";
	return NULL;
    }
    ctFields = nodeSequence->iSubNodeCount;
    ctRows = pSequenceContent->ctRows;
    if (ctFields == 0) {
	*sError = "sequence has no fields";
	return NULL;
    }
    if (field1 == 0) {
	*sError =
	    "field number was 0, must be between 1 and the number of fields";
	return NULL;
    }
    if (field1 > ctFields) {
	*sError = "field number exceeded field count";
	return NULL;
    }
    if (row1 == 0) {
	*sError =
	    "row number was 0, must be between 1 and the number of fields";
	return NULL;
    }
    if (row1 > ctRows) {
	*sError = "row number exceeded row count";
	return NULL;
    }
    eNodeType = *(pSequenceContent->aeFieldType + field1 - 1);
    pFieldVector = *(pSequenceContent->aFieldValue + field1 - 1);
    if (DEBUG)
	printf
	    ("field vector is: %p, value pointer is %p, and points to %p\n",
	     (void *) pFieldVector, (void *) (pFieldVector + row1 - 1),
	     *(pFieldVector + row1 - 1));
    if (zIsPrimitive(eNodeType)) {
	switch (eNodeType) {
	case NODE_Byte:
	    return (DBYTE *) pFieldVector + row1 - 1;
	case NODE_Int16:
	    return (DINT16 *) pFieldVector + row1 - 1;
	case NODE_UInt16:
	    return (DUINT16 *) pFieldVector + row1 - 1;
	case NODE_Int32:
	    return (DINT32 *) pFieldVector + row1 - 1;
	case NODE_UInt32:
	    return (DUINT32 *) pFieldVector + row1 - 1;
	case NODE_Float32:
	    return (DFLOAT32 *) pFieldVector + row1 - 1;
	case NODE_Float64:
	    return (DFLOAT64 *) pFieldVector + row1 - 1;
	default:
	    *sError = "unknown primitive type";
	    return NULL;
	}
    } else {
	return *(pFieldVector + row1 - 1);
    }
}

void *getSequenceValue_Nested(DAPNode * nodeSequence,
			      /* not used jhrg 4/3/07 NodeContent_Sequence* nodeContent, unsigned int field1, unsigned int row1, */
			      char **sError)
{
    if (nodeSequence == NULL) {
	*sError = "supplied sequence was null";
	return NULL;
    }
    return NULL;
}

int zIsPrimitive(eNODE_TYPE node_type)
{
    switch (node_type) {
    case NODE_Byte:;
    case NODE_Int16:;
    case NODE_UInt16:;
    case NODE_Int32:;
    case NODE_UInt32:;
    case NODE_Float32:;
    case NODE_Float64:;
	return 1;
    default:
	return 0;
    }
}

eRESULT zFile_LoadStructure(const char *sFilename, DAPObject **hStructure, StringBuffer *sbError)  {

    Buffer *pInputBuffer;
    ParserBuffer *pParserBuffer;
    char *sServerError;
    eRESULT result;
    int posDataHeader;
    int posData;
    int posDataLength;
    DAPNode *nodeRoot;

    if (sFilename == NULL)
      {
        sbAppend(sbError, "internal error, no file name provided");
        return FAILURE;
      }
    if (hStructure == NULL)
      {
        sbAppend(sbError, "internal error, no structure handle provided");
        return FAILURE;
      }

    pInputBuffer = NULL;
    pParserBuffer = malloc(sizeof(ParserBuffer));
    if (pParserBuffer == NULL)
      {
        sbAppend(sbError, "unable to create parser buffer, out of memory?");
        /* printf("freeing input buffer2: %p", pInputBuffer); */
        imp_buffer_Free(pInputBuffer /*, sbError */); /* ignore any errors -- todo generate warning */
        return FAILURE;
      }
    pParserBuffer->iStatus = TRUE;
#if 0
    pParserBuffer->eGridMode = GM_undefined;
#endif
    pParserBuffer->eReadMode = RM_Buffered;
    pParserBuffer->pOStructure = NULL;
    pParserBuffer->sError = NULL;
    pParserBuffer->pInputBuffer = NULL;
    pParserBuffer->eStructureType = STRUCTURE_undefined;
    pParserBuffer->pCurlRequest = NULL;

    pInputBuffer = bufferLoadFile(sFilename);
    if (pInputBuffer == NULL)
      {
        sbAppend(sbError, "failed to load file");
        goto FAIL;
      }

    /* determine if the file contains an error rather than data */
    sServerError = findError(pInputBuffer->pcharContent
        + pInputBuffer->iContentPosition0); /* todo don't use the raw buffer */
    if (sServerError != NULL)
      {
        sbAppend_string(sbError, "Server returned error: %s", sServerError);
        goto FAIL;
      }

    pParserBuffer->pInputBuffer = pInputBuffer;

    posDataHeader = indexOf_binary(pInputBuffer->pcharContent,
	    pInputBuffer->iContentLength, DATA_HEADER, strlen(DATA_HEADER));

    posData = posDataHeader + strlen(DATA_HEADER);
    posDataLength = pInputBuffer->iContentLength - posData + 1; /* remember how much data there is so far */

    /* This can likely be replaced or removed and the code in dap.y that
     * looks at the initial text in the input buffer used to decide if the
     * input is a DDS or DAS usied instead. jhrg 8/23/08 */

    /* read the structural header (DDS/DAS) */
    if (bufferStartsWith(pInputBuffer, "DATASET"))
      {
        pParserBuffer->eStructureType = STRUCTURE_DDS;
      }
    else if (bufferStartsWith(pInputBuffer, "ATTRIBUTES"))
      {
        pParserBuffer->eStructureType = STRUCTURE_DAS;
      }
    else
      {
        sbAppend(sbError,
            "Unknown buffer content start. Expected 'DATASET' or 'ATTRIBUTES'.");
        goto FAIL;
      }

    if (pParserBuffer->eStructureType == STRUCTURE_DDS)
      {
        int parser_return_value;
        if (DEBUG)
          printf("parsing DDS buffer\n");
        parser_return_value = parse_buffer((void *) pParserBuffer);
        if (parser_return_value == 0)
          { /* bison parser returns 0 on success, 1 on failure */
            if (pParserBuffer->pOStructure == NULL)
              {
                sbInsert_string(
                    sbError,
                    "Internal error, parser returned true but structure does not exist (%s)",
                    sFilename);
                goto FAIL;
              }
            else
              {
                pParserBuffer->pOStructure->eStructureType = STRUCTURE_DDS; /* todo support other object types */
                *hStructure = pParserBuffer->pOStructure;
                if(posDataHeader == 0) goto SUCCEED;
              }
          }
        else
          {
            sbInsert_string(sbError, "Error parsing: %s", pParserBuffer->
            sError);
            goto FAIL;
          }
      }
    else if (pParserBuffer->eStructureType == STRUCTURE_DAS)
      {
        if (zParseDAS(hStructure, pParserBuffer->pInputBuffer, sbError) == SUCCESS)
          {
            goto SUCCEED;
          }
        else
          {
            sbInsert(sbError, "Error parsing: ");
            goto FAIL;
          }
      }
    else
      {
        sbAppend(sbError, "unable to determine structure type");
        goto FAIL;
      }

    nodeRoot = pParserBuffer->pOStructure->nodeRoot;
    if (nodeRoot == NULL)
      {
        sbAppend(sbError, "about to read data but root node unexpectedly null");
        goto FAIL;
      }
    /*
     printf("buffer content %p:\n", pInputBuffer);
     bufferDump( pInputBuffer, stdout, 1, 500 );
     */

    pInputBuffer->iContentPosition0 = posData-1; // remember: 1 based
    pInputBuffer->iContentLength = posDataLength;

    if (read_Node(nodeRoot, NULL, pInputBuffer, sbError) == SUCCESS)
      {
        pParserBuffer->pOStructure->eStructureType = STRUCTURE_DataDDS;
        *hStructure = pParserBuffer->pOStructure;
        goto SUCCEED;
      }
    else
      {
        sbInsert(sbError, "error reading data: ");
        goto FAIL;
      }

    FAIL: result = FAILURE;
    goto FINALLY;
    SUCCEED: result = SUCCESS;
    goto FINALLY;
    FINALLY:
    /* printf("freeing input buffer3: %p", pInputBuffer); */
    imp_buffer_Free(pInputBuffer /*, sbError */); /* ignore any errors -- todo generate warning */
    if (pParserBuffer != NULL)
      free(pParserBuffer);
    return result;
  }

void vFreeConnectionParameters(ConnectionParameters * parameters) {
	/*	Proxy* pstructProxy;
	 URL*   pstructURL;
	 */
	/* if( parameters->sProtocol != NULL ) free( parameters->sProtocol ); do not free this -- it is a global string by default */
	if (parameters->sUserAgent != NULL)
		free(parameters->sUserAgent);
	if (parameters->sClientHeader != NULL)
		free(parameters-> sClientHeader);
	if (parameters->pstructProxy != NULL)
		if (parameters->cook != NULL)
				free(parameters->cook);
		free(parameters->pstructProxy);
	if (parameters != NULL)
		free(parameters);


}
