/*
 * This file is part of the "Ocapi" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005 OPeNDAP, Inc.
 * Author: Dennis Heimbigner
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */

#include "parse.h"

extern int dapdebug;

StringBuffer* sbError;
extern ParserBuffer* dapParserState;

int dap_parse_init(ParserBuffer*);
int isattribute(void*);

/* This is the wrapper around dapparse for dds*/
int
ddsparse(void* arg)
{
    ParserBuffer* buf = (ParserBuffer*)arg;
    if(sbError == NULL) sbError = sbNew(NULL);
    if(!dap_parse_init(buf)) return 1;
    if(dapparse() == 1) return 1;
    if(stackDepth(dapstack) != 1) {
        dap_parse_error("Stack length != 1",NULL,0);
	return 1;
    }
    buf->pOStructure = (DAPObject*)stackPop(dapstack);
    /* If the data hdr tag ("Data:") was encountered,*/
    /* then change the root type*/
    if(dap_datahdr && buf->pOStructure->eStructureType == STRUCTURE_DDS)
	buf->pOStructure->eStructureType = STRUCTURE_DataDDS;
    return 0;
}

/* Wrapper for DAS parsing*/
eRESULT
zParseDAS(DAPObject** rootp , Buffer *pBuffer, StringBuffer *sberr )
{
    if(sbError != NULL) sbFree(sbError);
    sbError = sberr;
    if(dapParserState == NULL)
	dapParserState = (ParserBuffer*)malloc(sizeof(ParserBuffer));
/*This buffer will have been freed by caller
*   else if(dapParserState->pInputBuffer != NULL)
*	imp_buffer_Free(dapParserState->pInputBuffer);*/
    dapParserState->pInputBuffer = pBuffer;
    if(ddsparse((void*)dapParserState) == 1) return FAILURE;
    *rootp = dapParserState->pOStructure;
    return SUCCESS;
}

int
dap_parse_init(ParserBuffer *parser_state)
{
    dapParserState = parser_state;
    if( parser_state->pInputBuffer == NULL ){
        dap_parse_error("parser state has no input buffer", NULL, 0);
        dap_cleanup();
	return 0;
    }
    if( sbError == NULL ){
        dap_parse_error("failed to create new error buffer", NULL, 0);
        dap_cleanup();
        return 0;
    }
    parser_state->iStatus = TRUE;
#if 0
    parser_state->eGridMode = GM_undefined;
#endif
    parser_state->sError = NULL;
    return 1;
}


void
dap_parse_error(char* msg, char* token, int lineno)
{
    fprintf(stderr,"parse error: %s: line %d",(msg?msg:"error"),lineno);
    if(token != NULL) fprintf(stderr,"; token=%s",token);
    fprintf(stderr,"\n");
}

void
free_Attribute(DAPAttribute* attr)
{
    unsigned int i;
    if(attr->sName != NULL) free(attr->sName);
    for(i=0;i<attr->ctValues;i++) {
	void* vp = attr->pValueList[i];
	if(vp != NULL) free(vp);
    }
    if(attr->pValueList != NULL) free(attr->pValueList);
}

void
dap_cleanup(void)
{
    /* Clear out the stacks and reclaim space*/
    int i;
    for(i=0;i<stackDepth(dapdimnames);i++) {
	char* s = (char*)stackPop(dapdimnames);
	if(s != NULL) free(s);
    }
    /* the primary stack may have either DAPNodes or DAPAttributes*/
    for(i=0;i<stackDepth(dapstack);i++) {
	void* p = (void*)stackPop(dapstack);
	if(isattribute(p)) {
	    DAPAttribute* attr = (DAPAttribute*)p;
	    free_Attribute(attr);
	} else {
	    imp_node_Free((DAPNode*)p,sbError);
	}
    }
    /* the daptemp stack may have DAPNodes*/
    for(i=0;i<stackDepth(daptemp);i++) {
	DAPNode* node = (DAPNode*)stackPop(daptemp);
	imp_node_Free(node,sbError);
    }
    /* the dapvalues stack may have strings*/
    for(i=0;i<stackDepth(dapvalues);i++) {
	char* s = (char*)stackPop(dapvalues);
	if(s != NULL) free(s);
    }
    dapParserState = NULL;
}

int OCAPI_Input( char* buf, unsigned int max_size ){
	Buffer* pBuffer;
	int iCharsAvailable;
	if( dapParserState == NULL ){
		fputs( "no parser state\n", stderr );
		return 0;
	}
	pBuffer = dapParserState->pInputBuffer;
	if( pBuffer->iContentPosition0 == 0 ){
		vResetScanner(); /* reset to initial state on new buffer */
	}
	iCharsAvailable = max_size < pBuffer->iContentLength ? max_size : pBuffer->iContentLength;
	if( iCharsAvailable > 0 ){
		memcpy( buf, pBuffer->pcharContent + pBuffer->iContentPosition0, iCharsAvailable );
		pBuffer->iContentPosition0 += iCharsAvailable;
		pBuffer->iContentLength    -= iCharsAvailable;
	}
	return iCharsAvailable;
}

