/*
 * This file is part of the "Ocapi" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005 OPeNDAP, Inc.
 * Author: Dennis Heimbigner <jchamberlain at opendap.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "stack.h"

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define DEFAULTALLOC 16
#define ALLOCINCR 16

#define DATANULL ((elem_t)0)

int stackSetalloc(Stack* stack, long sz);

Stack* stackNew(long initial)
{
  Stack* stack = (Stack*)malloc(sizeof(Stack));
  if(stack == NULL) return NULL;
  stack->alloc=0;
  stack->length=0;
  stack->content=NULL;
  if(initial > 0) stackSetalloc(stack,initial);
  return stack;
}

int stackFree(Stack* stack)
{
  if(!stack) return FALSE;
  stack->length=0;
  stack->alloc=0;
  if(stack->content != NULL) free(stack->content);
  stack->content=NULL;
  free(stack);
  return TRUE;
}

int stackSetalloc(Stack* stack, long sz)
{
  elem_t* newcontent;
  if(stack == NULL) return FALSE;
  if(sz <= 0) sz=DEFAULTALLOC;
  if(stack->alloc >= sz) return TRUE;
  newcontent=(elem_t*)calloc(sz,sizeof(elem_t));
  if(stack->alloc > 0 && stack->length > 0 && stack->content != NULL) {
	  memmove(stack->content, newcontent, sizeof(elem_t)*stack->length);
	  free(stack->content);
  }
  stack->content=newcontent;
  stack->alloc=sz;
  return TRUE;
}

int stackClear(Stack* stack)
{
  if(stack == NULL) return FALSE;
  stack->length = 0;
  return TRUE;
}

int stackDepth(Stack* stack)
{
  if(stack == NULL) return -1;
  return stack->length;
}

int stackSetdepth(Stack* stack, long sz)
{
  if(stack == NULL || sz < 0) return FALSE;
  if(!stackSetalloc(stack,sz)) return FALSE;
  stack->length = sz;
  return TRUE;
}


elem_t stackGet(Stack* stack, long index)
{
  if(stack == NULL) return DATANULL;
  if(index < 0 || index >= stack->length) return DATANULL;
  return stack->content[index];
}

int stackPush(Stack* stack, elem_t elem)
{
  if(stack == NULL) return FALSE;
  if(stack->length >= stack->alloc)
	stackSetalloc(stack,stack->alloc+ALLOCINCR);
  stack->content[stack->length] = elem;
  stack->length++;
  return TRUE;
}

elem_t stackPop(Stack* stack)
{
  if(stack == NULL) return DATANULL;
  if(stack->length == 0) return DATANULL;
  stack->length--;
  return stack->content[stack->length];
}

elem_t stackTop(Stack* stack)
{
  if(stack == NULL) return DATANULL;
  if(stack->length == 0) return DATANULL;
  return stack->content[stack->length - 1];
}

int
zTestStack(){
    Stack* pStack = stackNew(50);
    /* I added this because I was tired of seeing the warnings from gcc
     * about pointer sizes. The tests should really include verifying that
     * long and void * are the same size, but since this is test code...
     * jhrg 5/25/06 */
#if SIZEOF_VOID_P == 4
    unsigned int xItem, xPop;
#else
    unsigned long xItem, xPop;
#endif
    if( pStack == NULL ) return FALSE;
    /* int to pointer cast may fail on 64-bit machines. jhrg 2/1/06 */
    for( xItem = 1; xItem <= 10000; xItem++ )
	stackPush( pStack, (elem_t)xItem );
    if( ((elem_t)stackTop( pStack )) != 10000 ) return FALSE;
    for( xPop = 10000; xPop > 0; xPop-- ){
        elem_t pObject = stackPop( pStack );
        if( pObject == ELEMNULL ) return FALSE;
        /* Pointer to int cast may fail on 64-bit machines. jhrg 2/1/06 */
        if( pObject != (elem_t)xPop ) return FALSE;
    }
    if( stackTop( pStack ) != ELEMNULL ) return FALSE;
    if( stackDepth(pStack) != 0 ) return FALSE;
    return TRUE;
}

