/*
 * This file is part of the "OCAPI" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005 OPeNDAP, Inc.
 * Author: John Chamberlain <jchamberlain at opendap.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "utility.h"
#include "buffers.h"

/*Turn this on so the DBG statements will print for debugging purposes--remove if do not want the printout*/
/*#define DODS_DEBUG*/

#include "debug.h"

/* not case sensitive, returns TRUE or FALSE, NULL sensitive */
int equals(char *s1, char *s2)
{
    unsigned int len;
    register unsigned int pos;
    if (s1 == NULL || s2 == NULL)
	return s1 == NULL && s2 == NULL ? TRUE : FALSE;
    len = strlen(s1);
    if (strlen(s2) != len)
	return FALSE;
    for (pos = 0; pos < len; pos++) {
	if (toupper(*(s1 + pos)) != toupper(*(s2 + pos)))
	    return FALSE;
    }
    return TRUE;
}

char *toUpper(char *s)
{
    register int pos;
    int lenS;
    char *sUpper;
    if (s == NULL)
	return NULL;
    lenS = strlen(s);
    sUpper = malloc(lenS + 1);	/* accomodate null terminator */
    if (sUpper == NULL)
	return NULL;		/* out of memory */
    for (pos = 0; pos < lenS; pos++) {
	*(sUpper + pos) = toupper(*(s + pos));
    }
    *(sUpper + lenS) = 0;	/* null terminator */
    return sUpper;
}

/* finds the index of the first instance of sSearchString in sBuffer
 * - not case-sensitive
 * - one-based, returns 0 if the search string is not found
 */
int indexOf(char *sBuffer, char *sSearchString)
{
    DBG(fprintf
	(stderr, "sBuffer: xxx%sxxx (%u)\n", sBuffer,
	 (unsigned) strlen(sBuffer)));
    DBG(fprintf
	(stderr, "sSearchString: xxx%sxxx (%u)\n", sSearchString,
	 (unsigned) strlen(sSearchString)));
    return indexOf_binary(sBuffer, strlen(sBuffer), sSearchString,
			  strlen(sSearchString));
}
/* Finds the index binary of the sSearchString with lenSearchString
 * in sBuffer with lenBuffer. Called from indexOf.*/
int indexOf_binary(char *sBuffer, unsigned int lenBuffer,
		   char *sSearchString, unsigned int lenSearchString)
{
    unsigned int posBuffer;
    unsigned int posSearchString;
    unsigned int posMatchLocation;
    if (sBuffer == NULL || sSearchString == NULL)
	return 0;
    posBuffer = 1;
    posSearchString = 1;
    posMatchLocation = 0;
    while (posBuffer <= lenBuffer) {
	char cBuffer = *(sBuffer + posBuffer - 1);
	char cSearchString = *(sSearchString + posSearchString - 1);
/*The following 4 lines are being completed in order to make the two comparisons case insenstive within ASCII. rd 06/16/08*/
	if (cBuffer > 96 && cBuffer < 123)
	    cBuffer -= 32;
	if (cSearchString > 96 && cSearchString < 123)
	    cSearchString -= 32;
	if (cBuffer == cSearchString) {
	    if (posMatchLocation == 0)
		posMatchLocation = posBuffer;
	    if (posSearchString == lenSearchString)
		return posMatchLocation;
	    posSearchString++;
	} else {
	    if (posMatchLocation > 0) {
		posBuffer = posMatchLocation;
		posMatchLocation = 0;
	    }
	    posSearchString = 1;
	}
	posBuffer++;
    }
    return 0;
}

/*  splits a command string into white space-separated values
 *  quotation marks are respected; use \" to escape
 *  if the string is null, null is returned
 *  if the string is the empty string an array containing only one
 *  element, the empty string, is returned
 *  the CommandBuffer structure is defined in utility.h
 */
CommandBuffer *parseCommand(char *s)
{
    StringBuffer *sbBuffer;
    unsigned int ctElements;
    unsigned int len;
    unsigned int eState;
    int zFilling;
    char **as;

    if (s == NULL)
	return NULL;

    len = strlen(s);
    if (len == 0)
	return NULL;

    sbBuffer = sbNew_size(120);
    if (sbBuffer == NULL)
	return NULL;

    ctElements = 0;
    /* before word */
    eState = 1;
    /* first size the content array then in second pass fill it */
    zFilling = FALSE;
    as = NULL;
    while (TRUE) {
		unsigned int pos;
		char c;
		for (pos = 0; pos < len; pos++) {
			c = *(s + pos);
			switch (eState) {
				case 1:		/* before word */
				if (c == '"') {	/* quoted string */
					ctElements++;
					eState = 3;
				} else if (c == '\\') {
					ctElements++;	/* no matter what's next, its a word */
					eState = 4;
				} else if (isspace(c)) {
					/* ignore */
				}
				else {	/* in unquoted word */
					ctElements++;
					if (zFilling)
					sbAppend_char(sbBuffer, c);
					eState = 2;	/* in word */
				}
				break;

				case 2:		/* in unquoted word */
				if (isspace(c)) {	/* found end of word */
					if (zFilling) {
						unsigned int xArrayIndex;
						char *sWord;
						xArrayIndex = ctElements - 1;
						sWord = sbCopyContent(sbBuffer);
						as[xArrayIndex] = sWord;
						sbBuffer->iContentLength = 0;
					}
					eState = 1;	/* after word */
				} else if (c == '\\') {
					eState = 4;
				} else {
					if (zFilling) {
						sbAppend_char(sbBuffer, c);	/* still in word */
					}
				}
				break;

				case 3:		/* in quoted string */
				if (c == '"') {	/* found end of string */
					if (zFilling) {
						as[ctElements - 1] = sbCopyContent(sbBuffer);
						sbBuffer->iContentLength = 0;
					}
					eState = 1;	/* after word */
				} else if (c == '\\') {
					eState = 5;	/* after backslash in quoted string */
				} else {
					if (zFilling)
					sbAppend_char(sbBuffer, c);	/* still in string */
				}
				break;

				case 4:		/* after backslash in word */
				if (zFilling) {
					if (c == '"') {	/* literal quotation mark */
						sbAppend_char(sbBuffer, c);
					}
					else {	/* plain backslash */
						sbAppend_char(sbBuffer, '\\');
						sbAppend_char(sbBuffer, c);
					}
				}
				eState = 2;	/* in word */
				break;

				case 5:		/* after backslash in string */
				if (zFilling) {
					if (c == '"') {	/* literal quotation mark */
						sbAppend_char(sbBuffer, c);
					}
					else {	/* plain backslash */
						sbAppend_char(sbBuffer, '\\');
						sbAppend_char(sbBuffer, c);
					}
				}
				eState = 3;	/* in string */
				break;
			}
		}
		/*fill the content array*/
		if (zFilling) {
			CommandBuffer *pCB;

			/* terminal conditions */
			switch (eState) {
				case 1:		/* before word */
				/* normal terminatiion */
				break;

				case 2:		/* in unquoted word, end of word reached */
				as[ctElements - 1] = sbCopyContent(sbBuffer);
				sbBuffer->iContentLength = 0;
				break;

				case 3:		/* in quoted string, string is unterminated, accept anyway */
				as[ctElements - 1] = sbCopyContent(sbBuffer);
				sbBuffer->iContentLength = 0;
				break;

				case 4:
					/* after backslash in word, accept anyway */
				case 5:		/* after backslash in string, input error, but accept anyway */
				sbAppend_char(sbBuffer, '\\');
				as[ctElements - 1] = sbCopyContent(sbBuffer);
				sbBuffer->iContentLength = 0;
				break;
			}

			pCB = malloc(sizeof(CommandBuffer));

			if (pCB == NULL)
			return NULL;	/* out of memory */

			pCB->sCommand = as[0];
			pCB->asArguments = as;
			pCB->ctArguments = ctElements - 1;
			sbFree(sbBuffer);
			return pCB;		/* we are done */
		}
		/* size array and start filling pass */
		else {
			unsigned int xArrayZero;
			as = malloc(ctElements * sizeof(char *));
			for (xArrayZero = 0; xArrayZero < ctElements; xArrayZero++)
			as[xArrayZero] = NULL;
			if (as == NULL) {
				sbFree(sbBuffer);
				return NULL;	/* out of memory */
			}
			zFilling = TRUE;
			ctElements = 0;
			eState = 1;
		}
    }
}

void vFreeCommandBuffer(CommandBuffer * pCB)
{
    int xArgument;
    if (pCB == NULL)
	return;
    if (pCB->sCommand != NULL) {
	free(pCB->sCommand);
    }
    if (pCB->asArguments != NULL) {	/* note that arg 1 will be the command (already freed above) */
	for (xArgument = 1; xArgument <= pCB->ctArguments; xArgument++) {
	    if (pCB->asArguments + xArgument != NULL) {
		free(*(pCB->asArguments + xArgument));
	    }
	}
	free(pCB->asArguments);
	pCB->asArguments = NULL;
    }
    free(pCB);
}

/* clones a string on the heap */
char *sCloneString(const char *s)
{
    int lenS, xChar;
    char *sNew;
    if (s == NULL)
	return NULL;
    lenS = strlen(s);
    sNew = malloc(lenS + 1);
    if (sNew == NULL)
    return NULL;
    for (xChar = 0; xChar < lenS; xChar++)
	*(sNew + xChar) = *(s + xChar);
    *(sNew + xChar) = 0;	/* null terminator for string */
    return sNew;
}

/*  splits a string into dot-separated values
 *  if the string is null, null is returned
 *  if the string is the empty string an array containing only one
 *  element, the empty string, is returned
 */
StringList *parseQualifiedName(char *s)
{
    StringBuffer *sbBuffer;
    int ctElements, len, eState, zFilling;
    char **as = NULL;
    DBG(fprintf(stderr, "parseQualifiedName: %s\n", s));
    if (s == NULL)
	return NULL;
    len = strlen(s);
    if (len == 0)
	return NULL;
    sbBuffer = sbNew_size(120);
    if (sbBuffer == NULL)
	return NULL;
    ctElements = 0;
    eState = 1;			/* before word */
    zFilling = FALSE;		/* first size the content array then in second pass fill it */
    while (TRUE) {
	int pos;
	for (pos = 0; pos < len; pos++) {
	    char c = *(s + pos);
	    switch (eState) {
	    case 1:		/* before word */
		if (c == '.') {
		    /* ignore leading dots */
		} else if (isspace(c)) {
		    /* ignore */
		} else {	/* in unquoted word */
		    ctElements++;
		    if (zFilling)
			sbAppend_char(sbBuffer, c);
		    eState = 2;	/* in word */
		}
		break;
	    case 2:		/* in unquoted word */
		if (c == '.') {	/* found end of word */
		    if (zFilling) {
			as[ctElements - 1] = sbCopyContent(sbBuffer);
			sbBuffer->iContentLength = 0;
		    }
		    eState = 1;	/* before word */
		} else {
		    if (zFilling) {
			sbAppend_char(sbBuffer, c);	/* still in word */
		    }
		}
		break;
	    }
	}

	if (zFilling) {		/* resolve terminal conditions */
	    switch (eState) {
	    case 1:		/* before word */
		/* the string ends with a dot */
		break;
	    case 2:		/* in unquoted word, end of word reached */
		as[ctElements - 1] = sbCopyContent(sbBuffer);
		sbBuffer->iContentLength = 0;
		break;
	    }
	}

	if (zFilling) {
	    StringList *pStringList;
	    /* In the following I changed 'sizeof(pStringList)' to
	     * 'sizeof(StringList)' jhrg 8/15/06 */
	    pStringList = malloc(sizeof(StringList));
	    if (pStringList == NULL)
		return NULL;	/* out of memory */
	    pStringList->asStrings = as;
	    pStringList->ctStrings = ctElements;
	    sbFree(sbBuffer);
	    return pStringList;	/* we are done */
	} else {		/* size array and start filling pass */

	    as = malloc(ctElements * sizeof(char *));
	    if (as == NULL) {
		sbFree(sbBuffer);
		return NULL;	/* out of memory */
	    }

	    zFilling = TRUE;
	    ctElements = 0;
	    eState = 1;
	}
    }
}

/* returns the error in the buffer if it is found, NULL otherwise */
char *findError(char *sBuffer)
{
    int posError, posMessage, posTermination, iMessageLength;
    char *sMessageCopy;
    posError = indexOf(sBuffer, "Error {");
    /*do not need to check for all three conditions because if posError returns 0 the other two will as well. rd 06/16/08*/

    if (posError == 0) /* || posMessage == 0 || posTermination == 0)*/
	return NULL;

	posMessage = indexOf(sBuffer, "message = \"");
    posTermination = indexOf(sBuffer, "\";\n};");

    if (!(posError < posMessage && posMessage < posTermination))
	return NULL;

    posMessage = posMessage + 11;	/* move start past initial quote */
    iMessageLength = posTermination - posMessage;
    sMessageCopy = malloc(iMessageLength + 1);
    if (sMessageCopy == NULL)
	return "[out of memory copying message]";
    strncpy(sMessageCopy, sBuffer + posMessage - 1, iMessageLength);
    *(sMessageCopy + iMessageLength) = 0;
    return sMessageCopy;
}
